// *****************************************************************************
// Build settings
// *****************************************************************************

import java.time.*

// The current year is used to update the license copyright headers in the  sources.
val currentYear: Int = OffsetDateTime.now().getYear
// The initial starting year of the project.
val projectStartYear: Int = 2022

addCommandAlias("check", "Compile/scalafix --check; Test/scalafix --check; headerCheckAll; scalafmtCheckAll")
addCommandAlias("fix", "Compile/scalafix; Test/scalafix; headerCreateAll; scalafmtSbt; scalafmtAll")

// Enable the semanticdb compiler plugin needed by the metals language server.
Global / semanticdbEnabled := true

inThisBuild(
    Seq(
        scalaVersion     := "3.3.3",
        organization     := "de.smederee",
        organizationName := "Contributors as noted in the AUTHORS.md file",
        version          := "0.13.0-SNAPSHOT",
        scalacOptions ++= Seq(
            "-deprecation",
            "-explain",
            "-explain-types",
            "-feature",
            "-language:higherKinds",
            "-language:implicitConversions",
            "-no-indent",  // Prevent usage of indent based syntax.
            "-old-syntax", // Enforce classic syntax.
            "-unchecked",
            "-Wunused:imports",             // Warn on unused imports including given and wildcard imports.
            "-Wunused:linted",              // TODO: Find out what this does!
            "-Wunused:locals",              // Warn on unused local definitions.
            "-Wunused:nowarn",              // Warn on unused (useless) `@nowarn` annotations.
            "-Wunused:params",              // Warn on unused parameters.
            "-Wunused:privates",            // Warn on unused private definitions.
            "-Wunused:unsafe-warn-patvars", // TODO: Find out what this does!
            "-Wvalue-discard",              // Warn on discarding computed values.
            // "-Xfatal-warnings", // FIXME: Make this work despite of Twirl!
            "-Ykind-projector"
        ),
        resolvers += "jitpack" at "https://jitpack.io", // for JANSI fork
        Compile / console / scalacOptions --= Seq("-Xfatal-warnings"),
        Test / console / scalacOptions --= Seq("-Xfatal-warnings"),
        Test / fork              := true,
        Test / parallelExecution := false,
        Test / testOptions += Tests.Argument(TestFrameworks.MUnit, "-b")
    )
)

// *****************************************************************************
// Projects
// *****************************************************************************

lazy val smederee =
    project
        .in(file("."))
        .settings(commonSettings)
        .settings(
            name         := "smederee",
            publish      := {},
            publishLocal := {}
        )
        .aggregate(darcs, email, htmlUtils, hub, i18n, security, twirl)

lazy val darcs =
    project
        .in(file("modules/darcs"))
        .enablePlugins(AutomateHeaderPlugin)
        .settings(commonSettings)
        .settings(
            name := "darcs",
            scalacOptions += "-Xfatal-warnings",
            libraryDependencies ++= Seq(
                library.catsCore,
                library.catsEffect,
                library.logback,
                library.osLib,
                library.munit           % Test,
                library.munitCatsEffect % Test,
                library.munitScalaCheck % Test,
                library.scalaCheck      % Test
            )
        )

lazy val email =
    project
        .in(file("modules/email"))
        .enablePlugins(AutomateHeaderPlugin)
        .settings(commonSettings)
        .settings(
            name := "email",
            scalacOptions += "-Xfatal-warnings",
            libraryDependencies ++= Seq(
                library.catsCore,
                library.catsEffect,
                library.ip4sCore,
                library.logback,
                library.simpleJavaMail,
                library.munit           % Test,
                library.munitCatsEffect % Test,
                library.munitScalaCheck % Test,
                library.scalaCheck      % Test
            )
        )

lazy val htmlUtils =
    project
        .in(file("modules/html-utils"))
        .enablePlugins(AutomateHeaderPlugin)
        .settings(commonSettings)
        .settings(
            name := "html-utils",
            scalacOptions += "-Xfatal-warnings",
            libraryDependencies ++= Seq(
                library.catsCore,
                library.commonMark,
                library.commonMarkExtHeadingAnchor,
                library.commonMarkExtImageAttrs,
                library.commonMarkExtStrikethrough,
                library.commonMarkExtTables,
                library.commonMarkExtTaskListItems,
                library.http4sCore,
                library.ip4sCore,
                library.logback,
                library.munit           % Test,
                library.munitScalaCheck % Test,
                library.scalaCheck      % Test
            )
        )

lazy val hub =
    project
        .in(file("modules/hub"))
        .dependsOn(darcs, email, htmlUtils, i18n, security, twirl)
        .enablePlugins(
            AutomateHeaderPlugin,
            BuildInfoPlugin,
            DebianPlugin,
            JavaServerAppPackaging,
            JDebPackaging,
            RpmPlugin,
            SbtTwirl,
            SystemdPlugin
        )
        .settings(commonSettings)
        .settings(
            name             := "smederee-hub",
            buildInfoKeys    := Seq[BuildInfoKey](name, version, scalaVersion),
            buildInfoPackage := "de.smederee.hub",
            libraryDependencies ++= Seq(
                library.apacheSshdCore,
                library.apacheSshdSftp,
                library.apacheSshdScp,
                library.bouncyCastleProvider,
                library.catsCore,
                library.circeCore,
                library.circeGeneric,
                library.circeParser,
                library.doobieCore,
                library.doobieHikari,
                library.doobiePostgres,
                library.ed25519Java, // FIXME: Replace with BouncyCastle once implemented in Mina-SSHD!
                library.flywayCore,
                library.flywayPostgreSQL,
                library.http4sCirce,
                library.http4sCore,
                library.http4sDsl,
                library.http4sEmberClient,
                library.http4sEmberServer,
                // library.http4sTwirl,
                library.jansi,
                library.jclOverSlf4j, // Bridge Java Commons Logging to SLF4J.
                library.log4catsSlf4j,
                library.logback,
                library.osLib,
                library.postgresql,
                library.pureConfig,
                library.springSecurityCrypto,
                library.munit           % Test,
                library.munitCatsEffect % Test,
                library.munitScalaCheck % Test,
                library.scalaCheck      % Test
            ),
            TwirlKeys.templateImports ++= Seq(
                "cats.*",
                "cats.data.*",
                "cats.syntax.all.*",
                "de.smederee.html.*",
                "de.smederee.i18n.*",
                "de.smederee.security.{ CsrfToken, UserId, Username }",
                "org.http4s.Uri"
            )
        )
        .settings(
            Seq(
                daemonUser  := "smederee",
                daemonGroup := "smederee",
                Debian / debianPackageProvides += "smederee-hub",
                Debian / debianPackageDependencies += "openjdk-17-jre-headless",
                defaultLinuxInstallLocation := "/usr/local/share",
                maintainer                  := "Wegtam GmbH <devops@wegtam.com>",
                rpmLicense                  := Option("AGPL-3.0 or later"),
                rpmVendor                   := "Wegtam GmbH <devops@wegtam.com>",
                // Require a service restart after installation / update.
                Debian / maintainerScripts := maintainerScriptsAppend((Debian / maintainerScripts).value)(
                    DebianConstants.Postinst -> s"restartService ${normalizedName.value}"
                ),
                // Require a service restart after installation / update.
                Rpm / maintainerScripts := maintainerScriptsAppend((Rpm / maintainerScripts).value)(
                    RpmConstants.Post -> s"restartService ${normalizedName.value}"
                ),
                packageSummary := "Smederee Hub Service - Software collaboration platform.",
                packageDescription := "Leverage the power of the darcs vcs to handle your projects with ease and confidence, this is the central hub service",
                // Debian / requiredStartFacilities := Option("$local_fs $remote_fs $network $postgresql"),
                // Do not package API docs.
                Compile / packageDoc / publishArtifact := false,
                Compile / doc / sources                := Seq.empty,
                // Prevent a customised local application.conf file to be packaged!
                Compile / packageBin / mappings ~= { files =>
                    files.filterNot { case (_, name) =>
                        name == "application.conf"
                    }
                }
            )
        )

lazy val i18n =
    project
        .in(file("modules/i18n"))
        .enablePlugins(AutomateHeaderPlugin)
        .settings(commonSettings)
        .settings(
            name := "i18n",
            scalacOptions += "-Xfatal-warnings",
            libraryDependencies ++= Seq(
                library.catsCore,
                library.logback,
                library.munit           % Test,
                library.munitScalaCheck % Test,
                library.scalaCheck      % Test
            )
        )

lazy val security =
    project
        .in(file("modules/security"))
        .enablePlugins(AutomateHeaderPlugin)
        .settings(commonSettings)
        .settings(
            name := "security",
            scalacOptions += "-Xfatal-warnings",
            libraryDependencies ++= Seq(
                library.bouncyCastleProvider % Runtime,
                library.catsCore,
                library.jclOverSlf4j, // Bridge Java Commons Logging to SLF4J.
                library.logback,
                library.springSecurityCrypto,
                library.munit           % Test,
                library.munitScalaCheck % Test,
                library.scalaCheck      % Test
            )
        )

// FIXME: This is a workaround until http4s-twirl gets published properly for Scala 3!
lazy val twirl =
    project
        .in(file("twirl"))
        .enablePlugins(SbtTwirl)
        .settings(commonSettings)
        .settings(
            name    := "twirl",
            version := "0.1.0",
            libraryDependencies += library.http4sCore
        )

// *****************************************************************************
// Library dependencies
// *****************************************************************************

lazy val library =
    new {
        object Version {
            val apacheSshd      = "2.12.1"
            val bouncyCastle    = "1.78"
            val cats            = "2.10.0"
            val catsEffect      = "3.5.4"
            val circe           = "0.14.6"
            val commonMark      = "0.22.0"
            val doobie          = "1.0.0-RC5"
            val flyway          = "10.11.0"
            val fs2             = "3.5.0"
            val http4s          = "1.0.0-M41"
            val ip4s            = "3.5.0"
            val jansi           = "2.4.2"
            val jclOverSlf4j    = "2.0.13"
            val log4cats        = "2.6.0"
            val logback         = "1.5.6"
            val munit           = "1.0.0-M12"
            val munitCatsEffect = "2.0.0-M5"
            val osLib           = "0.10.0"
            val postgresql      = "42.7.3"
            val pureConfig      = "0.17.6"
            val scalaCheck      = "1.18.0"
            val simpleJavaMail  = "8.8.3"
            val springSecurity  = "6.2.4"
        }
        val apacheSshdCore             = "org.apache.sshd"  % "sshd-core"                        % Version.apacheSshd
        val apacheSshdSftp             = "org.apache.sshd"  % "sshd-sftp"                        % Version.apacheSshd
        val apacheSshdScp              = "org.apache.sshd"  % "sshd-scp"                         % Version.apacheSshd
        val bouncyCastleProvider       = "org.bouncycastle" % "bcprov-jdk15to18"                 % Version.bouncyCastle
        val catsCore                   = "org.typelevel"   %% "cats-core"                        % Version.cats
        val catsEffect                 = "org.typelevel"   %% "cats-effect"                      % Version.catsEffect
        val circeCore                  = "io.circe"        %% "circe-core"                       % Version.circe
        val circeGeneric               = "io.circe"        %% "circe-generic"                    % Version.circe
        val circeParser                = "io.circe"        %% "circe-parser"                     % Version.circe
        val commonMark                 = "org.commonmark"   % "commonmark"                       % Version.commonMark
        val commonMarkExtHeadingAnchor = "org.commonmark"   % "commonmark-ext-heading-anchor"    % Version.commonMark
        val commonMarkExtImageAttrs    = "org.commonmark"   % "commonmark-ext-image-attributes"  % Version.commonMark
        val commonMarkExtStrikethrough = "org.commonmark"   % "commonmark-ext-gfm-strikethrough" % Version.commonMark
        val commonMarkExtTables        = "org.commonmark"   % "commonmark-ext-gfm-tables"        % Version.commonMark
        val commonMarkExtTaskListItems = "org.commonmark"   % "commonmark-ext-task-list-items"   % Version.commonMark
        val doobieCore                 = "org.tpolecat"    %% "doobie-core"                      % Version.doobie
        val doobieHikari               = "org.tpolecat"    %% "doobie-hikari"                    % Version.doobie
        val doobiePostgres             = "org.tpolecat"    %% "doobie-postgres"                  % Version.doobie
        val doobieScalaTest            = "org.tpolecat"    %% "doobie-scalatest"                 % Version.doobie
        val ed25519Java                = "net.i2p.crypto"   % "eddsa"                            % "0.3.0"
        val flywayCore                 = "org.flywaydb"     % "flyway-core"                      % Version.flyway
        val flywayPostgreSQL           = "org.flywaydb"     % "flyway-database-postgresql"       % Version.flyway
        val fs2Core                    = "co.fs2"          %% "fs2-core"                         % Version.fs2
        val fs2IO                      = "co.fs2"          %% "fs2-io"                           % Version.fs2
        val http4sCirce                = "org.http4s"      %% "http4s-circe"                     % Version.http4s
        val http4sCore                 = "org.http4s"      %% "http4s-core"                      % Version.http4s
        val http4sDsl                  = "org.http4s"      %% "http4s-dsl"                       % Version.http4s
        val http4sEmberServer          = "org.http4s"      %% "http4s-ember-server"              % Version.http4s
        val http4sEmberClient          = "org.http4s"      %% "http4s-ember-client"              % Version.http4s
        val ip4sCore                   = "com.comcast"     %% "ip4s-core"                        % Version.ip4s
        val jansi                = "com.github.Osiris-Team"       % "jansi"                  % Version.jansi
        val jclOverSlf4j         = "org.slf4j"                    % "jcl-over-slf4j"         % Version.jclOverSlf4j
        val log4catsSlf4j        = "org.typelevel"               %% "log4cats-slf4j"         % Version.log4cats
        val logback              = "ch.qos.logback"               % "logback-classic"        % Version.logback
        val munit                = "org.scalameta"               %% "munit"                  % Version.munit
        val munitCatsEffect      = "org.typelevel"               %% "munit-cats-effect"      % Version.munitCatsEffect
        val munitScalaCheck      = "org.scalameta"               %% "munit-scalacheck"       % Version.munit
        val osLib                = "com.lihaoyi"                 %% "os-lib"                 % Version.osLib
        val postgresql           = "org.postgresql"               % "postgresql"             % Version.postgresql
        val pureConfig           = "com.github.pureconfig"       %% "pureconfig-core"        % Version.pureConfig
        val scalaCheck           = "org.scalacheck"              %% "scalacheck"             % Version.scalaCheck
        val simpleJavaMail       = "org.simplejavamail"           % "simple-java-mail"       % Version.simpleJavaMail
        val springSecurityCrypto = "org.springframework.security" % "spring-security-crypto" % Version.springSecurity
    }

// *****************************************************************************
// Settings
// *****************************************************************************

/** Calculate the copyright year string e.g. xxxx or xxxx - yyyy depending on the given start year and the current year.
  *
  * @param start
  *   The start year of the project.
  * @param current
  *   The current year which must be equal or larger than the start year.
  * @return
  *   A string containing either only the start year or "start year - current year"
  */
def calculateCopyrightYears(start: Int, current: Int): String =
    if (start < current)
        s"$start - $current"
    else
        start.toString

lazy val commonSettings =
    codeSettings ++
        scalafmtSettings

lazy val codeSettings =
    Seq(
        startYear := Option(projectStartYear),
        licenses += ("AGPL-3.0-or-later", new URL("https://www.gnu.org/licenses/agpl-3.0.txt"))
    )

lazy val scalafmtSettings =
    Seq(
        scalafmtOnCompile := false
    )
