/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.darcs

import java.nio.file.*

import cats.*
import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import org.slf4j.LoggerFactory

import scala.util.matching.Regex

opaque type DarcsHash = String
object DarcsHash {
    val Format: Regex = "^[a-f0-9]{40}$".r

    /** Create an instance of DarcsHash from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a DarcsHash.
      * @return
      *   The appropriate instance of DarcsHash.
      */
    def apply(source: String): DarcsHash = source

    /** Try to create an instance of DarcsHash from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a DarcsHash.
      * @return
      *   An option to the successfully converted DarcsHash.
      */
    def from(source: String): Option[DarcsHash] = Option(source).filter(s => Format.matches(s))

}

/** Wrapper for output generated by an external darcs command.
  *
  * @param exitValue
  *   The exit value return by the command (usually anything other than 0 indicates an error).
  * @param stdout
  *   The standard output produced by the command.
  * @param stderr
  *   The standard error output produced by the command.
  */
final case class DarcsCommandOutput(exitValue: Int, stdout: Chain[String], stderr: Chain[String])

/** Provide an interface to executing darcs commands via the external darcs binary provided.
  *
  * @param darcsBinary
  *   The path to the darcs binary which is invoked for commands.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class DarcsCommands[F[_]: Sync](val darcsBinary: Path) {
    private val log = LoggerFactory.getLogger(getClass)

    /** Apply a patch to a darcs repository.
      *
      * @param basePath
      *   The base path under which the repository is located.
      * @param repositoryName
      *   The name of the repository.
      * @param patchFile
      *   The path to the file containing the patch. This may also be a complete email. In the latter case the email
      *   will be parsed for a MIME attachment with a darcs patch.
      * @param options
      *   Additional options for the initialize command. Please note that you should not specify options which need user
      *   interaction and also you might want to set the `--dont-allow-conflicts` options.
      * @return
      *   The output of the darcs command.
      */
    def applyPatch(basePath: Path)(repositoryName: String)(patchFile: Path)(
        options: Chain[String]
    ): F[DarcsCommandOutput] = {
        log.trace(s"Execute $darcsBinary apply $patchFile for $repositoryName with $options")
        val repositoryDirectory = basePath.resolve(repositoryName)
        val darcsOptions        = List("apply") ::: options.toList ::: List(patchFile.toAbsolutePath().toString)
        val externalCommand     = os.proc(darcsBinary.toString, darcsOptions)
        for {
            process <- Sync[F].delay(externalCommand.call(cwd = os.Path(repositoryDirectory), check = false))
        } yield DarcsCommandOutput(process.exitCode, Chain(process.out.text()), Chain(process.err.text()))
    }

    /** Clone a darcs repository from one directory into another one. This function is intended to be used for server
      * side forking or repositories.
      *
      * @param sourceRepositoryPath
      *   The path of the source repository.
      * @param targetRespositoryPath
      *   The path of the target repository.
      * @param options
      *   Additional options for the clone command.
      * @return
      *   The output of the darcs command.
      */
    def clone(sourceRepositoryPath: Path, targetRespositoryPath: Path)(
        options: Chain[String]
    ): F[DarcsCommandOutput] = {
        val source = sourceRepositoryPath.toAbsolutePath
        val target = targetRespositoryPath.toAbsolutePath
        log.trace(s"Execute $darcsBinary clone $sourceRepositoryPath to $targetRespositoryPath with $options")
        val darcsOptions    = List("clone") ::: options.toList ::: List(source.toString, target.toString)
        val externalCommand = os.proc(darcsBinary.toString, darcsOptions)
        for {
            process <- Sync[F].delay(externalCommand.call(check = false))
        } yield DarcsCommandOutput(process.exitCode, Chain(process.out.text()), Chain(process.err.text()))
    }

    /** Return the diff of the specified patch from the given darcs repository.
      *
      * @param basePath
      *   The base path under which the repository is located.
      * @param repositoryName
      *   The name of the repository.
      * @param hash
      *   The hash of the patch whose diff shall be generated.
      * @param options
      *   Additional options for the initialize command.
      * @return
      *   The output of the darcs command.
      */
    def diff(basePath: Path)(repositoryName: String)(hash: DarcsHash)(options: Chain[String]): F[DarcsCommandOutput] = {
        log.trace(s"Execute $darcsBinary diff --hash $hash for $repositoryName with $options")
        val repositoryDirectory = Paths.get(basePath.toString, repositoryName)
        val darcsOptions        = List("diff", s"--hash=${hash.toString}") ::: options.toList
        val externalCommand     = os.proc(darcsBinary.toString, darcsOptions)
        for {
            process <- Sync[F].delay(externalCommand.call(cwd = os.Path(repositoryDirectory), check = false))
        } yield DarcsCommandOutput(process.exitCode, Chain(process.out.text()), Chain(process.err.text()))
    }

    /** Create a distribution archive from the given darcs repository.
      *
      * @param basePath
      *   The base path under which the repository is located.
      * @param repositoryName
      *   The name of the repository.
      * @param options
      *   Additional options for the initialize command.
      * @return
      *   The output of the darcs command.
      */
    def dist(basePath: Path)(repositoryName: String)(options: Chain[String]): F[DarcsCommandOutput] = {
        log.trace(s"Execute $darcsBinary dist for $repositoryName")
        val repositoryDirectory = Paths.get(basePath.toString, repositoryName)
        val darcsOptions        = List("dist") ::: options.toList
        val externalCommand     = os.proc(darcsBinary.toString, darcsOptions)
        for {
            process <- Sync[F].delay(externalCommand.call(cwd = os.Path(repositoryDirectory), check = false))
        } yield DarcsCommandOutput(process.exitCode, Chain(process.out.text()), Chain(process.err.text()))
    }

    /** Initialize a darcs repository under the given base path with the provided name. This is done by running the
      * external darcs binary with the appropriate parameters.
      *
      * @param basePath
      *   The base path under which the repository is located.
      * @param repositoryName
      *   The name of the repository.
      * @param options
      *   Additional options for the initialize command.
      * @return
      *   The output of the darcs command.
      */
    def initialize(basePath: Path)(repositoryName: String)(options: Chain[String]): F[DarcsCommandOutput] = {
        log.trace(s"Execute $darcsBinary initialize $basePath/$repositoryName with $options")
        val directory       = Paths.get(basePath.toString, repositoryName)
        val darcsOptions    = List("initialize") ::: options.toList ::: List(directory.toString)
        val externalCommand = os.proc(darcsBinary.toString, darcsOptions)
        for {
            process <- Sync[F].delay(externalCommand.call(check = false))
        } yield DarcsCommandOutput(process.exitCode, Chain(process.out.text()), Chain(process.err.text()))
    }

    /** Run the darcs log command on the given repository and return the output.
      *
      * @param basePath
      *   The base path under which the repository is located.
      * @param repositoryName
      *   The name of the repository.
      * @param options
      *   Additional options for the log command.
      * @return
      *   The output of the darcs command.
      */
    def log(basePath: Path)(repositoryName: String)(options: Chain[String]): F[DarcsCommandOutput] = {
        log.trace(s"Execute $darcsBinary log in $basePath/$repositoryName with $options")
        val directory       = Paths.get(basePath.toString, repositoryName)
        val darcsOptions    = List("log") ::: options.toList
        val externalCommand = os.proc(darcsBinary.toString, darcsOptions)
        for {
            process <- Sync[F].delay(externalCommand.call(cwd = os.Path(directory), check = false))
        } yield DarcsCommandOutput(process.exitCode, Chain(process.out.text()), Chain(process.err.text()))
    }

    /** Run the darcs show dependencies command on the given repository and return the output.
      *
      * @param basePath
      *   The base path under which the repository is located.
      * @param repositoryName
      *   The name of the repository.
      * @param options
      *   Additional options for the show command.
      * @return
      *   The output of the darcs command.
      */
    def showDependencies(
        basePath: Path
    )(repositoryName: String)(options: Chain[String]): F[DarcsCommandOutput] = {
        log.trace(s"Execute $darcsBinary show dependencies in $basePath/$repositoryName with $options")
        val directory       = Paths.get(basePath.toString, repositoryName)
        val darcsOptions    = List("show", "dependencies") ::: options.toList
        val externalCommand = os.proc(darcsBinary.toString, darcsOptions)
        for {
            process <- Sync[F].delay(externalCommand.call(cwd = os.Path(directory), check = false))
        } yield DarcsCommandOutput(process.exitCode, Chain(process.out.text()), Chain(process.err.text()))
    }

    /** Run the darcs whatsnew command on the given repository and return the output.
      *
      * @param basePath
      *   The base path under which the repository is located.
      * @param repositoryName
      *   The name of the repository.
      * @param options
      *   Additional options for the whatsnew command.
      * @return
      *   The output of the darcs command.
      */
    def whatsnew(basePath: Path)(repositoryName: String)(options: Chain[String]): F[DarcsCommandOutput] = {
        log.trace(s"Execute $darcsBinary whatsnew in $basePath/$repositoryName with $options")
        val directory       = Paths.get(basePath.toString, repositoryName)
        val darcsOptions    = List("whatsnew") ::: options.toList
        val externalCommand = os.proc(darcsBinary.toString, darcsOptions)
        for {
            process <- Sync[F].delay(externalCommand.call(cwd = os.Path(directory), check = false))
        } yield DarcsCommandOutput(process.exitCode, Chain(process.out.text()), Chain(process.err.text()))
    }

}
