/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.darcs

import java.nio.file.*

import cats.data.*
import cats.effect.*
import cats.syntax.all.*

import munit.*

final class DarcsCommandsTest extends CatsEffectSuite with TestHelpers {
    val darcsBinary = Paths.get("darcs")

    val tempWorkingDirectory = ResourceFunFixture(
        Resource.make(IO(Files.createTempDirectory("darcs-cmd-test-").toAbsolutePath()))(path =>
            IO(deleteDirectory(path)) *> IO.unit
        )
    )

    tempWorkingDirectory.test("darcs apply must apply a proper patch") { workingDirectory =>
        val cmd               = new DarcsCommands[IO](darcsBinary)
        val repo              = "test-apply"
        val expectedDirectory = workingDirectory.resolve(repo)
        val testPatch         = Paths.get(getClass().getClassLoader().getResource("test-patch.dpatch").toURI())
        val createRepository  = cmd.initialize(workingDirectory)(repo)(Chain.empty)
        val test = for {
            init  <- createRepository
            patch <- cmd.applyPatch(workingDirectory)(repo)(testPatch)(Chain.empty)
        } yield (init, patch)
        test.map { output =>
            val (init, patch) = output
            assert(init.exitValue === 0, "darcs init did not finish with exit code 0!")
            assert(patch.exitValue === 0, "darcs apply did not finish with exit code 0!")
            assert(
                Files.exists(expectedDirectory.resolve("README.md")),
                "Expected file README.md missing after applying the patch!"
            )
        }
    }

    tempWorkingDirectory.test("darcs clone must create a proper copy") { workingDirectory =>
        val cmd    = new DarcsCommands[IO](darcsBinary)
        val source = workingDirectory.resolve("source")
        val target = workingDirectory.resolve("target")
        val test =
            for {
                init  <- cmd.initialize(workingDirectory)("source")(Chain.empty)
                clone <- cmd.clone(source, target)(Chain.empty)
            } yield (init, clone)
        test.map { output =>
            val (init, clone) = output
            assert(init.exitValue === 0, "darcs init did not finish with exit code 0!")
            assert(Files.exists(source.toAbsolutePath), "Source repository directory does not exist!")
            assert(clone.exitValue === 0, "darcs clone did not finish with exit code 0!")
            assert(Files.exists(target.toAbsolutePath), "Target repository directory does not exist!")
        }
    }

    tempWorkingDirectory.test("darcs clone must fail if the source does not exist") { workingDirectory =>
        val cmd    = new DarcsCommands[IO](darcsBinary)
        val source = workingDirectory.resolve("source-should-not-exist")
        val target = workingDirectory.resolve("target-should-not-be-created")
        val test =
            for {
                clone <- cmd.clone(source, target)(Chain.empty)
            } yield clone
        test.map { clone =>
            assert(!Files.exists(source.toAbsolutePath), "Source repository directory must not exist!")
            assert(clone.exitValue =!= 0, "darcs clone must not finish with exit code 0!")
            assert(!Files.exists(target.toAbsolutePath), "Target repository directory must not exist!")
        }
    }

    tempWorkingDirectory.test("darcs initialize must create a new repository") { workingDirectory =>
        val cmd               = new DarcsCommands[IO](darcsBinary)
        val repo              = "test-repository"
        val expectedDirectory = workingDirectory.resolve(repo)
        val test              = cmd.initialize(workingDirectory)(repo)(Chain.empty)
        test.map { output =>
            assert(output.exitValue === 0, "Exit code of darcs command is expected to be 0!")
            assert(Files.exists(expectedDirectory), "Expected directory does not exist!")
            assert(Files.isDirectory(expectedDirectory), "Expected directory is not a directory!")
            val darcsDirectory = Paths.get(expectedDirectory.toString, "_darcs")
            assert(Files.exists(darcsDirectory), "The _darcs directory does not exist in the repository!")
        }
    }

    tempWorkingDirectory.test("darcs initialize must create a new repository in an existing directory") {
        workingDirectory =>
            val cmd               = new DarcsCommands[IO](darcsBinary)
            val repo              = "test-repository-in-directory"
            val expectedDirectory = workingDirectory.resolve(repo)
            val _                 = Files.createDirectories(expectedDirectory)
            val test              = cmd.initialize(workingDirectory)(repo)(Chain.empty)
            test.map { output =>
                assert(output.exitValue === 0, "Exit code of darcs command is expected to be 0!")
                val darcsDirectory = Paths.get(expectedDirectory.toString, "_darcs")
                assert(Files.exists(darcsDirectory), "The _darcs directory does not exist in the repository!")
            }
    }

    tempWorkingDirectory.test("darcs initialize must fail if the repository already exists") { workingDirectory =>
        val cmd               = new DarcsCommands[IO](darcsBinary)
        val repo              = "test-repository-existing"
        val expectedDirectory = workingDirectory.resolve(repo)
        val _                 = Files.createDirectories(expectedDirectory)
        val test = for {
            _      <- cmd.initialize(workingDirectory)(repo)(Chain.empty)
            output <- cmd.initialize(workingDirectory)(repo)(Chain.empty)
        } yield output
        test.map { output =>
            assert(output.exitValue =!= 0, "The initialize command is expected to fail here!")
            assert(output.stderr.nonEmpty)
        }
    }
}
