/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.email

import cats.data.NonEmptyList

import org.scalacheck.*

/** A container for ScalaCheck generators for our data types.
  */
object Generators {
    private val validEmailAddressPrefixChars =
        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.!#$%&’'*+/=?^_`{|}~-".toList
    private val validDomainNameChars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-".toList

    val genValidFromAddress: Gen[FromAddress] = for {
        prefix <- Gen.nonEmptyListOf(Gen.oneOf(validEmailAddressPrefixChars)).map(_.take(64).mkString)
        domain <- Gen.nonEmptyListOf(Gen.oneOf(validDomainNameChars)).map(_.take(32).mkString)
        topLevelDomain <- Gen
            .nonEmptyListOf(Gen.oneOf(validDomainNameChars))
            .suchThat(_.length >= 2)
            .map(_.take(24).mkString)
        suffix = s"$domain.$topLevelDomain"
    } yield FromAddress(s"$prefix@$suffix")

    given Arbitrary[FromAddress] = Arbitrary(genValidFromAddress)

    val genValidSubjectLine: Gen[SubjectLine] =
        Gen.nonEmptyListOf(Gen.alphaNumChar).map(_.take(78).mkString).map(SubjectLine.apply)

    given Arbitrary[SubjectLine] = Arbitrary(genValidSubjectLine)

    val genValidTextBody: Gen[TextBody] = Gen.nonEmptyListOf(Gen.alphaNumChar).map(_.mkString).map(TextBody.apply)

    given Arbitrary[TextBody] = Arbitrary(genValidTextBody)

    val genValidToAddress: Gen[ToAddress] = for {
        prefix <- Gen.nonEmptyListOf(Gen.oneOf(validEmailAddressPrefixChars)).map(_.take(64).mkString)
        domain <- Gen.nonEmptyListOf(Gen.oneOf(validDomainNameChars)).map(_.take(32).mkString)
        topLevelDomain <- Gen
            .nonEmptyListOf(Gen.oneOf(validDomainNameChars))
            .suchThat(_.length >= 2)
            .map(_.take(24).mkString)
        suffix = s"$domain.$topLevelDomain"
    } yield ToAddress(s"$prefix@$suffix")

    given Arbitrary[ToAddress] = Arbitrary(genValidToAddress)

    val genEmailMessage: Gen[EmailMessage] = for {
        from     <- genValidFromAddress
        to       <- genValidToAddress
        too      <- Gen.listOf(genValidToAddress)
        cc       <- Gen.listOf(genValidToAddress)
        bcc      <- Gen.listOf(genValidToAddress)
        subject  <- genValidSubjectLine
        textBody <- genValidTextBody
    } yield EmailMessage(
        from,
        NonEmptyList.of(to, too: _*),
        cc,
        bcc,
        subject,
        textBody
    )

    given Arbitrary[EmailMessage] = Arbitrary(genEmailMessage)
}
