/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.html

import cats.syntax.all.*
import org.http4s.Uri

object LinkTools {

    extension (linkConfig: ExternalUrlConfiguration) {

        /** Take the given URI path and create a full URI using the specified configuration with a possible path prefix
          * and append the given path to it.
          *
          * @param path
          *   An URI containing a path with possible URL fragment and query parameters which will be used to construct
          *   the full URI.
          * @return
          *   A full URI created from the values of the ExternalUrlConfiguration (scheme, host, port, possible path
          *   prefix) and the path data from the given URI.
          */
        def createFullUri(path: Uri): Uri = {
            val completePath = linkConfig.path match {
                case None             => path.path
                case Some(pathPrefix) => pathPrefix.path |+| path.path
            }
            val baseUri = Uri(
                scheme = Option(linkConfig.scheme),
                authority = Option(
                    Uri.Authority(
                        userInfo = None,
                        host = Uri.Host.fromIp4sHost(linkConfig.host),
                        port = linkConfig.port.map(_.value)
                    )
                ),
                path = completePath
            ).withQueryParams(path.params)
            path.fragment.fold(baseUri)(fragment => baseUri.withFragment(fragment))
        }
    }

}
