CREATE SCHEMA IF NOT EXISTS hub;
COMMENT ON SCHEMA hub IS 'Central data storage containing accounts, projects and so on.';

CREATE TABLE hub.accounts
(
  uid              UUID                     NOT NULL,
  name             CHARACTER VARYING(32)    NOT NULL,
  email            CHARACTER VARYING(128)   NOT NULL,
  password         TEXT,
  failed_attempts  INTEGER                  DEFAULT 0,
  locked_at        TIMESTAMP WITH TIME ZONE DEFAULT NULL,
  unlock_token     TEXT                     DEFAULT NULL,
  reset_expiry     TIMESTAMP WITH TIME ZONE DEFAULT NULL,
  reset_token      TEXT                     DEFAULT NULL,
  created_at       TIMESTAMP WITH TIME ZONE NOT NULL,
  updated_at       TIMESTAMP WITH TIME ZONE NOT NULL,
  validated_email  BOOLEAN                  DEFAULT FALSE,
  validation_token TEXT                     DEFAULT NULL,
  CONSTRAINT accounts_pk           PRIMARY KEY (uid),
  CONSTRAINT accounts_unique_name  UNIQUE (name),
  CONSTRAINT accounts_unique_email UNIQUE (email)
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE hub.accounts IS 'All user accounts for the system live within this table.';
COMMENT ON COLUMN hub.accounts.uid IS 'A globally unique ID for the related user account.';
COMMENT ON COLUMN hub.accounts.name IS 'A username between 2 and 32 characters which must be globally unique, contain only lowercase alphanumeric characters and start with a character.';
COMMENT ON COLUMN hub.accounts.email IS 'A globally unique email address associated with the account.';
COMMENT ON COLUMN hub.accounts.password IS 'The hashed password for the account.';
COMMENT ON COLUMN hub.accounts.failed_attempts IS 'The number of failed login attempts since the last sucessful login.';
COMMENT ON COLUMN hub.accounts.locked_at IS 'A timestamp when the account was locked. If this value is not NULL then the account should be considered locked';
COMMENT ON COLUMN hub.accounts.unlock_token IS 'An unlock token which can be used to unlock the account.';
COMMENT ON COLUMN hub.accounts.reset_expiry IS 'The timestamp when the reset token will expire.';
COMMENT ON COLUMN hub.accounts.reset_token IS 'A token which can be used for a password reset.';
COMMENT ON COLUMN hub.accounts.created_at IS 'The timestamp of when the account was created.';
COMMENT ON COLUMN hub.accounts.updated_at IS 'A timestamp when the account was last changed.';
COMMENT ON COLUMN hub.accounts.validated_email IS 'This flag indicates if the email address of the user has been validated via a validation email.';
COMMENT ON COLUMN hub.accounts.validation_token IS 'A token used to validate the email address of the user.';

CREATE TABLE hub.sessions
(
  id         VARCHAR(32)              NOT NULL,
  uid        UUID                     NOT NULL,
  created_at TIMESTAMP WITH TIME ZONE NOT NULL,
  updated_at TIMESTAMP WITH TIME ZONE NOT NULL,
  CONSTRAINT sessions_pk     PRIMARY KEY (id),
  CONSTRAINT sessions_fk_uid FOREIGN KEY (uid)
    REFERENCES hub.accounts (uid) ON UPDATE CASCADE ON DELETE CASCADE
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE hub.sessions IS 'Keeps the sessions of users.';
COMMENT ON COLUMN hub.sessions.id IS 'A globally unique session ID.';
COMMENT ON COLUMN hub.sessions.uid IS 'The unique ID of the user account to whom the session belongs.';
COMMENT ON COLUMN hub.sessions.created_at IS 'The timestamp of when the session was created.';
COMMENT ON COLUMN hub.sessions.updated_at IS 'The session ID should be re-generated in regular intervals resulting in a copy of the old session entry with a new ID and the corresponding timestamp in this column.';

CREATE TABLE hub.ssh_keys
(
  id           UUID                     NOT NULL, 
  uid          UUID                     NOT NULL,
  key_type     CHARACTER VARYING(32)    NOT NULL,
  key          TEXT                     NOT NULL,
  fingerprint  CHARACTER VARYING(256)   NOT NULL,
  comment      CHARACTER VARYING(256)   DEFAULT NULL,
  created_at   TIMESTAMP WITH TIME ZONE NOT NULL,
  last_used_at TIMESTAMP WITH TIME ZONE DEFAULT NULL,
  CONSTRAINT ssh_keys_pk        PRIMARY KEY (id),
  CONSTRAINT ssh_keys_unique_fp UNIQUE (fingerprint),
  CONSTRAINT ssh_keys_fk_uid    FOREIGN KEY (uid)
    REFERENCES hub.accounts (uid) ON UPDATE CASCADE ON DELETE CASCADE
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE hub.ssh_keys IS 'SSH keys uploaded by users live within this table. Updates are not intended, so keys have to be deleted and re-uploaded upon changes.';
COMMENT ON COLUMN hub.ssh_keys.id IS 'The globally unique ID of the ssh key.';
COMMENT ON COLUMN hub.ssh_keys.uid IS 'The unique ID of the user account to whom the ssh key belongs.';
COMMENT ON COLUMN hub.ssh_keys.key_type IS 'The type of the key e.g. ssh-rsa or ssh-ed25519 and so on.';
COMMENT ON COLUMN hub.ssh_keys.key IS 'A base 64 string containing the public ssh key.';
COMMENT ON COLUMN hub.ssh_keys.fingerprint IS 'The fingerprint of the ssh key. It must be unique because a key can only be used by one account.';
COMMENT ON COLUMN hub.ssh_keys.comment IS 'An optional comment for the ssh key limited to 256 characters.';
COMMENT ON COLUMN hub.ssh_keys.created_at IS 'The timestamp of when the ssh key was created.';
COMMENT ON COLUMN hub.ssh_keys.last_used_at IS 'The timestamp of when the ssh key was last used by the user.';
