/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.util.UUID

import de.smederee.i18n.LanguageCode
import de.smederee.security.*
import de.smederee.ssh.PublicSshKey
import fs2.Stream

/** The base class for database operations related to account management for users.
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class AccountManagementRepository[F[_]] {

    /** Add the given ssh key to the database. The database MUST ensure that a key is unique across the system! Usually
      * by having a unique constraint on the fingerprint of the key.
      *
      * @param key
      *   The public ssh key to be saved to the database.
      * @return
      *   The number of affected database rows.
      */
    def addSshKey(key: PublicSshKey): F[Int]

    /** Delete the account with the given user id from the database.
      *
      * The internal database logic (foreign keys, cascading) SHALL ensure that everything related to the user is
      * deleted from the database.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   The number of affected database rows.
      */
    def deleteAccount(uid: UserId): F[Int]

    /** Delete the ssh key with the given id and owner from the database.
      *
      * @param keyId
      *   The unique id of the public ssh key.
      * @param ownerId
      *   The unique user id of the owner of the ssh key.
      * @return
      *   The number of affected database rows.
      */
    def deleteSshKey(keyId: UUID, ownerId: UserId): F[Int]

    /** Find the account with the given validation token.
      *
      * @param token
      *   A validation token.
      * @return
      *   An option to the account if it exists.
      */
    def findByValidationToken(token: ValidationToken): F[Option[Account]]

    /** Retrieve the password hash from the database.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   An option to the password hash.
      */
    def findPasswordHash(uid: UserId): F[Option[PasswordHash]]

    /** Return all public ssh keys for the given user.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   A stream of public ssh keys that may be empty.
      */
    def listSshKeys(uid: UserId): Stream[F, PublicSshKey]

    /** Mark the account with the given user id as validated. This includes the following operations in the database:
      *
      * {{{
      * 1. set the ´validated_email` column to `true`
      * 2. clear the `validation_token` column
      * }}}
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   The number of affected database rows.
      */
    def markAsValidated(uid: UserId): F[Int]

    /** Set the language for the user account.
      *
      * @param uid
      *   The unique id of the user account.
      * @param language
      *   An option to the preferred language that shall be set for the account.
      * @return
      *   The number of affected database rows.
      */
    def setLanguage(uid: UserId, language: Option[LanguageCode]): F[Int]

    /** Set the validation token for the account with the given user id.
      *
      * @param uid
      *   The unique id of the user account.
      * @param token
      *   The validation token to be stored in the database.
      * @return
      *   The number of affected database rows.
      */
    def setValidationToken(uid: UserId, token: ValidationToken): F[Int]

}
