/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.io.IOException
import java.nio.file.FileVisitResult
import java.nio.file.FileVisitor
import java.nio.file.Files
import java.time.OffsetDateTime
import java.time.ZoneOffset
import java.util.UUID

import cats.*
import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.email.*
import de.smederee.html.LinkTools.*
import de.smederee.hub.RequestHelpers.instances.given
import de.smederee.hub.config.*
import de.smederee.hub.forms.types.*
import de.smederee.i18n.LanguageCode
import de.smederee.security.*
import de.smederee.ssh.*
import de.smederee.tickets.TicketServiceApi
import de.smederee.tickets.TicketsUser
import org.http4s.*
import org.http4s.dsl.*
import org.http4s.headers.Location
import org.http4s.implicits.*
import org.http4s.twirl.TwirlInstances.*
import org.slf4j.LoggerFactory

import scala.util.Try

/** Routes for user account self management which should provide every functionality needed by users to manage their
  * account.
  *
  * @param accountManagementRepo
  *   The repository providing all needed database functionality for account management.
  * @param configuration
  *   The hub service configuration.
  * @param emailMiddleware
  *   Middleware layer needed to send emails.
  * @param organisationsRepo
  *   The repository providing functionality to handle organisations.
  * @param signAndValidate
  *   A class providing functions to handle session token signing and validation.
  * @param ticketServiceApi
  *   The ticket service API which is needed to synchronise changes for the account.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class AccountManagementRoutes[F[_]: Async](
    accountManagementRepo: AccountManagementRepository[F],
    configuration: ServiceConfig,
    emailMiddleware: EmailMiddleware[F],
    organisationsRepo: OrganisationRepository[F],
    signAndValidate: SignAndValidate,
    ticketServiceApi: TicketServiceApi[F]
) extends Http4sDsl[F] {
    private val log = LoggerFactory.getLogger(getClass)

    /** Delete the given directory recursively. It is checked if the given directory is a direct sub directory of the
      * `repositoriesDirectory` and only if this is the case is the directory removed.
      *
      * @param userDirectory
      *   The path on the filesystem to the directory that shall be deleted.
      * @return
      *   `true` if the directory was deleted.
      */
    protected def deleteUserDirectory(userDirectory: java.nio.file.Path): F[Boolean] =
        for {
            _            <- Sync[F].delay(log.debug(s"Request to delete user dir: $userDirectory"))
            reposDirPath <- Sync[F].delay(configuration.darcs.repositoriesDirectory.toPath)
            isSubDir     <- Sync[F].delay(reposDirPath.equals(userDirectory.getParent()))
            deleted <-
                Sync[F].delay {
                    if (isSubDir) {
                        Files.walkFileTree(
                            userDirectory,
                            new FileVisitor[java.nio.file.Path] {
                                override def visitFileFailed(
                                    file: java.nio.file.Path,
                                    exc: IOException
                                ): FileVisitResult = FileVisitResult.CONTINUE

                                override def visitFile(
                                    file: java.nio.file.Path,
                                    attrs: java.nio.file.attribute.BasicFileAttributes
                                ): FileVisitResult = {
                                    Files.delete(file)
                                    FileVisitResult.CONTINUE
                                }

                                override def preVisitDirectory(
                                    dir: java.nio.file.Path,
                                    attrs: java.nio.file.attribute.BasicFileAttributes
                                ): FileVisitResult = FileVisitResult.CONTINUE

                                override def postVisitDirectory(
                                    dir: java.nio.file.Path,
                                    exc: IOException
                                ): FileVisitResult = {
                                    Files.delete(dir)
                                    FileVisitResult.CONTINUE
                                }
                            }
                        )
                        Files.deleteIfExists(userDirectory)
                    } else {
                        log.warn(
                            s"Refused requested removal of directory $userDirectory which is not a direct sub directory of the configured repositories directory!"
                        )
                        false
                    }
                }
        } yield deleted

    private val addSshKey: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / "user" / "settings" / "ssh" / "add" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    formData <- Sync[F].delay {
                        urlForm.values.map { t =>
                            val (key, values) = t
                            (
                                key,
                                values.headOption.getOrElse("")
                            ) // Pick the first value (a field might get submitted multiple times)!
                        }
                    }
                    form          <- Sync[F].delay(AddPublicSshKeyForm.validate(formData))
                    actionBaseUri <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings"))
                    addAction     <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings/ssh/add"))
                    deleteAction  <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings/ssh/delete"))
                    keys          <- accountManagementRepo.listSshKeys(user.uid).compile.toList
                    resp <- form match {
                        case Validated.Invalid(errors) =>
                            BadRequest(
                                views.html.account
                                    .sshSettings(lang = language)(
                                        csrf,
                                        Option(s"Smederee/~${user.name} - SSH Settings"),
                                        user
                                    )(
                                        actionBaseUri,
                                        addAction,
                                        deleteAction,
                                        keys
                                    )(formData, FormErrors.fromNec(errors))
                            )
                        case Validated.Valid(validSshKeyForm) =>
                            for {
                                id        <- Sync[F].delay(UUID.randomUUID())
                                createdAt <- Sync[F].delay(OffsetDateTime.now(ZoneOffset.UTC))
                                convertedKey <- Sync[F].delay(
                                    PublicSshKey.from(id)(user.uid)(createdAt)(validSshKeyForm.keyString)
                                )
                                key <- Sync[F].delay(
                                    convertedKey.map(key =>
                                        validSshKeyForm.name.fold(key)(name => key.copy(comment = Option(name)))
                                    )
                                ) // Override comment with the one from the form if given.
                                written <- key.traverse(key =>
                                    accountManagementRepo.addSshKey(key).recoverWith(_ => Sync[F].pure(0))
                                )
                                resp <- written match {
                                    case None =>
                                        // There has been no write at all, implying that the conversion failed.
                                        BadRequest(
                                            views.html.account
                                                .sshSettings(lang = language)(
                                                    csrf,
                                                    Option(s"Smederee/~${user.name} - SSH Settings"),
                                                    user
                                                )(
                                                    actionBaseUri,
                                                    addAction,
                                                    deleteAction,
                                                    keys
                                                )(
                                                    formData,
                                                    Map(
                                                        AddPublicSshKeyForm.fieldGlobal -> List(
                                                            FormFieldError("The key could not be properly converted!")
                                                        )
                                                    )
                                                )
                                        )
                                    case Some(1) =>
                                        // One row was written to the database implying that everything went well.
                                        SeeOther(Location(actionBaseUri.addSegment("ssh")))
                                    case Some(_) =>
                                        // Any other result implies that there has been an error.
                                        accountManagementRepo.listSshKeys(user.uid).compile.toList.flatMap { keys =>
                                            BadRequest(
                                                views.html.account
                                                    .sshSettings(lang = language)(
                                                        csrf,
                                                        Option(s"Smederee/~${user.name} - SSH Settings"),
                                                        user
                                                    )(
                                                        actionBaseUri,
                                                        addAction,
                                                        deleteAction,
                                                        keys
                                                    )(
                                                        formData,
                                                        Map(
                                                            AddPublicSshKeyForm.fieldGlobal -> List(
                                                                FormFieldError(
                                                                    "An error occured while saving the key to the database!"
                                                                )
                                                            )
                                                        )
                                                    )
                                            )
                                        }
                                }
                            } yield resp
                    }
                } yield resp
            }
    }

    private val deleteAccount: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / "user" / "settings" / "delete" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    formData <- Sync[F].delay {
                        urlForm.values.map { t =>
                            val (key, values) = t
                            (
                                key,
                                values.headOption.getOrElse("")
                            ) // Pick the first value (a field might get submitted multiple times)!
                        }
                    }
                    passwordField  <- Sync[F].delay(formData.get("password").flatMap(Password.from))
                    userIsSure     <- Sync[F].delay(formData.get("i-am-sure").exists(_ === "yes"))
                    rootUri        <- Sync[F].delay(configuration.external.createFullUri(uri"/"))
                    actionBaseUri  <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings"))
                    deleteAction   <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings/delete"))
                    languageAction <- Sync[F].delay(configuration.external.createFullUri(uri"/user/settings/language"))
                    validateAction <- Sync[F].delay(
                        configuration.external.createFullUri(uri"user/settings/email/validate")
                    )
                    passwordHash <- accountManagementRepo.findPasswordHash(user.uid)
                    passwordCorrect <- Sync[F].delay(
                        (passwordField, passwordHash)
                            .mapN { case (enteredPassword, hashFromDatabase) =>
                                enteredPassword.matches(hashFromDatabase)
                            }
                            .getOrElse(false)
                    )
                    response <-
                        if (passwordCorrect && userIsSure) {
                            // First we delete the user directory and all sub directories.
                            // Afterwards we delete the user account from the database and redirect the user
                            // to the start page, removing the authentication cookie in the process.
                            for {
                                _ <- Sync[F].delay(
                                    log.info(s"Going to delete account ${user.name} as requested by the user.")
                                )
                                userDir <- Sync[F].delay(
                                    java.nio.file.Paths
                                        .get(
                                            configuration.darcs.repositoriesDirectory.toPath.toString,
                                            user.name.toString
                                        )
                                )
                                _ <- deleteUserDirectory(userDir)
                                response <- ticketServiceApi.deleteUser(user.uid) *> accountManagementRepo
                                    .deleteAccount(
                                        user.uid
                                    ) *> SeeOther(Location(rootUri)).map(
                                        _.removeCookie(Constants.authenticationCookieName.toString)
                                    )
                            } yield response
                        } else
                            BadRequest(
                                views.html.account.settings(lang = language)(
                                    csrf,
                                    Option(s"Smederee/~${user.name} - Settings"),
                                    user
                                )(
                                    actionBaseUri,
                                    deleteAction,
                                    languageAction,
                                    validateAction
                                )
                            )
                } yield response
            }
    }

    private val deleteSshKey: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / "user" / "settings" / "ssh" / "delete" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf <- Sync[F].delay(ar.req.getCsrfToken)
                    formData <- Sync[F].delay {
                        urlForm.values.map { t =>
                            val (key, values) = t
                            (
                                key,
                                values.headOption.getOrElse("")
                            ) // Pick the first value (a field might get submitted multiple times)!
                        }
                    }
                    actionBaseUri <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings"))
                    userIsSure    <- Sync[F].delay(formData.get("i-am-sure").exists(_ === "yes"))
                    sshKeyId <- Sync[F].delay {
                        Try(formData.get("ssh-key-id").map(UUID.fromString)) match {
                            case scala.util.Failure(error) =>
                                log.error("Error while parsing ssh key id upon key delete request.", error)
                                None
                            case scala.util.Success(Some(keyId)) => Option(keyId)
                            case _                               => None
                        }
                    }
                    _ <- userIsSure match {
                        case false => Sync[F].pure(None)
                        case true  => sshKeyId.traverse(id => accountManagementRepo.deleteSshKey(id, user.uid))
                    }
                    resp <- SeeOther(Location(actionBaseUri.addSegment("ssh")))
                } yield resp
            }
    }

    private val sendValidationEmail: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / "user" / "settings" / "email" / "validate" as user =>
            for {
                csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                token    <- Sync[F].delay(ValidationToken.generate)
                _        <- accountManagementRepo.setValidationToken(user.uid, token)
                from     <- Sync[F].delay(FromAddress("noreply@smeder.ee"))
                to       <- Sync[F].delay(NonEmptyList.of(user.email.toToAddress))
                uri      <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings/email/validate"))
                subject  <- Sync[F].delay(SubjectLine("Smederee - Please validate your email address."))
                body <- Sync[F].delay(
                    TextBody(views.txt.emails.validate(user, token, uri).toString)
                ) // TODO: extension method?
                message <- Sync[F].delay(EmailMessage(from, to, List.empty, List.empty, subject, body))
                result  <- emailMiddleware.send(message)
                response <- result match {
                    case Left(error) => InternalServerError(s"An error occured: $error")
                    case Right(_)    => SeeOther(Location(configuration.external.createFullUri(uri"user/settings")))
                }
            } yield response
    }

    private val setLanguage: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / "user" / "settings" / "language" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf <- Sync[F].delay(ar.req.getCsrfToken)
                    formData <- Sync[F].delay {
                        urlForm.values.map { t =>
                            val (key, values) = t
                            (
                                key,
                                values.headOption.getOrElse("")
                            ) // Pick the first value (a field might get submitted multiple times)!
                        }
                    }
                    languageCode   <- Sync[F].delay(formData.get("language").flatMap(LanguageCode.from))
                    rootUri        <- Sync[F].delay(configuration.external.createFullUri(uri"/"))
                    actionBaseUri  <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings"))
                    deleteAction   <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings/delete"))
                    languageAction <- Sync[F].delay(configuration.external.createFullUri(uri"/user/settings/language"))
                    validateAction <- Sync[F].delay(
                        configuration.external.createFullUri(uri"user/settings/email/validate")
                    )
                    _ <- accountManagementRepo.setLanguage(user.uid, languageCode)
                    _ <- ticketServiceApi.createOrUpdateUser(TicketsUser(user.uid, user.name, user.email, languageCode))
                    resp <- SeeOther(Location(actionBaseUri))
                } yield resp
            }
    }

    private val showAccountSettings: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "user" / "settings" as user =>
            for {
                csrf           <- Sync[F].delay(ar.req.getCsrfToken)
                language       <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                actionBaseUri  <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings"))
                deleteAction   <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings/delete"))
                languageAction <- Sync[F].delay(configuration.external.createFullUri(uri"/user/settings/language"))
                validateAction <- Sync[F].delay(
                    configuration.external.createFullUri(uri"user/settings/email/validate")
                )
                resp <- Ok(
                    views.html.account
                        .settings(lang = language)(csrf, Option(s"Smederee/~${user.name} - Settings"), user)(
                            actionBaseUri,
                            deleteAction,
                            languageAction,
                            validateAction
                        )
                )
            } yield resp
    }

    private val showAccountOrganisations: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "user" / "settings" / "organisations" as user =>
            for {
                csrf          <- Sync[F].delay(ar.req.getCsrfToken)
                language      <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                actionBaseUri <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings"))
                orgs          <- organisationsRepo.allByOwner(user.uid).compile.toList
                organisationActionBaseUri <- Sync[F].delay(
                    configuration.external.createFullUri(uri"user/settings/organisations")
                )
                resp <- Ok(
                    views.html.account
                        .settingsOrganisations(lang = language)(
                            csrf,
                            Option(s"Smederee/~${user.name} - Organisations"),
                            user
                        )(actionBaseUri, orgs, organisationActionBaseUri)
                )
            } yield resp
    }

    private val showAccountSshSettings: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "user" / "settings" / "ssh" as user =>
            for {
                csrf          <- Sync[F].delay(ar.req.getCsrfToken)
                language      <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                actionBaseUri <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings"))
                addAction     <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings/ssh/add"))
                deleteAction  <- Sync[F].delay(configuration.external.createFullUri(uri"user/settings/ssh/delete"))
                keys          <- accountManagementRepo.listSshKeys(user.uid).compile.toList
                resp <- Ok(
                    views.html.account
                        .sshSettings(lang = language)(csrf, Option(s"Smederee/~${user.name} - SSH Settings"), user)(
                            actionBaseUri,
                            addAction,
                            deleteAction,
                            keys
                        )()
                )
            } yield resp
    }

    private val validateEmailAddress: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / "user" / "settings" / "email" / "validate" / ValidationTokenPathParameter(
                token
            ) =>
            for {
                csrf    <- Sync[F].delay(req.getCsrfToken)
                account <- accountManagementRepo.findByValidationToken(token)
                _       <- account.traverse(user => accountManagementRepo.markAsValidated(user.uid))
                _ <- account.traverse(user =>
                    ticketServiceApi.createOrUpdateUser(TicketsUser(user.uid, user.name, user.email, user.language))
                )
                resp <- SeeOther(Location(configuration.external.createFullUri(uri"/")))
            } yield resp
    }

    val protectedRoutes =
        addSshKey <+> deleteAccount <+> deleteSshKey <+> sendValidationEmail <+> setLanguage <+> showAccountSettings <+> showAccountOrganisations <+> showAccountSshSettings

    val routes = validateEmailAddress

}
