/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.data.*
import cats.syntax.all.*
import de.smederee.hub.forms.*
import de.smederee.hub.forms.types.*
import de.smederee.ssh.*

/** Data container for the form to add a new public ssh key.
  *
  * @param name
  *   An optional name for the key.
  * @param keyString
  *   A string containing a valid public key.
  */
final case class AddPublicSshKeyForm(name: Option[KeyComment], keyString: SshPublicKeyString)

object AddPublicSshKeyForm extends FormValidator[AddPublicSshKeyForm] {
    val fieldName: FormField = FormField("name")
    val fieldKey: FormField  = FormField("key")

    override def validate(data: Map[String, String]): ValidatedNec[FormErrors, AddPublicSshKeyForm] = {
        val name = data.get(fieldName).fold(None.validNec)(name => KeyComment.from(name).validNec)
        val key = data
            .get(fieldKey)
            .fold(FormFieldError("No ssh public key given!").invalidNec)(string =>
                SshPublicKeyString
                    .from(string.trim)
                    .fold(FormFieldError("Invalid ssh public key!").invalidNec)(_.validNec)
            )
            .leftMap(errors => NonEmptyChain.of(Map(fieldKey -> errors.toList)))
        (name, key).mapN { case (validName, validKey) =>
            AddPublicSshKeyForm(validName, validKey)
        }
    }

}
