/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.*

import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.hub.RequestHelpers.instances.given_RequestHelpers_Request
import de.smederee.hub.SessionHelpers.instances.*
import de.smederee.hub.config.AuthenticationTimeouts
import de.smederee.security.*
import org.http4s.*

/** Implementation of needed functionality to provide an authentication middleware layer for http4s.
  */
trait AuthenticationMiddleware {

    /** Helper function to find and validate the authentication token from a request and extract the session id from it.
      *
      * @param request
      *   A http4s request.
      * @param signAndValidate
      *   A class providing functionality to validate the authentication token.
      * @return
      *   Either an error message or the extracted session id.
      */
    protected def extractSessionId[F[_]: Sync](
        request: Request[F],
        signAndValidate: SignAndValidate
    ): Either[ErrorMessageKey, SessionId] =
        for {
            token  <- request.getAuthenticationToken.toRight(ErrorMessageKey("error.auth.token.not-found"))
            authId <- signAndValidate.validate(token).toRight(ErrorMessageKey("error.auth.token.invalid"))
            id     <- SessionId.from(authId).toRight(ErrorMessageKey("error.auth.token.invalid"))
        } yield id

    /** Extract an authentication token from a request, validate it and try to load the corresponding user account from
      * the database repository if the token is valid. The functionality is wrapped into a [[cats.data.Kleisli]] to be
      * composable within the http4s middleware.
      *
      * @param repo
      *   A database repository providing the functionality to find a user account in the database.
      * @param signAndValidate
      *   A class providing functionality to validate the authentication token.
      * @param timeouts
      *   Timeouts related to the authentication and session management.
      * @return
      *   Either an error or the loaded user account.
      */
    def authenticateUser[F[_]: Sync](
        repo: AuthenticationRepository[F],
        signAndValidate: SignAndValidate,
        timeouts: AuthenticationTimeouts
    ): Kleisli[F, Request[F], Either[ErrorMessageKey, Account]] =
        Kleisli { request =>
            val sessionId = extractSessionId(request, signAndValidate)
            sessionId.traverse(resolveUser(repo)(timeouts).run).map {
                case Left(error)          => error.asLeft
                case Right(None)          => ErrorMessageKey("error.auth.unauthorized.message").asLeft
                case Right(Some(account)) => account.asRight
            }
        }

    /** This is an Implementation of [[authenticateUser]] which in addition to [[cats.data.Kleisli]] also uses
      * [[cats.data.OptionT]] to make it useable with the fall through functionality of the http4s authentication
      * middleware. The function also extracts an authentication token from a requests, validates it and tries to load
      * the corresponding user account from the database repository if the token is valid.
      *
      * @param repo
      *   A database repository providing the functionality to find a user account in the database.
      * @param signAndValidate
      *   A class providing functionality to validate the authentication token.
      * @param timeouts
      *   Timeouts related to the authentication and session management.
      * @return
      *   An option to the account of a logged in user.
      */
    def authenticateUserWithFallThrough[F[_]: Sync](
        repo: AuthenticationRepository[F],
        signAndValidate: SignAndValidate,
        timeouts: AuthenticationTimeouts
    ): Kleisli[OptionT[F, *], Request[F], Account] =
        Kleisli { request =>
            val sessionId = extractSessionId(request, signAndValidate)
            OptionT(sessionId.traverse(resolveUser(repo)(timeouts).run).map(_.getOrElse(None)))
        }

    /** Try to resolve a user account using the provided database repository. It is wrapped into a [[cats.data.Kleisli]]
      * to be composable within the http4s middleware. The function expects a [[SessionId]] to be passed to it which is
      * used to obtain a session and the related user account.
      *
      * @param repo
      *   A database repository providing the functionality to find a user account in the database.
      * @param timeouts
      *   Timeouts related to the authentication and session management.
      * @return
      *   An option to the user account if it exists.
      */
    def resolveUser[F[_]: Sync](
        repo: AuthenticationRepository[F]
    )(timeouts: AuthenticationTimeouts): Kleisli[F, SessionId, Option[Account]] =
        Kleisli { sessionId =>
            for {
                session      <- repo.findUserSession(sessionId)
                currentTime  <- Sync[F].delay(OffsetDateTime.now(ZoneOffset.UTC))
                validSession <- Sync[F].delay(session.filterNot(_.hasReachedAbsoluteTimeout(timeouts)(currentTime)))
                account      <- session.traverse(s => repo.findAccount(s.uid)).map(_.getOrElse(None))
            } yield account
        }
}
