/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import de.smederee.email.EmailAddress
import de.smederee.security.*
import fs2.Stream

/** A base class for database functionality related to the authentication process.
  *
  * ### General notes ###
  *
  * {{{
  * 1. An account is considered *locked* **NOT** by the presence of an unlock token **BUT** by the presence
  *    of the `locked_at` date!
  * 2. All functions that are used to find accounts by email, id or name must return only *unlocked* accounts!
  * 3. Also the function `findPasswordHashAndAttempts` must only consider *unlocked* accounts!
  * }}}
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class AuthenticationRepository[F[_]] {

    /** Return all accounts found in the database sorted ascending by username.
      *
      * @return
      *   A stream containing all accounts sorted ascending by username which may be empty.
      */
    def allAccounts(): Stream[F, Account]

    /** Create a user session in the database.
      *
      * @param session
      *   The session information that shall be persisted.
      * @return
      *   The number of affected database rows.
      */
    def createUserSession(session: Session): F[Int]

    /** Delete all user sessions of the given user from the database.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   The number of affected database rows.
      */
    def deleteAllUserSessions(uid: UserId): F[Int]

    /** Delete the user session with the given ID from the database.
      *
      * @param id
      *   The globally unique ID of the session.
      * @return
      *   The number of affected database rows.
      */
    def deleteUserSession(id: SessionId): F[Int]

    /** Search for the unlocked account with the given unique id from the database.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   An option to the loaded account if it exists.
      */
    def findAccount(uid: UserId): F[Option[Account]]

    /** Search for the unlocked account with the given email address in the database and return the first found result.
      *
      * @param email
      *   The email address of the user account which must be unique according to our requirements.
      * @return
      *   An option to the found account if it exists.
      */
    def findAccountByEmail(email: EmailAddress): F[Option[Account]]

    /** Search for the unlocked account with the given name in the database and return the first found result.
      *
      * @param name
      *   The username which must be unique according to our requirements.
      * @return
      *   An option to the found account if it exists.
      */
    def findAccountByName(name: Username): F[Option[Account]]

    /** Search for a locked account with the given unlock token and return the first found result.
      *
      * An account is considered *locked* **NOT** by the presence of an unlock token **BUT** by the presence of the
      * `locked_at` date!
      *
      * @param name
      *   The username which must be unique according to our requirements.
      * @param token
      *   An unlock token which must be present if given.
      * @return
      *   An option to the account if it exists and is locked using the given token.
      */
    def findLockedAccount(name: Username)(token: Option[UnlockToken]): F[Option[Account]]

    /** Retrieve the password hash and the number of failed login attempts from the database.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   An option to a tuple containing the password hash and the number of failed login attempts.
      */
    def findPasswordHashAndAttempts(uid: UserId): F[Option[(PasswordHash, Int)]]

    /** Find the session with the given ID in the database.
      *
      * @param id
      *   A session id which is supposed to be globally unique.
      * @return
      *   An option to the desired session if it exists.
      */
    def findUserSession(id: SessionId): F[Option[Session]]

    /** Increment the number of failed login attempts for the user with the given name.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   The number of affected database rows.
      */
    def incrementFailedAttempts(uid: UserId): F[Int]

    /** Lock the account with the given id and set the value of the `unlock_token` column to the given token which might
      * be `None` (`null` in the database).
      *
      * @param uid
      *   The unique id of the user account.
      * @param token
      *   An unlock token which must be present.
      * @return
      *   The number of affected database rows.
      */
    def lockAccount(uid: UserId)(token: Option[UnlockToken]): F[Int]

    /** Reset the number of failed login attempts (set it to zero) for the user account with the given id.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   The number of affected database rows.
      */
    def resetFailedAttempts(uid: UserId): F[Int]

    /** Unlock the user account with the given id. The process of unlocking clears both columns related to locking:
      * `locked_at` and `unlock_token`!
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   The number of affected database rows.
      */
    def unlockAccount(uid: UserId): F[Int]

}
