/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.effect.*
import cats.syntax.all.*
import org.flywaydb.core.Flyway
import org.flywaydb.core.api.configuration.FluentConfiguration
import org.flywaydb.core.api.output.MigrateResult

/** Provide functionality to migrate the database used by the service.
  */
final class DatabaseMigrator[F[_]: Sync] {

    /** Apply pending migrations to the database if needed using the underlying Flyway library.
      *
      * @param url
      *   The JDBC connection URL **without** username and password.
      * @param user
      *   The username (login) needed to authenticate against the database.
      * @param pass
      *   The password needed to authenticate against the database.
      * @return
      *   A migrate result object holding information about executed migrations and the schema. See the Java-Doc of
      *   Flyway for details.
      */
    def migrate(url: String, user: String, pass: String): F[MigrateResult] =
        for {
            flyway <- Sync[F].delay(DatabaseMigrator.configureFlyway(url, user, pass).load())
            result <- Sync[F].delay(flyway.migrate())
        } yield result
}

object DatabaseMigrator {

    /** Configure an instance of flyway with our defaults and the given credentials for a database connection. The
      * returned instance must be activated by calling the `.load()` method.
      *
      * @param url
      *   The JDBC connection URL **without** username and password.
      * @param user
      *   The username (login) needed to authenticate against the database.
      * @param pass
      *   The password needed to authenticate against the database.
      * @return
      *   An instance configuration which can be turned into a flyway database migrator by calling the `.load()` method.
      */
    def configureFlyway(url: String, user: String, pass: String): FluentConfiguration =
        Flyway.configure().defaultSchema("hub").locations("classpath:db/migration/hub").dataSource(url, user, pass)

}
