/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.util.UUID

import cats.effect.*
import cats.syntax.all.*
import de.smederee.email.EmailAddress
import de.smederee.i18n.LanguageCode
import de.smederee.security.*
import doobie.*
import doobie.Fragments.*
import doobie.implicits.*
import doobie.postgres.implicits.*
import fs2.Stream

final class DoobieAuthenticationRepository[F[_]: Sync](tx: Transactor[F]) extends AuthenticationRepository[F] {
    given Meta[EmailAddress] = Meta[String].timap(EmailAddress.apply)(_.toString)
    given Meta[FullName]     = Meta[String].timap(FullName.apply)(_.toString)
    given Meta[LanguageCode] = Meta[String].timap(LanguageCode.apply)(_.toString)
    given Meta[PasswordHash] = Meta[String].timap(PasswordHash.apply)(_.toString)
    given Meta[SessionId]    = Meta[String].timap(SessionId.apply)(_.toString)
    given Meta[UnlockToken]  = Meta[String].timap(UnlockToken.apply)(_.toString)
    given Meta[UserId]       = Meta[UUID].timap(UserId.apply)(_.toUUID)
    given Meta[Username]     = Meta[String].timap(Username.apply)(_.toString)

    private val lockedFilter    = fr"""locked_at IS NOT NULL"""
    private val notLockedFilter = fr"""locked_at IS NULL"""
    private val selectAccountColumns =
        fr"""SELECT uid, name, email, full_name, validated_email, language FROM hub.accounts"""

    override def allAccounts(): Stream[F, Account] = {
        val query = selectAccountColumns ++ fr"""ORDER BY name ASC"""
        query.query[Account].stream.transact(tx)
    }

    override def createUserSession(session: Session): F[Int] =
        sql"""INSERT INTO hub.sessions (id, uid, created_at, updated_at) VALUES (${session.id}, ${session.uid}, ${session.createdAt}, ${session.updatedAt})""".update.run
            .transact(tx)

    override def deleteAllUserSessions(uid: UserId): F[Int] =
        sql"""DELETE FROM hub.sessions WHERE uid = $uid""".update.run.transact(tx)

    override def deleteUserSession(id: SessionId): F[Int] =
        sql"""DELETE FROM hub.sessions WHERE id = $id""".update.run.transact(tx)

    override def findAccount(uid: UserId): F[Option[Account]] = {
        val uidFilter = fr"""uid = $uid"""
        val query     = selectAccountColumns ++ whereAnd(notLockedFilter, uidFilter) ++ fr"""LIMIT 1"""
        query.query[Account].option.transact(tx)
    }

    override def findAccountByEmail(email: EmailAddress): F[Option[Account]] = {
        val emailFilter = fr"""email = $email"""
        val query       = selectAccountColumns ++ whereAnd(notLockedFilter, emailFilter) ++ fr"""LIMIT 1"""
        query.query[Account].option.transact(tx)
    }

    override def findAccountByName(name: Username): F[Option[Account]] = {
        val nameFilter = fr"""name = $name"""
        val query      = selectAccountColumns ++ whereAnd(notLockedFilter, nameFilter) ++ fr"""LIMIT 1"""
        query.query[Account].option.transact(tx)
    }

    override def findLockedAccount(name: Username)(token: Option[UnlockToken]): F[Option[Account]] = {
        val nameFilter  = fr"""name = $name"""
        val tokenFilter = token.map(token => fr"""unlock_token = $token""")
        val query =
            selectAccountColumns ++ whereAndOpt(lockedFilter.some, nameFilter.some, tokenFilter) ++ fr"""LIMIT 1"""
        query.query[Account].option.transact(tx)
    }

    override def findPasswordHashAndAttempts(uid: UserId): F[Option[(PasswordHash, Int)]] = {
        val uidFilter = fr"""uid = $uid"""
        val query = fr"""SELECT password, failed_attempts FROM hub.accounts""" ++ whereAnd(
                notLockedFilter,
                uidFilter
            ) ++ fr"""LIMIT 1"""
        query.query[(PasswordHash, Int)].option.transact(tx)
    }

    override def findUserSession(id: SessionId): F[Option[Session]] =
        sql"""SELECT id, uid, created_at, updated_at FROM hub.sessions WHERE id = $id LIMIT 1"""
            .query[Session]
            .option
            .transact(tx)

    override def incrementFailedAttempts(uid: UserId): F[Int] =
        sql"""UPDATE hub.accounts SET failed_attempts = failed_attempts + 1 WHERE uid = $uid""".update.run
            .transact(tx)

    override def lockAccount(uid: UserId)(token: Option[UnlockToken]): F[Int] =
        sql"""UPDATE hub.accounts SET locked_at = NOW(), unlock_token = $token WHERE uid = $uid""".update.run
            .transact(tx)

    override def resetFailedAttempts(uid: UserId): F[Int] =
        sql"""UPDATE hub.accounts SET failed_attempts = 0 WHERE uid = $uid""".update.run.transact(tx)

    override def unlockAccount(uid: UserId): F[Int] =
        sql"""UPDATE hub.accounts SET locked_at = NULL, unlock_token = NULL WHERE uid = $uid""".update.run
            .transact(tx)
}
