/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.effect.*
import cats.syntax.all.*
import de.smederee.html.LinkTools.*
import de.smederee.hub.RequestHelpers.instances.given
import de.smederee.hub.config.*
import de.smederee.i18n.LanguageCode
import org.http4s.*
import org.http4s.dsl.Http4sDsl
import org.http4s.implicits.*
import org.http4s.twirl.TwirlInstances.*

/** This bundles all routing related to possible landing pages for users entering the site.
  *
  * Please note that due to the routing logic of http4s catch-all pages (`-> Root`) should be put last in the list of
  * routes!
  *
  * @param configuration
  *   The hub service configuration.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class LandingPageRoutes[F[_]: Async](configuration: ServiceConfig) extends Http4sDsl[F] {
    private val linkConfig = configuration.external
    // The base URI for our site which that be passed into some templates which create links themselfes.
    private val baseUri = linkConfig.createFullUri(Uri())
    // The URL that shall be used in the `action` field of the form.
    private val signupUri = linkConfig.createFullUri(uri"/signup")

    private val contact: AuthedRoutes[Account, F] = AuthedRoutes.of { case ar @ GET -> Root / "contact" as user =>
        for {
            csrf     <- Sync[F].delay(ar.req.getCsrfToken)
            language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
            resp     <- Ok(views.html.contact(baseUri, lang = language)(csrf, "Smederee - Contact".some, user.some))
        } yield resp
    }

    private val contactForGuests: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root / "contact" =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            resp <- Ok(views.html.contact(baseUri)(csrf, "Smederee - Contact".some))
        } yield resp
    }

    private val imprint: AuthedRoutes[Account, F] = AuthedRoutes.of { case ar @ GET -> Root / "imprint" as user =>
        for {
            csrf     <- Sync[F].delay(ar.req.getCsrfToken)
            language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
            resp <- Ok(
                views.html.imprint(baseUri, lang = language)(csrf, "Smederee - Imprint / Impressum".some, user.some)
            )
        } yield resp
    }

    private val imprintForGuests: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root / "imprint" =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            resp <- Ok(views.html.imprint(baseUri)(csrf, "Smederee - Imprint / Impressum".some))
        } yield resp
    }

    private val mainSite: AuthedRoutes[Account, F] = AuthedRoutes.of { case ar @ GET -> Root as user =>
        for {
            csrf     <- Sync[F].delay(ar.req.getCsrfToken)
            language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
            resp <- Ok(
                views.html
                    .index(baseUri, lang = language)()(signupUri, csrf, "Welcome to the Smederee!".some, user.some)
            )
        } yield resp
    }

    private val mainSiteForGuests: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            resp <- Ok(views.html.index(baseUri)()(signupUri, csrf, "Welcome to the Smederee!".some))
        } yield resp
    }

    private val privacyPolicy: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "privacy-policy" as user =>
            for {
                csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                resp <- Ok(
                    views.html
                        .privacyPolicy(baseUri, lang = language)(csrf, "Smederee - Privacy Policy".some, user.some)
                )
            } yield resp
    }

    private val privacyPolicyForGuests: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root / "privacy-policy" =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            resp <- Ok(views.html.privacyPolicy(baseUri)(csrf, "Smederee - Privacy Policy".some))
        } yield resp
    }

    private val publicAlpha: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "public-alpha" as user =>
            for {
                csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                resp <- Ok(
                    views.html
                        .publicAlpha(baseUri, lang = language)(
                            csrf,
                            "Smederee - Details about our public alpha.".some,
                            user.some
                        )
                )
            } yield resp
    }

    private val publicAlphaForGuests: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root / "public-alpha" =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            resp <- Ok(views.html.publicAlpha(baseUri)(csrf, "Smederee - Details about our public alpha.".some))
        } yield resp
    }

    private val termsOfUse: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "terms-of-use" as user =>
            for {
                csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                resp <- Ok(
                    views.html.termsOfUse(baseUri, lang = language)(csrf, "Smederee - Terms of Use".some, user.some)
                )
            } yield resp
    }

    private val termsOfUseForGuests: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root / "terms-of-use" =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            resp <- Ok(views.html.termsOfUse(baseUri)(csrf, "Smederee - Terms of Use".some))
        } yield resp
    }

    val protectedRoutes = contact <+> imprint <+> privacyPolicy <+> publicAlpha <+> termsOfUse <+> mainSite

    val routes =
        contactForGuests <+> imprintForGuests <+> privacyPolicyForGuests <+> publicAlphaForGuests <+> termsOfUseForGuests <+> mainSiteForGuests

}
