/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.data.*
import cats.syntax.all.*
import de.smederee.hub.forms.*
import de.smederee.hub.forms.types.*
import de.smederee.security.*

/** A data container for the login form.
  *
  * @param name
  *   The username which is supposed to be globally unique.
  * @param password
  *   The password which must match the one in the database.
  */
final case class LoginForm(name: Username, password: Password)

object LoginForm extends FormValidator[LoginForm] {
    val fieldName: FormField     = FormField("name")
    val fieldPassword: FormField = FormField("password")

    override def validate(data: Map[String, String]): ValidatedNec[FormErrors, LoginForm] = {
        val genericError = FormFieldError("Invalid credentials!") // We just return a generic error.
        val name: ValidatedNec[FormErrors, Username] = data
            .get(fieldName)
            .fold(genericError.invalidNec)(s => Username.from(s).fold(genericError.invalidNec)(_.validNec))
            .leftMap(es => NonEmptyChain.of(Map(fieldName -> es.toList)))
        val password: ValidatedNec[FormErrors, Password] = data
            .get(fieldPassword)
            .fold(genericError.invalidNec)(s => Password.from(s).fold(genericError.invalidNec)(_.validNec))
            .leftMap(es => NonEmptyChain.of(Map(fieldPassword -> es.toList)))
        (name, password).mapN { case (n, pw) =>
            LoginForm(n, pw)
        }
    }
}
