/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import de.smederee.security.UserId
import de.smederee.security.Username
import fs2.Stream

/** A base class for a database repository that should handle all organisation related functionality.
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class OrganisationRepository[F[_]] {

    /** Add an administrator to the organisation.
      *
      * @param organisationId
      *   The unique ID of the organisation.
      * @param user
      *   The globally unique ID of the user to be added as an administrator.
      * @return
      *   The number of affected database rows.
      */
    def addAdministrator(organisationId: OrganisationId)(user: UserId): F[Int]

    /** Find all organisations of which the user with the given id is an owner.
      *
      * @param owner
      *   The unique user id of an account.
      * @return
      *   A stream with all organisation of which the user is an owner.
      */
    def allByOwner(owner: UserId): Stream[F, Organisation]

    /** Create the database entry for the given organisation.
      *
      * @param org
      *   The organisation that shall be written to the database.
      * @return
      *   The number of affected database rows.
      */
    def create(org: Organisation): F[Int]

    /** Delete the organisation with the given id.
      *
      * @param organisationId
      *   The unique ID of the organisation.
      * @return
      *   The number of affected database rows.
      */
    def delete(organisationId: OrganisationId): F[Int]

    /** Find the organisation with the given name.
      *
      * @param organisationId
      *   The unique ID of the organisation.
      * @return
      *   An option to the found organisation.
      */
    def find(organisationId: OrganisationId): F[Option[Organisation]]

    /** Find the organisation with the given name.
      *
      * @param name
      *   An organisation name which is actually a valid username.
      * @return
      *   An option to the found organisation.
      */
    def findByName(name: Username): F[Option[Organisation]]

    /** Find the account of an organisation owner.
      *
      * @param organisationId
      *   The unique ID of the organisation.
      * @return
      *   An option to the found account.
      */
    def findOwner(organisationId: OrganisationId): F[Option[Account]]

    /** Return the accounts of all users that are administrators of the given organisation.
      *
      * @param organisationId
      *   The unique ID of the organisation.
      * @return
      *   A stream of accounts that may be empty.
      */
    def getAdministrators(organisationId: OrganisationId): Stream[F, Account]

    /** Remove an administrator from the organisation.
      *
      * @param organisationId
      *   The unique ID of the organisation.
      * @param user
      *   The globally unique ID of the user to be removed.
      * @return
      *   The number of affected database rows.
      */
    def removeAdministrator(organisationId: OrganisationId)(user: UserId): F[Int]

    /** Update the database entry for the given organisation.
      *
      * @param org
      *   The organisation that shall be written to the database.
      * @return
      *   The number of affected database rows.
      */
    def update(org: Organisation): F[Int]
}
