/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.OffsetDateTime

import de.smederee.security.*

/** A base class for database functionality related to resetting a user password.
  *
  * ### General notes ###
  *
  * {{{
  * 1. An account is considered *locked* **NOT** by the presence of an unlock token **BUT** by the presence
  *    of the `locked_at` date!
  * 2. All functions that are used to find accounts by email, id or name must return only *unlocked* accounts!
  * 3. Also the function `findPasswordHashAndAttempts` must only consider *unlocked* accounts!
  * }}}
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class ResetPasswordRepository[F[_]] {

    /** Find a user account using the given name and reset token. This function shall not check for the expiration date
      * but for the expiration date being NULL and for the given combination of username and reset token.
      *
      * The nulled out expiration date is considered proof that the link with the reset url has been called.
      *
      * @param name
      *   The username which must be unique according to our requirements.
      * @param token
      *   A token that must be present in the reset token column.
      * @return
      *   An option to the found user account.
      */
    def findByNameAndResetPasswordToken(name: Username, token: ResetToken): F[Option[Account]]

    /** Find a user account via the given password reset token which must not be expired.
      *
      * @param token
      *   A token that must be present in the reset token column.
      * @return
      *   An option to the found user account.
      */
    def findByResetPasswordToken(token: ResetToken): F[Option[Account]]

    /** Remove just the expiration date for the reset password token from the database. This is used to prevent opening
      * the reset password link a second time while still being able to perform a password change.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   The number of affected database rows.
      */
    def removeResetPasswordExpirationDate(uid: UserId): F[Int]

    /** Remove the reset password token and the expiration date from the database.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   The number of affected database rows.
      */
    def removeResetPasswordToken(uid: UserId): F[Int]

    /** Set the password hash for the account with the given user id.
      *
      * @param uid
      *   The unique id of the user account.
      * @param hash
      *   The password hash for the account.
      * @return
      *   The number of affected database rows.
      */
    def setPassword(uid: UserId)(hash: PasswordHash): F[Int]

    /** Set the password reset token for the given user and the related expiration time.
      *
      * @param uid
      *   The unique id of the user account.
      * @param token
      *   A token that is written into the reset token column and must be present in the reset request uri.
      * @param tokenExpiration
      *   The timestamp when the token shall expire.
      * @return
      *   The number of affected database rows.
      */
    def setResetPasswordToken(uid: UserId)(token: ResetToken, tokenExpiration: OffsetDateTime): F[Int]

}
