/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.OffsetDateTime

import cats.*
import cats.syntax.all.*
import de.smederee.security.UserId

/** A user session which is used to track logged in users.
  *
  * @param id
  *   A globally unique ID for the session.
  * @param uid
  *   The unique ID of the user account related to the session.
  * @param createdAt
  *   The timestamp when the session was created.
  * @param updatedAt
  *   The session ID should be re-generated in regular intervals resulting in a copy of the old session entry with a new
  *   ID and the corresponding timestamp in this field.
  */
final case class Session(id: SessionId, uid: UserId, createdAt: OffsetDateTime, updatedAt: OffsetDateTime)

object Session {
    given Eq[Session] =
        Eq.instance[Session] { (a, b) =>
            a.id === b.id && a.uid === b.uid && a.createdAt.compareTo(b.createdAt) === 0 && a.updatedAt.compareTo(
                b.updatedAt
            ) === 0
        }

    /** Create a new session object for the given user and generate a new session id.
      *
      * @param uid
      *   The unique ID of the user account related to the session.
      * @param createdAt
      *   The creation timestamp to be used for the session.
      * @param updatedAt
      *   The session ID should be re-generated in regular intervals resulting in a copy of the old session entry with a
      *   new ID and the corresponding timestamp in this field.
      * @return
      *   A new session with a generated id for the given user and timestamp.
      */
    def create(uid: UserId, createdAt: OffsetDateTime, updatedAt: OffsetDateTime): Session =
        Session(
            id = SessionId.generate,
            uid = uid,
            createdAt = createdAt,
            updatedAt = updatedAt
        )
}
