/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.data.*
import cats.syntax.all.*
import de.smederee.email.EmailAddress
import de.smederee.hub.forms.*
import de.smederee.hub.forms.types.*
import de.smederee.security.*

/** A data container for our signup form.
  *
  * Please note that both `name` and `email` have to be checked against possibly existing entries in the database!
  *
  * @param name
  *   The username which is desired to be taken by the user.
  * @param email
  *   The email address of the user.
  * @param password
  *   The password of the user.
  */
final case class SignupForm(name: Username, email: EmailAddress, password: Password)

object SignupForm extends FormValidator[SignupForm] {
    val fieldName: FormField     = FormField("name")
    val fieldEmail: FormField    = FormField("email")
    val fieldPassword: FormField = FormField("password")
    override def validate(data: Map[String, String]): ValidatedNec[FormErrors, SignupForm] = {
        val email: ValidatedNec[FormErrors, EmailAddress] = data
            .get(fieldEmail)
            .fold(FormFieldError("No email address given!").invalidNec)(s =>
                EmailAddress.from(s).fold(FormFieldError("Invalid email address!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldEmail -> es.toList)))
        val name: ValidatedNec[FormErrors, Username] = data
            .get(fieldName)
            .fold(FormFieldError("No username given!").invalidNec)(s =>
                Username.from(s).fold(FormFieldError("Invalid username!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldName -> es.toList)))
        val password: ValidatedNec[FormErrors, Password] = data
            .get(fieldPassword)
            .fold("No password given!".invalidNec)(Password.validate)
            .leftMap(es => NonEmptyChain.of(Map(fieldPassword -> es.toList.map(FormFieldError.apply))))
        (email, name, password).mapN { case (validEmail, validName, validPassword) =>
            SignupForm(validName, validEmail, validPassword)
        }
    }
}
