/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import de.smederee.email.EmailAddress
import de.smederee.security.*

/** A base class for our signup repository which provides the needed database functions.
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class SignupRepository[F[_]] {

    /** Create the given user account in the database.
      *
      * @param account
      *   The account to be created.
      * @param hash
      *   The password hash for the account.
      * @return
      *   The number of affected database rows.
      */
    def createAccount(account: Account, hash: PasswordHash): F[Int]

    /** Find the given email address in the database accounts table. This function can be used to check if an email
      * address has already been taken.
      *
      * @param address
      *   An email address which shall be searched for.
      * @return
      *   An option which is either empty or contains the email address if it exists in the database.
      */
    def findEmail(address: EmailAddress): F[Option[EmailAddress]]

    /** Find the given username in the database accounts table. This function can be used to check if a username has
      * already been taken.
      *
      * @param name
      *   The username which shall be searched for.
      * @return
      *   An option which is either empty or contains the username if it exists in the database.
      */
    def findUsername(name: Username): F[Option[Username]]

}
