/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import de.smederee.security.Username
import fs2.Stream

/** A base class for a database repository that should handle all functionality regarding vcs repositories and their
  * metadata in the database.
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class VcsMetadataRepository[F[_]] {

    /** Create an entry in the forks table to save the relationship between to repositories.
      *
      * @param source
      *   The ID of the original repository from which was forked.
      * @param target
      *   The ID of the forked repository.
      * @return
      *   The number of affected database rows.
      */
    def createFork(source: VcsRepositoryId, target: VcsRepositoryId): F[Int]

    /** Create a database entry for the given vcs repository.
      *
      * @param repository
      *   The vcs repository metadata that shall be written to the database.
      * @return
      *   The number of affected database rows.
      */
    def createVcsRepository(repository: VcsRepository): F[Int]

    /** Delete the repository from the database.
      *
      * @param repository
      *   The vcs repository metadata that shall be deleted from the database.
      * @return
      *   The number of affected database rows.
      */
    def deleteVcsRepository(repository: VcsRepository): F[Int]

    /** Search for the vcs repository entry with the given owner and name.
      *
      * @param owner
      *   Data about the owner of the repository containing information needed to query the database.
      * @param name
      *   The repository name which must be unique in regard to the owner.
      * @return
      *   An option to the successfully found vcs repository entry.
      */
    def findVcsRepository(owner: VcsRepositoryOwner, name: VcsRepositoryName): F[Option[VcsRepository]]

    /** Find all branches (created via the `fork` functionality) for the repository with the given id.
      *
      * @param originalRepositoryId
      *   The id from the original repository from which the branches (forks) were created.
      * @return
      *   A stream of tuples holding the username of the branch owner and the branch (repository) names.
      */
    def findVcsRepositoryBranches(originalRepositoryId: VcsRepositoryId): Stream[F, (Username, VcsRepositoryName)]

    /** Search for the internal database specific (auto generated) ID of the given owner / repository combination which
      * serves as a primary key for the database table.
      *
      * @param owner
      *   Data about the owner of the repository containing information needed to query the database.
      * @param name
      *   The repository name which must be unique in regard to the owner.
      * @return
      *   An option to the internal database ID.
      */
    def findVcsRepositoryId(owner: VcsRepositoryOwner, name: VcsRepositoryName): F[Option[VcsRepositoryId]]

    /** Search for a repository owner of whom we only know the name.
      *
      * @param name
      *   The name of the repository owner which is the username of the actual owners account.
      * @return
      *   An option to successfully found repository owner.
      */
    def findVcsRepositoryOwner(name: Username): F[Option[VcsRepositoryOwner]]

    /** Search for a possible parent repository (i.e. a repository forked from) of the one described by the given name
      * and owner.
      *
      * @param owner
      *   Data about the owner of the repository containing information needed to query the database.
      * @param name
      *   The repository name which must be unique in regard to the owner.
      * @return
      *   An option to the successfully found parent vcs repository entry.
      */
    def findVcsRepositoryParentFork(owner: VcsRepositoryOwner, name: VcsRepositoryName): F[Option[VcsRepository]]

    /** Return a list of all repositories from all users.
      *
      * @param requester
      *   An optional account (None if guest i.e. not logged in user) for whom the list shall be created. This will
      *   affect which repositories will be returned in regard to access rights.
      * @param ordering
      *   The desired ordering of the list.
      * @return
      *   A stream of vcs repository entries in the requested order.
      */
    def listAllRepositories(requester: Option[Account])(
        ordering: VcsMetadataRepositoriesOrdering
    ): Stream[F, VcsRepository]

    /** Return a list of all repositories of the given owner.
      *
      * @param requester
      *   An optional account (None if guest i.e. not logged in user) for whom the list shall be created. This will
      *   affect which repositories will be returned in regard to access rights.
      * @param owner
      *   Data about the owner of the repositories containing information needed to query the database.
      * @return
      *   A stream of vcs repository entries ordered by name which may be empty.
      */
    def listRepositories(requester: Option[Account])(owner: VcsRepositoryOwner): Stream[F, VcsRepository]

    /** Update the database entry for the given vcs repository.
      *
      * @param repository
      *   The vcs repository metadata that shall be updated within the database.
      * @return
      *   The number of affected database rows.
      */
    def updateVcsRepository(repository: VcsRepository): F[Int]

}

/** Helper types to provide sorting instructions to several functions.
  */
enum VcsMetadataRepositoriesOrdering {
    case NameAscending  extends VcsMetadataRepositoriesOrdering
    case NameDescending extends VcsMetadataRepositoriesOrdering
}
