/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.*
import java.time.format.DateTimeFormatter

import cats.*
import cats.data.*
import cats.syntax.all.*
import de.smederee.email.EmailAddress
import de.smederee.security.UserId
import de.smederee.security.Username
import org.http4s.Uri
import org.slf4j.LoggerFactory

import scala.util.Try
import scala.util.matching.Regex

/** The types of version control software that we support.
  */
enum VcsType {

    /** The darcs DVCS.
      */
    case Darcs
}

opaque type VcsRepositoryDescription = String
object VcsRepositoryDescription {
    val MaximumLength: Int = 254

    /** Create an instance of VcsRepositoryDescription from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a VcsRepositoryDescription.
      * @return
      *   The appropriate instance of VcsRepositoryDescription.
      */
    def apply(source: String): VcsRepositoryDescription = source

    /** Try to create an instance of VcsRepositoryDescription from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a VcsRepositoryDescription.
      * @return
      *   An option to the successfully converted VcsRepositoryDescription.
      */
    def from(source: String): Option[VcsRepositoryDescription] = Option(source).map(_.take(MaximumLength))

}

opaque type VcsRepositoryId = Long
object VcsRepositoryId {
    given Eq[VcsRepositoryId] = Eq.fromUniversalEquals

    val Format: Regex = "^-?\\d+$".r

    /** Create an instance of VcsRepositoryId from the given Long type.
      *
      * @param source
      *   An instance of type Long which will be returned as a VcsRepositoryId.
      * @return
      *   The appropriate instance of VcsRepositoryId.
      */
    def apply(source: Long): VcsRepositoryId = source

    /** Try to create an instance of VcsRepositoryId from the given Long.
      *
      * @param source
      *   A Long that should fulfil the requirements to be converted into a VcsRepositoryId.
      * @return
      *   An option to the successfully converted VcsRepositoryId.
      */
    def from(source: Long): Option[VcsRepositoryId] = Option(source)

    /** Try to create an instance of VcsRepositoryId from the given String.
      *
      * @param source
      *   A string that should fulfil the requirements to be converted into a VcsRepositoryId.
      * @return
      *   An option to the successfully converted VcsRepositoryId.
      */
    def fromString(source: String): Option[VcsRepositoryId] =
        Option(source).filter(Format.matches).map(_.toLong).flatMap(from)

    extension (id: VcsRepositoryId) {
        def toLong: Long = id
    }
}

opaque type VcsRepositoryName = String
object VcsRepositoryName {

    given Eq[VcsRepositoryName] = Eq.fromUniversalEquals

    given Order[VcsRepositoryName] = Order.from((a, b) => a.toString.compareTo(b.toString))

    // TODO: Can we rewrite this in a Scala-3 way (i.e. without implicitly)?
    given Ordering[VcsRepositoryName] = implicitly[Order[VcsRepositoryName]].toOrdering

    val Format: Regex = "^[a-zA-Z0-9][a-zA-Z0-9\\-_]{1,63}$".r

    /** Create an instance of VcsRepositoryName from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a VcsRepositoryName.
      * @return
      *   The appropriate instance of VcsRepositoryName.
      */
    def apply(source: String): VcsRepositoryName = source

    /** Try to create an instance of VcsRepositoryName from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a VcsRepositoryName.
      * @return
      *   An option to the successfully converted VcsRepositoryName.
      */
    def from(source: String): Option[VcsRepositoryName] = validate(source).toOption

    /** Validate the given string and return either the validated repository name or a list of errors.
      *
      * @param s
      *   An arbitrary string which should be a repository name.
      * @return
      *   Either a list of errors or the validated repository name.
      */
    def validate(s: String): ValidatedNec[String, VcsRepositoryName] =
        Option(s).map(_.trim.nonEmpty) match {
            case Some(true) =>
                val input = s.trim
                val miniumLength =
                    if (input.length > 1)
                        input.validNec
                    else
                        "Repository name too short (min. 2 characters)!".invalidNec
                val maximumLength =
                    if (input.length < 65)
                        input.validNec
                    else
                        "Repository name too long (max. 64 characters)!".invalidNec
                val validFormat =
                    if (Format.matches(input))
                        input.validNec
                    else
                        "Repository name must start with a letter or number and is only allowed to contain alphanumeric characters plus .".invalidNec
                (miniumLength, maximumLength, validFormat).mapN { case (_, _, name) =>
                    name
                }
            case _ => "Repository name must not be empty!".invalidNec
        }
}

/** Extractor to retrieve a VcsRepositoryName from a path parameter.
  */
object VcsRepositoryNamePathParameter {
    def unapply(str: String): Option[VcsRepositoryName] = Option(str).flatMap(VcsRepositoryName.from)
}

/** Descriptive information about the owner of a vcs repository.
  *
  * @param owner
  *   The user ID of the account that owns the repository.
  * @param name
  *   The name of the account that owns the repository.
  * @param email
  *   The email address of the project owner.
  */
final case class VcsRepositoryOwner(uid: UserId, name: Username, email: EmailAddress)

object VcsRepositoryOwner {

    given Eq[VcsRepositoryOwner] = Eq.fromUniversalEquals

}

opaque type VcsPatchAuthor = String
object VcsPatchAuthor {
    val RemoveEmail: Regex = """<?([a-zA-Z0-9.!#$%&’'*+/=?^_`{|}~-]+)@([a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*)>?""".r

    /** Create an instance of VcsPatchAuthor from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a VcsPatchAuthor.
      * @return
      *   The appropriate instance of VcsPatchAuthor.
      */
    def apply(source: String): VcsPatchAuthor = source

    /** Try to create an instance of VcsPatchAuthor from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a VcsPatchAuthor.
      * @return
      *   An option to the successfully converted VcsPatchAuthor.
      */
    def from(source: String): Option[VcsPatchAuthor] = Option(source).filter(_.nonEmpty)

}

extension (author: VcsPatchAuthor) {

    /** Remove everything from the author string which could be an email address.
      *
      * @return
      *   A string with every occurence of a potential email address replaced by an empty string.
      */
    def withoutEmail: String = VcsPatchAuthor.RemoveEmail.replaceAllIn(author.toString, "")

}

opaque type VcsPatchComment = String
object VcsPatchComment {
    val FixesHashMarker: Regex = "\\bFixes: (?<hash>[a-f0-9]{40})\\b".r

    /** Create an instance of VcsPatchComment from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a VcsPatchComment.
      * @return
      *   The appropriate instance of VcsPatchComment.
      */
    def apply(source: String): VcsPatchComment = source

    /** Try to create an instance of VcsPatchComment from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a VcsPatchComment.
      * @return
      *   An option to the successfully converted VcsPatchComment.
      */
    def from(source: String): Option[VcsPatchComment] = Option(source).filter(_.nonEmpty)

    extension (comment: VcsPatchComment) {
        def length: Int = comment.length

        /** Find trailing `Fixes: <HASH>` lines in a patch comment as described in the contributing guidelines and
          * replace them with a link that is supposed to point where the patch details will be displayed.
          *
          * @param baseUri
          *   The base URI to which the actual hash value will be appended.
          * @return
          *   All lines of the comment string with every line containing the patch comment with all occurrences of
          *   `Fixes: <HASH>` replaced by `Fixes: <baseUri/HASH>`.
          */
        def withFixesHashUrl(baseUri: Uri): Array[String] = {
            def replace(m: Regex.Match): String =
                s"""Fixes: <a href="${baseUri.addSegment(m.group("hash")).toString}">${m.group("hash")}</a>"""
            val lines = comment.toString.split("\\r?\\n")
            lines.map(line => FixesHashMarker.replaceAllIn(line, replace))
        }
    }
}

opaque type VcsPatchFilename = String
object VcsPatchFilename {

    /** Create an instance of VcsPatchFilename from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a VcsPatchFilename.
      * @return
      *   The appropriate instance of VcsPatchFilename.
      */
    def apply(source: String): VcsPatchFilename = source

    /** Try to create an instance of VcsPatchFilename from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a VcsPatchFilename.
      * @return
      *   An option to the successfully converted VcsPatchFilename.
      */
    def from(source: String): Option[VcsPatchFilename] = Option(source).filter(_.nonEmpty)

}

opaque type VcsPatchHash = String
object VcsPatchHash {

    /** Create an instance of VcsPatchHash from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a VcsPatchHash.
      * @return
      *   The appropriate instance of VcsPatchHash.
      */
    def apply(source: String): VcsPatchHash = source

    /** Try to create an instance of VcsPatchHash from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a VcsPatchHash.
      * @return
      *   An option to the successfully converted VcsPatchHash.
      */
    def from(source: String): Option[VcsPatchHash] = Option(source).filter(_.nonEmpty)

}

opaque type VcsPatchName = String
object VcsPatchName {

    /** Create an instance of VcsPatchName from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a VcsPatchName.
      * @return
      *   The appropriate instance of VcsPatchName.
      */
    def apply(source: String): VcsPatchName = source

    /** Try to create an instance of VcsPatchName from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a VcsPatchName.
      * @return
      *   An option to the successfully converted VcsPatchName.
      */
    def from(source: String): Option[VcsPatchName] = Option(source).filter(_.nonEmpty)

}

/** A VcsPatchTimestamp is a OffsetDateTime with an `UTC` zone offset.
  */
opaque type VcsPatchTimestamp = OffsetDateTime
object VcsPatchTimestamp {

    /** Create an instance of VcsPatchTimestamp from the given OffsetDateTime type.
      *
      * @param source
      *   An instance of type OffsetDateTime which will be returned as a VcsPatchTimestamp.
      * @return
      *   The appropriate instance of VcsPatchTimestamp.
      */
    def apply(source: OffsetDateTime): VcsPatchTimestamp = source.withOffsetSameLocal(ZoneOffset.UTC)

    /** Try to create an instance of VcsPatchTimestamp from the given OffsetDateTime.
      *
      * @param source
      *   A OffsetDateTime that should fulfil the requirements to be converted into a VcsPatchTimestamp.
      * @return
      *   An option to the successfully converted VcsPatchTimestamp.
      */
    def from(source: OffsetDateTime): Option[VcsPatchTimestamp] =
        source.getOffset() match {
            case ZoneOffset.UTC => Option(source)
            case _              => None
        }

}

extension (timestamp: VcsPatchTimestamp) {

    /** Converts this date-time to an Instant. This returns an Instant representing the same point on the time-line as
      * this date-time.
      *
      * @return
      *   an Instant representing the same instant, not null
      */
    def toInstant: Instant = timestamp.toInstant

    def toOffsetDateTime: OffsetDateTime = timestamp
}

/** Summary construct for a file modification within a patch.
  *
  * @param added
  *   The number of lines added to the file which might be 0.
  * @param name
  *   The name of the file within the repository.
  * @param removed
  *   The number of lines removed from the file which might be 0.
  */
final case class VcsPatchSummaryFileModification(added: Int, name: VcsPatchFilename, removed: Int)

/** Summary of file operations within a patch.
  *
  * @param added
  *   A list of file names that are added within this patch.
  * @param modified
  *   A list of file modification summaries within this patch.
  * @param removed
  *   A list of file names that are removed within this patch.
  */
final case class VcsPatchSummary(
    added: List[VcsPatchFilename],
    modified: List[VcsPatchSummaryFileModification],
    removed: List[VcsPatchFilename]
)

object VcsPatchSummary {
    private val log = LoggerFactory.getLogger(classOf[VcsPatchSummary])

    /** Parse the given xml element which SHALL originate from the xml output of a darcs log command. Any exceptional
      * errors which occur during conversion are logged.
      *
      * @param darcsLogEntry
      *   A log entry from a darcs log using the xml output (i.e. the `<patch>` element).
      * @return
      *   An option to a successfully extracted patch metadata object.
      */
    def fromDarcsXmlLog(darcsLogEntry: scala.xml.Node): Option[VcsPatchSummary] =
        Try {
            val summary = (darcsLogEntry \ "summary").headOption
            val added = summary
                .map(summary => (summary \ "add_file").flatMap(file => VcsPatchFilename.from(file.text)).toList)
                .getOrElse(List.empty)
            val modified = summary
                .map(summary =>
                    (summary \ "modify_file").toList.flatMap { file =>
                        val filename      = VcsPatchFilename.from(file.text)
                        val added_lines   = file \ "added_lines" \@ "num"
                        val removed_lines = file \ "removed_lines" \@ "num"
                        val added =
                            if (added_lines.nonEmpty && added_lines.forall(_.isDigit))
                                added_lines.toInt
                            else
                                0
                        val removed =
                            if (removed_lines.nonEmpty && removed_lines.forall(_.isDigit))
                                removed_lines.toInt
                            else
                                0
                        filename.map(filename => VcsPatchSummaryFileModification(added, filename, removed))
                    }
                )
                .getOrElse(List.empty)
            val removed = summary
                .map(summary => (summary \ "remove_file").flatMap(file => VcsPatchFilename.from(file.text)).toList)
                .getOrElse(List.empty)
            summary.map(_ => VcsPatchSummary(added, modified, removed))
        } match {
            case scala.util.Failure(exception) =>
                log.error(
                    "Error occured while trying to extract patch summary from darcs log xml output!",
                    exception
                )
                None
            case scala.util.Success(patchSummary) => patchSummary
        }
}

/** Data describing a patch in a vcs repository.
  *
  * @param author
  *   The author of the patch which is usually a line of the format `Jane Doe <jane@example.com>`.
  * @param comment
  *   A comment (which should be a detailed description) for the patch.
  * @param hash
  *   A unique hash identifying the patch.
  * @param name
  *   The name (title) of the patch.
  * @param summary
  *   A summary of the file operations that were done with this patch.
  * @param timestamp
  *   The timestamp in UTC when the patch was recorded.
  */
final case class VcsRepositoryPatchMetadata(
    author: VcsPatchAuthor,
    comment: Option[VcsPatchComment],
    hash: VcsPatchHash,
    name: VcsPatchName,
    summary: Option[VcsPatchSummary],
    timestamp: VcsPatchTimestamp
)

object VcsRepositoryPatchMetadata {
    private val log = LoggerFactory.getLogger(classOf[VcsRepositoryPatchMetadata])

    private val replaceEmail = """<([a-zA-Z0-9.!#$%&’'*+/=?^_`{|}~-]+)@([a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*)>""".r

    val DarcsCommentFilter: Regex          = "^Ignore-this: [0-9a-f]+".r
    val DarcsDateFormat: DateTimeFormatter = DateTimeFormatter.ofPattern("yyyyMMddHHmmss")

    /** Parse the given xml element which SHALL originate from the xml output of a darcs log command. Any exceptional
      * errors which occur during conversion are logged.
      *
      * @param darcsLogEntry
      *   A log entry from a darcs log using the xml output (i.e. the `<patch>` element).
      * @return
      *   An option to a successfully extracted patch metadata object.
      */
    def fromDarcsXmlLog(darcsLogEntry: scala.xml.Node): Option[VcsRepositoryPatchMetadata] =
        Try {
            val author = VcsPatchAuthor.from(darcsLogEntry \@ "author")
            val comment = (darcsLogEntry \ "comment").headOption
                .map(comment => DarcsCommentFilter.replaceFirstIn(comment.text, "").trim)
                .flatMap(VcsPatchComment.from)
            val hash    = VcsPatchHash.from(darcsLogEntry \@ "hash")
            val name    = (darcsLogEntry \ "name").headOption.map(_.text).flatMap(VcsPatchName.from)
            val summary = VcsPatchSummary.fromDarcsXmlLog(darcsLogEntry)
            val timestamp =
                VcsPatchTimestamp.from(
                    LocalDateTime.parse(darcsLogEntry \@ "date", DarcsDateFormat).atOffset(ZoneOffset.UTC)
                )
            (author, hash, name, timestamp).mapN { case (author, hash, name, timestamp) =>
                VcsRepositoryPatchMetadata(author, comment, hash, name, summary, timestamp)
            }
        } match {
            case scala.util.Failure(exception) =>
                log.error(
                    "Error occured while trying to extract patch information from darcs log xml output!",
                    exception
                )
                None
            case scala.util.Success(patchMetadata) => patchMetadata
        }

    extension (patch: VcsRepositoryPatchMetadata) {

        /** Convert the patch metadata information into an RSS item that may be included in a RSS feed.
          *
          * @param baseUri
          *   The URI needed to access the patch on platform i.e. the address to which only the hash needs to be
          *   appended.
          * @return
          *   An RSS item describing the patch.
          */
        def toRssItem(baseUri: Uri): RssItem = {
            val uri = baseUri.addSegment(patch.hash.toString).toString
            RssItem(
                title = patch.name.toString,
                link = uri,
                description = patch.comment.map(_.toString).getOrElse(patch.name.toString).replaceAll("\n", "<br />"),
                author = replaceEmail.replaceAllIn(patch.author.toString, "").trim.some,
                guid = uri.some,
                pubDate = patch.timestamp.toOffsetDateTime.some
            )
        }
    }
}

/** Data about a VCS respository.
  *
  * @param name
  *   The name of the repository. A repository name must start with a letter or number and must contain only
  *   alphanumeric ASCII characters as well as minus or underscore signs. It must be between 2 and 64 characters long.
  * @param owner
  *   The owner of the repository.
  * @param isPrivate
  *   A flag indicating if this repository is private i.e. only visible / accessible for accounts with appropriate
  *   permissions.
  * @param description
  *   An optional short text description of the repository.
  * @param ticketsEnabled
  *   A flag indicating if ticket tracking support via the ticket service is enabled for this repository.
  * @param vcsType
  *   The type of the underlying DVCS that manages the repository.
  * @param website
  *   An optional uri pointing to a website related to the repository / project.
  */
final case class VcsRepository(
    name: VcsRepositoryName,
    owner: VcsRepositoryOwner,
    isPrivate: Boolean,
    description: Option[VcsRepositoryDescription],
    ticketsEnabled: Boolean,
    vcsType: VcsType,
    website: Option[Uri]
)

object VcsRepository {
    given Eq[VcsRepository] = Eq.fromUniversalEquals
}
