/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.io.IOException
import java.nio.file.*
import java.util.Locale

import cats.*
import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.darcs.*
import de.smederee.html.LinkTools.*
import de.smederee.html.MarkdownRenderer
import de.smederee.hub.RelatedTypesConverter.given
import de.smederee.hub.RequestHelpers.instances.given
import de.smederee.hub.config.*
import de.smederee.hub.forms.types.FormErrors
import de.smederee.i18n.LanguageCode
import de.smederee.security.CsrfToken
import de.smederee.security.Username
import de.smederee.tickets.ProjectRepository
import org.fusesource.jansi.utils.UtilsAnsiHtml
import org.http4s.*
import org.http4s.dsl.Http4sDsl
import org.http4s.dsl.impl.*
import org.http4s.headers.*
import org.http4s.implicits.*
import org.http4s.twirl.TwirlInstances.*
import org.slf4j.LoggerFactory
import org.typelevel.ci.CIStringSyntax

/** Routes for handling VCS repositories, including creation, management and public serving.
  *
  * @param configuration
  *   The hub service configuration.
  * @param darcs
  *   A class providing darcs VCS operations.
  * @param vcsMetadataRepo
  *   A repository for handling database operations regarding our vcs repositories and their metadata.
  * @param ticketsProjectRepo
  *   A repository for handling the synchronisation with the ticket service.
  * @param orgRepo
  *   A repository for handling database functionality for organisations.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class VcsRepositoryRoutes[F[_]: Async](
    configuration: ServiceConfig,
    darcs: DarcsCommands[F],
    vcsMetadataRepo: VcsMetadataRepository[F],
    ticketsProjectRepo: ProjectRepository[F],
    orgRepo: OrganisationRepository[F]
) extends Http4sDsl[F] {
    private val log                               = LoggerFactory.getLogger(getClass)
    given org.typelevel.log4cats.LoggerFactory[F] = org.typelevel.log4cats.slf4j.Slf4jFactory.create[F]

    private val MaximumFileSize     = configuration.renderMaximumFileSize
    private val createRepoPath      = uri"/repo/create"
    private val darcsConfig         = configuration.darcs
    private val linkConfig          = configuration.external
    private val linkToTicketService = if (configuration.tickets.enabled) configuration.tickets.baseUri.some else None
    private val sshConfig           = configuration.ssh

    // The base URI for our site which that be passed into some templates which create links themselfes.
    private val baseUri = linkConfig.createFullUri(Uri())

    /** Load the repository metadata with the given owner and name from the database and return it and its primary key
      * id if the repository exists and is readable by the given user account.
      *
      * @param currentUser
      *   The user account that is requesting access to the repository or None for a guest user.
      * @param repositoryOwnerName
      *   The name of the account that owns the repository.
      * @param repositoryName
      *   The name of the repository. A repository name must start with a letter or number and must contain only
      *   alphanumeric ASCII characters as well as minus or underscore signs. It must be between 2 and 64 characters
      *   long.
      * @return
      *   An option to a tuple holding the [[VcsRepository]] and its primary key id.
      */
    private def loadRepo(
        currentUser: Option[Account]
    )(repositoryOwnerName: Username, repositoryName: VcsRepositoryName): F[Option[(VcsRepository, VcsRepositoryId)]] =
        for {
            owner <- vcsMetadataRepo.findVcsRepositoryOwner(repositoryOwnerName)
            loadedRepo <- owner match {
                case None => Sync[F].pure(None)
                case Some(owner) =>
                    (
                        vcsMetadataRepo.findVcsRepository(owner, repositoryName),
                        vcsMetadataRepo.findVcsRepositoryId(owner, repositoryName)
                    ).mapN {
                        case (Some(repo), Some(repoId)) => (repo, repoId).some
                        case _                          => None
                    }
            }
            // TODO: Replace with whatever we implement as proper permission model. ;-)
            repoAndId = currentUser match {
                case None => loadedRepo.filter(tuple => tuple._1.isPrivate === false)
                case Some(user) =>
                    loadedRepo.filter(tuple =>
                        tuple._1.isPrivate === false || tuple._1.owner === user.toVcsRepositoryOwner
                    )
            }
        } yield repoAndId

    /** Delete the given directory recursively. It is checked if the given directory is a direct sub directory of the
      * owner's directory under `repositoriesDirectory` and only if this is the case is the directory removed.
      *
      * @param ownerName
      *   The name of the repository owner which is used for the directory check.
      * @param repoDirectory
      *   The path on the filesystem to the directory that shall be deleted.
      * @return
      *   `true` if the directory was deleted.
      */
    protected def deleteRepositoryDirectory(
        ownerName: Username
    )(repoDirectory: java.nio.file.Path): F[Boolean] =
        for {
            _ <- Sync[F].delay(log.debug(s"Request to delete repository dir: $repoDirectory"))
            reposDirPath <- Sync[F].delay(
                configuration.darcs.repositoriesDirectory.toPath.resolve(ownerName.toString)
            )
            isSubDir <- Sync[F].delay(reposDirPath.equals(repoDirectory.getParent()))
            deleted <-
                Sync[F].delay {
                    if (isSubDir) {
                        Files.walkFileTree(
                            repoDirectory,
                            new FileVisitor[java.nio.file.Path] {
                                override def visitFileFailed(
                                    file: java.nio.file.Path,
                                    exc: IOException
                                ): FileVisitResult = FileVisitResult.CONTINUE

                                override def visitFile(
                                    file: java.nio.file.Path,
                                    attrs: java.nio.file.attribute.BasicFileAttributes
                                ): FileVisitResult = {
                                    Files.delete(file)
                                    FileVisitResult.CONTINUE
                                }

                                override def preVisitDirectory(
                                    dir: java.nio.file.Path,
                                    attrs: java.nio.file.attribute.BasicFileAttributes
                                ): FileVisitResult = FileVisitResult.CONTINUE

                                override def postVisitDirectory(
                                    dir: java.nio.file.Path,
                                    exc: IOException
                                ): FileVisitResult = {
                                    Files.delete(dir)
                                    FileVisitResult.CONTINUE
                                }
                            }
                        )
                        Files.deleteIfExists(repoDirectory)
                    } else {
                        log.warn(
                            s"Refused requested removal of directory $repoDirectory which is not a direct sub directory of the configured repositories directory!"
                        )
                        false
                    }
                }
        } yield deleted

    /** Logic for creating a distribution file and serving it as a download to the user.
      *
      * @param user
      *   An optional user account for whom the list of repositories shall be rendered.
      * @param repositoryOwnerName
      *   The name of the user who owns the repository.
      * @param repositoryName
      *   The actual name of the repository.
      * @return
      *   An option to the file name that can be streamed to the user.
      */
    private def doDownloadDistribution(
        user: Option[Account]
    )(repositoryOwnerName: Username, repositoryName: VcsRepositoryName): F[Option[fs2.io.file.Path]] = {
        // TODO: Clean this up and refactor into proper helpers.

        /** Helper function to package a repository into a target directory because the darcs dist task has problems
          * with unicode characters in filenames. The archive will be created with a temporary name and afterwards it
          * will be moved to the given target directory and will be named `$repositoryName.tar.gz`.
          *
          * @param respositoryOwnerDirectory
          *   The directory in which the repository resides and in which the command will be executed.
          * @param repositoryName
          *   The name of the repository (and the name of the directory to be packaged).
          * @param targetDirectory
          *   The target (destination) directory into which the archive will be created.
          */
        def packageRepository(
            basePath: os.Path,
            repositoryName: String,
            targetDirectory: os.Path
        ): F[(Int, Chain[String], Chain[String])] = {
            val targetFile = targetDirectory / s"${repositoryName}.tar.gz"
            for {
                _ <- Sync[F].delay(Files.createDirectories(targetDirectory.toNIO))
                tempTargetFile <- Sync[F].delay(
                    Files.createTempFile(targetDirectory.toNIO, "download", s"${repositoryName}.tar.gz")
                )
                _ <- Sync[F].delay(log.debug(s"Packaging repository $repositoryName into $targetFile."))
                tarOptions      = List("-czf", tempTargetFile.toString, "--exclude=_darcs", repositoryName.toString)
                externalCommand = os.proc("tar", tarOptions)
                process <- Sync[F].delay(externalCommand.call(cwd = basePath, check = false))
                _       <- Sync[F].delay(os.remove(targetFile))
                _       <- Sync[F].delay(os.move(os.Path(tempTargetFile), targetFile))
            } yield (process.exitCode, Chain(process.out.text()), Chain(process.err.text()))
        }

        for {
            repoAndId <- loadRepo(user)(repositoryOwnerName, repositoryName)
            repo = repoAndId.map(_._1)
            repositoryOwnerDirectory <- Sync[F].delay(
                os.Path(
                    Paths.get(
                        darcsConfig.repositoriesDirectory.toPath.toString,
                        repositoryOwnerName.toString
                    )
                )
            )
            targetDirectory <- Sync[F].delay(
                os.Path(
                    Paths.get(
                        configuration.downloadDirectory.toPath.toString,
                        repositoryOwnerName.toString
                    )
                )
            )
            createDist <- repo.traverse(repo =>
                packageRepository(repositoryOwnerDirectory, repositoryName.toString, targetDirectory)
            )
            _ = createDist.map { case (exitCode, stdout, stderr) =>
                if (exitCode =!= 0)
                    log.error(
                        s"An error occured while packaging $repositoryName: ${stderr.toList.mkString}, ${stdout.toList.mkString}"
                    )
            }
            requestedFilePath <- repo.traverse(_ =>
                Sync[F].delay(
                    fs2.io.file.Path.fromNioPath((targetDirectory / s"${repositoryName.toString}.tar.gz").toNIO)
                )
            )
        } yield requestedFilePath
    }

    /** Logic for rendering a list of all repositories visible to the given user account.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of repositories shall be rendered.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    private def doShowAllRepositories(csrf: Option[CsrfToken])(user: Option[Account]): F[Response[F]] =
        for {
            language <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            repos <- vcsMetadataRepo
                .listAllRepositories(user)(VcsMetadataRepositoriesOrdering.NameAscending)
                .compile
                .toList
            actionBaseUri <- Sync[F].delay(linkConfig.createFullUri(uri"projects"))
            resp <- Ok(
                views.html.showAllRepositories(lang = language)(actionBaseUri, csrf, s"Smederee - Projects".some, user)(
                    repos
                )
            )
        } yield resp

    /** Logic for rendering a list of repositories of the given owner for a specific user account. This function takes
      * visibility into account.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param repositoriesOwnerName
      *   The name of a user whos repositories shall be listed.
      * @param user
      *   An optional user account for whom the list of repositories shall be rendered.
      * @return
      *   An HTTP response containing the rendered HTML or an error.
      */
    private def doShowRepositories(
        csrf: Option[CsrfToken]
    )(repositoriesOwnerName: Username)(user: Option[Account]): F[Response[F]] =
        for {
            owner <- vcsMetadataRepo.findVcsRepositoryOwner(repositoriesOwnerName)
            // FIXME: Load with proper permission check!
            organisation <- orgRepo
                .findByName(repositoriesOwnerName)
                .map(_.filter(o => o.isPrivate === false || user.exists(_.uid === o.owner)))
            language <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            loadRepos = vcsMetadataRepo.listRepositories(user)
            actionBaseUri <- Sync[F].delay(
                linkConfig.createFullUri(Uri(path = Uri.Path.unsafeFromString(s"~$repositoriesOwnerName")))
            )
            organisationActionBaseUri = organisation.map(org =>
                configuration.external.createFullUri(
                    uri"user/settings/organisations".addSegment(s"~${org.name.toString}")
                )
            )
            resp <- (owner, organisation) match {
                case (Some(owner), organisation) =>
                    loadRepos(owner).compile.toList.flatMap { repos =>
                        Ok(
                            views.html.showRepositories(lang = language)(
                                actionBaseUri,
                                csrf,
                                s"Smederee/~$repositoriesOwnerName".some,
                                user
                            )(repos, repositoriesOwnerName, organisation, organisationActionBaseUri)
                        )
                    }
                case (None, Some(organisation)) =>
                    Ok(
                        views.html.showRepositories(lang = language)(
                            actionBaseUri,
                            csrf,
                            s"Smederee/~$repositoriesOwnerName".some,
                            user
                        )(Nil, repositoriesOwnerName, organisation.some, organisationActionBaseUri)
                    )
                case _ =>
                    NotFound(
                        views.html.errors.userOrOrganisationNotFound(lang = language)(
                            csrf,
                            s"Smederee/~$repositoriesOwnerName".some,
                            user
                        )(repositoriesOwnerName)
                    )
            }
        } yield resp

    /** Logic for rendering an overview of the existing branches of a repository.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of repositories shall be rendered.
      * @param repositoryOwnerName
      *   The name of the user who owns the repository.
      * @param repositoryName
      *   The actual name of the repository.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    private def doShowRepositoryBranches(
        csrf: Option[CsrfToken]
    )(user: Option[Account])(repositoryOwnerName: Username, repositoryName: VcsRepositoryName): F[Response[F]] =
        for {
            language  <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            repoAndId <- loadRepo(user)(repositoryOwnerName, repositoryName)
            repo = repoAndId.map(_._1)
            actionBaseUri <- Sync[F].delay(
                linkConfig.createFullUri(
                    Uri(path =
                        Uri.Path(
                            Vector(
                                Uri.Path.Segment(s"~$repositoryOwnerName"),
                                Uri.Path.Segment(repositoryName.toString)
                            )
                        )
                    )
                )
            )
            branches <- repoAndId.map(_._2) match {
                case Some(repoId) => vcsMetadataRepo.findVcsRepositoryBranches(repoId).compile.toList
                case _            => Sync[F].delay(List.empty)
            }
            parentFork <- repo match {
                case None       => Sync[F].pure(None)
                case Some(repo) => vcsMetadataRepo.findVcsRepositoryParentFork(repo.owner, repo.name)
            }
            resp <- repo match {
                case None => NotFound("Repository not found!")
                case Some(repo) =>
                    Ok(
                        views.html.showRepositoryBranches(baseUri, lang = language)(
                            actionBaseUri,
                            csrf,
                            linkToTicketService,
                            s"Smederee/~$repositoryOwnerName/$repositoryName".some,
                            user
                        )(repo, branches)
                    )
            }
        } yield resp

    /** Logic for rendering the content of a repository directory or file visible to the given user account.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of repositories shall be rendered.
      * @param repositoryOwnerName
      *   The name of the user who owns the repository.
      * @param repositoryName
      *   The actual name of the repository.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    private def doShowRepositoryOverview(
        csrf: Option[CsrfToken]
    )(user: Option[Account])(repositoryOwnerName: Username, repositoryName: VcsRepositoryName): F[Response[F]] =
        for {
            language  <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            repoAndId <- loadRepo(user)(repositoryOwnerName, repositoryName)
            repo = repoAndId.map(_._1)
            actionBaseUri <- Sync[F].delay(
                linkConfig.createFullUri(
                    Uri(path =
                        Uri.Path(
                            Vector(
                                Uri.Path.Segment(s"~$repositoryOwnerName"),
                                Uri.Path.Segment(repositoryName.toString)
                            )
                        )
                    )
                )
            )
            sshUri <- sshConfig.enabled match {
                case false => Sync[F].pure(None)
                case true =>
                    Sync[F].delay(
                        s"SSH_PORT=${sshConfig.port.toString} darcs clone darcs@${sshConfig.host.toString}:${repositoryOwnerName.toString}/${repositoryName.toString}".some
                    )
            }
            directory <- Sync[F].delay(
                os.Path(
                    Paths.get(
                        darcsConfig.repositoriesDirectory.toPath.toString,
                        repositoryOwnerName.toString
                    )
                )
            )
            branches <- repoAndId.map(_._2) match {
                case Some(repoId) => vcsMetadataRepo.findVcsRepositoryBranches(repoId).compile.toList
                case _            => Sync[F].delay(List.empty)
            }
            vcsLog <- darcs.log(directory.toNIO)(repositoryName.toString)(Chain(s"--max-count=2", "--xml-output"))
            xmlLog <- Sync[F].delay(scala.xml.XML.loadString(vcsLog.stdout.toList.mkString))
            patches <- Sync[F].delay(
                (xmlLog \ "patch").flatMap(VcsRepositoryPatchMetadata.fromDarcsXmlLog).toList.map { patch =>
                    // TODO: Move the hard coded number into a configuration value.
                    if (patch.comment.exists(_.length > 320))
                        patch.copy(comment = patch.comment.map(c => VcsPatchComment(c.toString.take(320) + "...")))
                    else
                        patch
                }
            )
            readmeData <- repo.traverse(repo => doLoadReadme(repo))
            readme <- readmeData match {
                case Some((lines, Some(filename))) =>
                    if (filename.matches("(?iu).*\\.(md|markdown)$")) {
                        Sync[F]
                            .delay(
                                MarkdownRenderer
                                    .renderRepositoryMarkdownFile(repo.map(_.name.toString))(lines.mkString("\n"))
                            )
                            .map(_.some)
                    } else {
                        Sync[F].delay(lines.mkString("\n").some)
                    }
                case _ => Sync[F].delay(None)
            }
            readmeName = readmeData.flatMap(_._2)
            parentFork <- repo match {
                case None       => Sync[F].pure(None)
                case Some(repo) => vcsMetadataRepo.findVcsRepositoryParentFork(repo.owner, repo.name)
            }
            resp <- repo match {
                case None => NotFound("Repository not found!")
                case Some(repo) =>
                    Ok(
                        views.html.showRepositoryOverview(baseUri, lang = language)(
                            actionBaseUri,
                            csrf,
                            linkToTicketService,
                            s"Smederee/~$repositoryOwnerName/$repositoryName".some,
                            user
                        )(
                            repo,
                            vcsRepositoryBranches = branches,
                            vcsRepositoryHistory = patches,
                            vcsRepositoryParentFork = parentFork,
                            vcsRepositoryReadme = readme,
                            vcsRepositoryReadmeFilename = readmeName,
                            vcsRepositorySshUri = sshUri
                        )
                    )
            }
        } yield resp

    /** Logic for rendering the content of a repository directory or file visible to the given user account.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of repositories shall be rendered.
      * @param repositoryOwnerName
      *   The name of the user who owns the repository.
      * @param repositoryName
      *   The actual name of the repository.
      * @param filePath
      *   An URI path which describes the path the the requested part of the repository (empty or `/` for the root
      *   directory of the repo).
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    private def doShowRepositoryFiles(csrf: Option[CsrfToken])(
        user: Option[Account]
    )(repositoryOwnerName: Username, repositoryName: VcsRepositoryName)(filePath: Uri.Path): F[Response[F]] =
        for {
            language  <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            repoAndId <- loadRepo(user)(repositoryOwnerName, repositoryName)
            repo = repoAndId.map(_._1)
            requestedFilePath <- repo
                .traverse(_ =>
                    Sync[F].delay(
                        os.Path(
                            Paths.get(
                                darcsConfig.repositoriesDirectory.toPath.toString,
                                repositoryOwnerName.toString,
                                repositoryName.toString,
                                filePath.segments.map(_.decoded()).mkString("/")
                            )
                        )
                    )
                )
            fileType <- requestedFilePath
                .filter(os.isFile) // Only determine the mime type for regular files.
                .fold(Sync[F].pure(none[MediaType]))(path =>
                    // TODO: Refactor this into a dedicated function and clean it up.
                    Sync[F].delay(
                        MediaType
                            .parse(
                                os.proc("file", "--mime-type", path.toString)
                                    .call()
                                    .out
                                    .text()
                                    .trim
                                    .reverse
                                    .takeWhile(char => !char.isWhitespace)
                                    .reverse
                            )
                            .toOption
                    )
                )
            _        <- Sync[F].delay(log.debug(s"Requested to view file type: $fileType"))
            viewFile <- Sync[F].delay(fileType.exists(_.isText)) // Only display certain file types.
            listing <-
                requestedFilePath match {
                    case None                => Sync[F].delay(IndexedSeq.empty)
                    case Some(_) if viewFile => Sync[F].delay(IndexedSeq.empty)
                    case Some(path)          => doListFiles(path)
                }
            content <-
                requestedFilePath match {
                    case None => Sync[F].delay(IndexedSeq.empty)
                    case Some(path) =>
                        if (viewFile)
                            for {
                                size <- Sync[F].delay(os.size(path))
                                stream <-
                                    if (size <= MaximumFileSize)
                                        Sync[F].delay(os.read.lines.stream(path))
                                    else
                                        Sync[F].delay(os.Generator("File is too big!"))
                                lines <- Sync[F].delay(stream.toVector)
                            } yield lines
                        else if (os.isFile(path))
                            Sync[F].delay(Vector("Sorry, but displaying such file types is currently not implemented!"))
                        else
                            Sync[F].delay(IndexedSeq.empty)
                }
            repositoryBaseUri <- Sync[F].delay(
                linkConfig.createFullUri(
                    Uri(path =
                        Uri.Path(
                            Vector(
                                Uri.Path.Segment(s"~$repositoryOwnerName"),
                                Uri.Path.Segment(repositoryName.toString)
                            )
                        )
                    )
                )
            )
            actionBaseUri <- Sync[F].delay(
                linkConfig.createFullUri(Uri(path = repositoryBaseUri.path.addSegment("files") |+| filePath))
            )
            goBackUri <- Sync[F].delay(
                linkConfig.createFullUri(
                    Uri(path = Uri.Path(actionBaseUri.path.segments.reverse.drop(1).reverse))
                )
            )
            fileContent <- content.isEmpty match {
                case false =>
                    if (actionBaseUri.path.toString.toLowerCase(java.util.Locale.ROOT).endsWith(".md"))
                        Sync[F].delay(
                            List(MarkdownRenderer.renderRepositoryMarkdownFile(None)(content.mkString("\n")))
                        )
                    else
                        Sync[F].delay(content.toList)
                case true =>
                    Sync[F].pure(List.empty)
            }
            branches <- repoAndId.map(_._2) match {
                case Some(repoId) => vcsMetadataRepo.findVcsRepositoryBranches(repoId).compile.toList
                case _            => Sync[F].delay(List.empty)
            }
            title <- Sync[F].delay(
                s"Smederee/~$repositoryOwnerName/$repositoryName/${filePath.segments.map(_.decoded()).mkString("/")}"
            )
            resp <-
                repo match {
                    case None => NotFound("Repository not found!")
                    case Some(repo) =>
                        if (
                            filePath.segments.mkString
                                .startsWith("_darcs") || filePath.segments.mkString.startsWith("/_darcs")
                        )
                            NotFound("File not found!")
                        else
                            Ok(
                                views.html.showRepositoryFiles(baseUri, lang = language)(
                                    actionBaseUri,
                                    csrf,
                                    goBackUri.some,
                                    linkToTicketService,
                                    title.some,
                                    user
                                )(fileContent, listing, repositoryBaseUri, repo, branches)
                            )
                }
        } yield resp

    /** Logic for showing the history (changes) for the requested repository.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of repositories shall be rendered.
      * @param repositoryOwnerName
      *   The name of the user who owns the repository.
      * @param repositoryName
      *   The actual name of the repository.
      * @param fromEntry
      *   The optional number of the change from which the history shall be shown.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    def doShowRepositoryHistory(
        csrf: Option[CsrfToken]
    )(user: Option[Account])(repositoryOwnerName: Username, repositoryName: VcsRepositoryName)(
        fromEntry: Option[Int]
    ): F[Response[F]] =
        for {
            language  <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            repoAndId <- loadRepo(user)(repositoryOwnerName, repositoryName)
            repo = repoAndId.map(_._1)
            directory <- Sync[F].delay(
                os.Path(
                    Paths.get(
                        darcsConfig.repositoriesDirectory.toPath.toString,
                        repositoryOwnerName.toString
                    )
                )
            )
            branches <- repoAndId.map(_._2) match {
                case Some(repoId) => vcsMetadataRepo.findVcsRepositoryBranches(repoId).compile.toList
                case _            => Sync[F].delay(List.empty)
            }
            countChanges    <- darcs.log(directory.toNIO)(repositoryName.toString)(Chain("--count"))
            numberOfChanges <- Sync[F].delay(countChanges.stdout.toList.mkString.trim.toInt)
            maxCount = 10
            from     = fromEntry.getOrElse(1)
            to =
                if ((from + maxCount > numberOfChanges) && (numberOfChanges > 0))
                    numberOfChanges
                else
                    from + maxCount
            next = if (to < numberOfChanges) Option(to + 1) else None
            vcsLog <- darcs.log(directory.toNIO)(repositoryName.toString)(
                Chain(s"--index=$from-$to", "--summary", "--xml-output")
            )
            xmlLog  <- Sync[F].delay(scala.xml.XML.loadString(vcsLog.stdout.toList.mkString))
            patches <- Sync[F].delay((xmlLog \ "patch").flatMap(VcsRepositoryPatchMetadata.fromDarcsXmlLog).toList)
            actionBaseUri <- Sync[F].delay(
                linkConfig.createFullUri(
                    Uri(path =
                        Uri.Path(
                            Vector(
                                Uri.Path.Segment(s"~$repositoryOwnerName"),
                                Uri.Path.Segment(repositoryName.toString)
                            )
                        )
                    )
                )
            )
            goBackUri <- Sync[F].delay(
                linkConfig.createFullUri(
                    Uri(path =
                        Uri.Path(
                            Vector(
                                Uri.Path.Segment(s"~$repositoryOwnerName"),
                                Uri.Path.Segment(repositoryName.toString)
                            )
                        )
                    )
                )
            )
            repositoryBaseUri <- Sync[F].delay(
                linkConfig.createFullUri(
                    Uri(path =
                        Uri.Path(
                            Vector(
                                Uri.Path.Segment(s"~$repositoryOwnerName"),
                                Uri.Path.Segment(repositoryName.toString)
                            )
                        )
                    )
                )
            )
            resp <- repo match {
                case None => NotFound()
                case Some(repo) =>
                    Ok(
                        views.html.showRepositoryHistory(baseUri, lang = language)(
                            actionBaseUri,
                            csrf,
                            goBackUri.some,
                            linkToTicketService,
                            s"Smederee - History of ~$repositoryOwnerName/$repositoryName".some,
                            user
                        )(patches, next, repositoryBaseUri, repo, branches)
                    )
            }
        } yield resp

    /** Logic for generating the RSS feed for the requested repository.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of repositories shall be rendered.
      * @param repositoryOwnerName
      *   The name of the user who owns the repository.
      * @param repositoryName
      *   The actual name of the repository.
      * @return
      *   An HTTP response containing the rendered RSS XML.
      */
    def doShowRepositoryRssFeed(
        csrf: Option[CsrfToken]
    )(user: Option[Account])(repositoryOwnerName: Username, repositoryName: VcsRepositoryName): F[Response[F]] =
        for {
            language  <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            repoAndId <- loadRepo(user)(repositoryOwnerName, repositoryName)
            repo = repoAndId.map(_._1)
            response <- repo match {
                case None => NotFound()
                case Some(repo) =>
                    for {
                        directory <- Sync[F].delay(
                            os.Path(
                                Paths.get(
                                    darcsConfig.repositoriesDirectory.toPath.toString,
                                    repositoryOwnerName.toString
                                )
                            )
                        )
                        countChanges    <- darcs.log(directory.toNIO)(repositoryName.toString)(Chain("--count"))
                        numberOfChanges <- Sync[F].delay(countChanges.stdout.toList.mkString.trim.toInt)
                        maxCount = 15
                        from     = 1
                        to =
                            if ((from + maxCount > numberOfChanges) && (numberOfChanges > 0))
                                numberOfChanges
                            else
                                from + maxCount
                        next = if (to < numberOfChanges) Option(to + 1) else None
                        vcsLog <- darcs.log(directory.toNIO)(repositoryName.toString)(
                            Chain(s"--index=$from-$to", "--summary", "--xml-output")
                        )
                        xmlLog <- Sync[F].delay(scala.xml.XML.loadString(vcsLog.stdout.toList.mkString))
                        patches <- Sync[F].delay(
                            (xmlLog \ "patch").flatMap(VcsRepositoryPatchMetadata.fromDarcsXmlLog).toList
                        )
                        repositoryBaseUri <- Sync[F].delay(
                            linkConfig.createFullUri(
                                Uri(path =
                                    Uri.Path(
                                        Vector(
                                            Uri.Path.Segment(s"~$repositoryOwnerName"),
                                            Uri.Path.Segment(repositoryName.toString)
                                        )
                                    )
                                )
                            )
                        )
                        items = patches.map(_.toRssItem(repositoryBaseUri.addSegment("patch")))
                        channel = RssChannel(
                            title = s"Smederee: ~$repositoryOwnerName/$repositoryName",
                            link = repositoryBaseUri.toString,
                            description = repo.description
                                .map(_.toString)
                                .getOrElse(s"RSS feed for $repositoryOwnerName/$repositoryName"),
                            copyright = None,
                            image = None,
                            items = items
                        )
                        resp <- Ok(views.xml.rss(channel))
                    } yield resp
            }
        } yield response

    /** Get a diff for the requested patch and return the rendered response of it.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of repositories shall be rendered.
      * @param repositoryOwnerName
      *   The name of the user who owns the repository.
      * @param repositoryName
      *   The actual name of the repository.
      * @param hash
      *   The unique hash of the patch identifying it.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    def doShowRepositoryPatchDetails(csrf: Option[CsrfToken])(
        user: Option[Account]
    )(repositoryOwnerName: Username, repositoryName: VcsRepositoryName)(hash: DarcsHash): F[Response[F]] =
        for {
            language  <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            repoAndId <- loadRepo(user)(repositoryOwnerName, repositoryName)
            repo = repoAndId.map(_._1)
            directory <- Sync[F].delay(
                os.Path(
                    Paths.get(
                        darcsConfig.repositoriesDirectory.toPath.toString,
                        repositoryOwnerName.toString
                    )
                )
            )
            vcsLog <- darcs.log(directory.toNIO)(repositoryName.toString)(
                Chain(s"--hash=${hash.toString}", "--summary", "--xml-output")
            )
            xmlLog <- Sync[F].delay(scala.xml.XML.loadString(vcsLog.stdout.toList.mkString))
            patch <- Sync[F].delay(
                (xmlLog \ "patch").flatMap(VcsRepositoryPatchMetadata.fromDarcsXmlLog).toList.headOption
            )
            darcsDiff        <- darcs.diff(directory.toNIO)(repositoryName.toString)(hash)(Chain("--no-pause-for-gui"))
            patchCutMarker   <- Sync[F].delay(s"patch ${hash.toString}")
            cleanedPatchDiff <- Sync[F].delay(darcsDiff.stdout.toList.mkString.split(patchCutMarker)(0))
            patchDetails     <- Sync[F].delay(cleanedPatchDiff)
            htmlPatchDetails <- Sync[F].delay(new UtilsAnsiHtml().convertAnsiToHtml(patchDetails))
            cleanedHtmlPatchDetails <- Sync[F].delay(
                htmlPatchDetails.replaceAll("""style="color: white;"""", """style="color: #121212;"""")
            )
            branches <- repoAndId.map(_._2) match {
                case Some(repoId) => vcsMetadataRepo.findVcsRepositoryBranches(repoId).compile.toList
                case _            => Sync[F].delay(List.empty)
            }
            actionBaseUri <- Sync[F].delay(
                linkConfig.createFullUri(
                    Uri(path =
                        Uri.Path(
                            Vector(
                                Uri.Path.Segment(s"~$repositoryOwnerName"),
                                Uri.Path.Segment(repositoryName.toString)
                            )
                        )
                    )
                )
            )
            resp <- repo match {
                case None => NotFound()
                case Some(repo) =>
                    Ok(
                        views.html
                            .showRepositoryPatch(baseUri, lang = language)(
                                actionBaseUri,
                                csrf,
                                linkToTicketService,
                                patch.map(_.name.toString),
                                user
                            )(
                                patch,
                                cleanedHtmlPatchDetails,
                                repo,
                                branches
                            )
                    )
            }
        } yield resp

    /** List walk the given directory at the first level and return all found files and directories and their stats
      * sorted by directory first and by name second. If the given path is _not_ a directory then no traversal is done
      * and an empty list is returned.
      *
      * @param directory
      *   The path to the directory that shall be traversed.
      * @return
      *   A list of tuples containing a relative path and the related stats.
      */
    private def doListFiles(directory: os.Path): F[IndexedSeq[(os.RelPath, os.StatInfo)]] =
        for {
            isDirectory <- Sync[F].delay(os.isDir(directory))
            listing <-
                if (isDirectory)
                    Sync[F].delay(
                        os.walk
                            .attrs(directory, skip = (path, _) => path.baseName === "_darcs", maxDepth = 1)
                            .map((path, attrs) => (path.relativeTo(directory), attrs))
                            .sortWith { (left, right) =>
                                val (leftPath, leftAttrs)  = left
                                val (rightPath, rightAttr) = right
                                (leftAttrs.fileType, rightAttr.fileType) match {
                                    case (os.FileType.Dir, os.FileType.Dir) => leftPath.last < rightPath.last
                                    case (os.FileType.Dir, _)               => true
                                    case (_, os.FileType.Dir)               => false
                                    case (_, _)                             => leftPath.last < rightPath.last
                                }
                            }
                    )
                else
                    Sync[F].delay(IndexedSeq.empty)
        } yield listing

    /** Load the content of the first file matching our "readme file" heuristic which does not exceed the maximum file
      * size and return it.
      *
      * @param repo
      *   A repository in which we should search.
      * @return
      *   A tuple containing a list of strings (lines) that may be empty and an option to the file name.
      */
    private def doLoadReadme(repo: VcsRepository): F[(Vector[String], Option[String])] =
        for {
            path <- Sync[F].delay(
                Paths.get(
                    darcsConfig.repositoriesDirectory.toPath.toString,
                    repo.owner.name.toString,
                    repo.name.toString
                )
            )
            _     <- Sync[F].delay(log.debug(s"Trying to find README file in $path."))
            files <- Sync[F].delay(os.list(os.Path(path)))
            readme = files.find(_.last.matches("(?iu)^readme(\\..+)?$"))
            _    <- Sync[F].delay(log.debug(s"Found README at $readme."))
            size <- readme.traverse(path => Sync[F].delay(os.size(path)))
            stream <- readme.traverse { path =>
                if (size.getOrElse(0L) <= MaximumFileSize)
                    Sync[F].delay(os.read.lines.stream(path))
                else
                    Sync[F].delay(os.Generator(""))
            }
            lines <- Sync[F].delay(stream.map(_.toVector).getOrElse(Vector.empty))
        } yield (lines, readme.map(_.last))

    private val cloneRepository: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> UsernamePathParameter(repositoryOwnerName) /: VcsRepositoryClonePathParameter(
                repositoryName
            ) /: filePath =>
            for {
                csrf  <- Sync[F].delay(req.getCsrfToken)
                owner <- vcsMetadataRepo.findVcsRepositoryOwner(repositoryOwnerName)
                loadedRepo <- owner match {
                    case None        => Sync[F].pure(None)
                    case Some(owner) => vcsMetadataRepo.findVcsRepository(owner, repositoryName)
                }
                // Cloning via HTTP protocol is public only, so we don't need more sophisticated checks here.
                repo = loadedRepo.filter(r => r.isPrivate === false)
                requestedFilePath <- Sync[F].delay(
                    repo.map(_ =>
                        fs2.io.file.Path.fromNioPath(
                            Paths.get(
                                darcsConfig.repositoriesDirectory.toPath.toString,
                                repositoryOwnerName.toString,
                                repositoryName.toString,
                                filePath.segments.mkString("/")
                            )
                        )
                    )
                )
                _ <- Sync[F].delay(log.info(s"Repository $repo ($requestedFilePath) is cloned."))
                resp <- (repo, requestedFilePath) match {
                    case (Some(_), Some(path)) => StaticFile.fromPath(path, req.some).getOrElseF(NotFound())
                    case _                     => NotFound()
                }
            } yield resp
    }

    private val downloadDistribution: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "download" as user =>
            for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                file <- doDownloadDistribution(user.some)(repositoryOwnerName, repositoryName)
                resp <- file match {
                    case None => NotFound()
                    case Some(filePath) =>
                        StaticFile
                            .fromPath(filePath, ar.req.some)
                            .map(
                                _.putHeaders(
                                    `Content-Disposition`(
                                        "attachment",
                                        Map(ci"filename" -> s"${repositoryName}.tar.gz")
                                    )
                                )
                            )
                            .getOrElseF(NotFound())
                }
            } yield resp
    }

    private val downloadDistributionForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "download" =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                file <- doDownloadDistribution(None)(repositoryOwnerName, repositoryName)
                resp <- file match {
                    case None => NotFound()
                    case Some(filePath) =>
                        StaticFile
                            .fromPath(filePath, req.some)
                            .map(
                                _.putHeaders(
                                    `Content-Disposition`(
                                        "attachment",
                                        Map(ci"filename" -> s"${repositoryName}.tar.gz")
                                    )
                                )
                            )
                            .getOrElseF(NotFound())
                }
            } yield resp
    }

    private val forkRepository: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "fork" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    resp <- user.validatedEmail match {
                        case false =>
                            Forbidden(
                                views.html.errors
                                    .unvalidatedAccount(lang = language)(
                                        csrf,
                                        "Smederee - Account not validated!".some,
                                        user
                                    )
                            )
                        case true =>
                            for {
                                repoAndId <- loadRepo(user.some)(repositoryOwnerName, repositoryName)
                                sourceRepository = repoAndId.map(_._1)
                                // Check if a repository with that name already exists for the user.
                                loadedTargetRepo <- vcsMetadataRepo.findVcsRepository(
                                    user.toVcsRepositoryOwner,
                                    repositoryName
                                )
                                // If no repo exists we copy and adjust the source one, otherwise we return `None`.
                                targetRepository = loadedTargetRepo.fold(
                                    sourceRepository.map(_.copy(owner = user.toVcsRepositoryOwner))
                                )(_ => None)
                                targetUri <- Sync[F].delay(
                                    linkConfig.createFullUri(
                                        Uri(path =
                                            Uri.Path(
                                                Vector(
                                                    Uri.Path.Segment(s"~${user.name.toString}"),
                                                    Uri.Path.Segment(repositoryName.toString)
                                                )
                                            )
                                        )
                                    )
                                )
                                sourceDirectory <- Sync[F].delay(
                                    Paths.get(
                                        darcsConfig.repositoriesDirectory.toString,
                                        repositoryOwnerName.toString,
                                        repositoryName.toString
                                    )
                                )
                                targetDirectory <- Sync[F].delay(
                                    Paths.get(
                                        darcsConfig.repositoriesDirectory.toString,
                                        user.name.toString,
                                        repositoryName.toString
                                    )
                                )
                                output <- targetRepository match {
                                    case None => Sync[F].pure(0)
                                    case Some(repoMetadata) =>
                                        for {
                                            _ <- Sync[F].delay(
                                                Files.createDirectories(
                                                    Paths.get(
                                                        darcsConfig.repositoriesDirectory.toString,
                                                        repoMetadata.owner.name.toString
                                                    )
                                                )
                                            )
                                            darcsClone <- darcs.clone(sourceDirectory, targetDirectory)(
                                                Chain("--complete")
                                            )
                                            createRepo <-
                                                if (darcsClone.exitValue === 0)
                                                    for {
                                                        create <- vcsMetadataRepo.createVcsRepository(repoMetadata)
                                                        // Get IDs to create the fork entry in the database.
                                                        source <- sourceRepository.traverse(sourceRepo =>
                                                            vcsMetadataRepo.findVcsRepositoryId(
                                                                sourceRepo.owner,
                                                                sourceRepo.name
                                                            )
                                                        )
                                                        target <- vcsMetadataRepo.findVcsRepositoryId(
                                                            repoMetadata.owner,
                                                            repoMetadata.name
                                                        )
                                                        fork <- (source, target) match {
                                                            case (Some(Some(source)), Some(target)) =>
                                                                vcsMetadataRepo.createFork(source, target)
                                                            case _ => Sync[F].pure(0)
                                                        }
                                                    } yield create + fork
                                                else
                                                    Sync[F].pure(0) // Do not create DB entry if darcs init failed!
                                        } yield createRepo
                                }
                                resp <- SeeOther(Location(targetUri))
                            } yield resp
                    }
                } yield resp
            }
    }

    private val createRepository: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / "repo" / "create" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                val response = for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    _ <- Sync[F].raiseUnless(user.validatedEmail)(
                        new Error(
                            "An unvalidated account is not allowed to create a repository!"
                        ) // FIXME: Proper error handling!
                    )
                    formData <- Sync[F].delay {
                        urlForm.values.map { t =>
                            val (key, values) = t
                            (
                                key,
                                values.headOption.getOrElse("")
                            ) // Pick the first value (a field might get submitted multiple times)!
                        }
                    }
                    form <- Sync[F].delay(NewVcsRepositoryForm.validate(formData))
                    resp <- form match {
                        case Validated.Invalid(es) =>
                            BadRequest(
                                views.html
                                    .createRepository(lang = language)(
                                        createRepoPath,
                                        csrf,
                                        "Smederee - Create a new repository".some,
                                        user
                                    )(
                                        formData,
                                        FormErrors.fromNec(es)
                                    )
                            )
                        case Validated.Valid(newVcsRepositoryForm) =>
                            for {
                                directory <- Sync[F].delay(
                                    Paths.get(darcsConfig.repositoriesDirectory.toPath.toString, user.name.toString)
                                )
                                repoInDb <- vcsMetadataRepo.findVcsRepository(
                                    user.toVcsRepositoryOwner,
                                    newVcsRepositoryForm.name
                                )
                                output <- repoInDb match {
                                    case None =>
                                        for {
                                            _ <- Sync[F].delay {
                                                if (repoInDb.isEmpty && !Files.exists(directory)) {
                                                    log.debug(
                                                        s"User repository directory does not exist, trying to create it: $directory"
                                                    )
                                                    val _ = Files.createDirectories(directory)
                                                }
                                            }
                                            newRepo = VcsRepository(
                                                newVcsRepositoryForm.name,
                                                user.toVcsRepositoryOwner,
                                                newVcsRepositoryForm.isPrivate,
                                                newVcsRepositoryForm.description,
                                                newVcsRepositoryForm.ticketsEnabled,
                                                VcsType.Darcs,
                                                newVcsRepositoryForm.website
                                            )
                                            output <- darcs.initialize(directory)(newRepo.name.toString)(Chain.empty)
                                            _ <-
                                                if (output.exitValue === 0)
                                                    for {
                                                        written <- vcsMetadataRepo.createVcsRepository(newRepo)
                                                        _ <- Option(newRepo)
                                                            .filter(_.ticketsEnabled)
                                                            .traverse(repo =>
                                                                ticketsProjectRepo.createProject(repo.convert)
                                                            )
                                                    } yield written
                                                else
                                                    Sync[F].pure(0) // Do not create DB entry if darcs init failed!
                                        } yield output
                                    case Some(_) =>
                                        Sync[F].delay(
                                            DarcsCommandOutput(1, Chain.empty, Chain("The repository already exists!"))
                                        )
                                }
                                resp <- output.exitValue match {
                                    case 0 =>
                                        SeeOther(
                                            Location(
                                                Uri(path =
                                                    Uri.Path.Root |+| Uri.Path(
                                                        Vector(Uri.Path.Segment(s"~${user.name.toString}"))
                                                    )
                                                )
                                            )
                                        )
                                    case _ =>
                                        for {
                                            _ <- Sync[F].delay(
                                                log.error(
                                                    s"Error creating the repository ${newVcsRepositoryForm.name} in directory $directory: ${output.stderr.toList.mkString}"
                                                )
                                            )
                                            resp <- InternalServerError(
                                                s"Error creating the repository: ${output.stderr.toList.mkString}"
                                            )
                                        } yield resp
                                }
                            } yield resp
                    }
                } yield resp
                response.recoverWith { error =>
                    log.error("Internal Server Error", error)
                    for {
                        csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                        language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                        resp <- InternalServerError(
                            views.html.errors.internalServerError(lang = language)(csrf, user.some)
                        )
                    } yield resp
                }
            }
    }

    private val deleteRepository: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "delete" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf <- Sync[F].delay(ar.req.getCsrfToken)
                    _ <- Sync[F].raiseUnless(user.validatedEmail)(
                        new Error(
                            "An unvalidated account is not allowed to edit a repository!"
                        ) // FIXME: Proper error handling!
                    )
                    owner <- vcsMetadataRepo.findVcsRepositoryOwner(repositoryOwnerName)
                    loadedRepo <- owner match {
                        case None        => Sync[F].pure(None)
                        case Some(owner) => vcsMetadataRepo.findVcsRepository(owner, repositoryName)
                    }
                    redirectUri <- Sync[F].delay(
                        linkConfig.createFullUri(
                            Uri(path = Uri.Path(Vector(Uri.Path.Segment(s"~$repositoryOwnerName"))))
                        )
                    )
                    // You can only delete repositories that you own!
                    repo = loadedRepo.filter(_.owner === user.toVcsRepositoryOwner)
                    _ <- Sync[F].raiseUnless(repo.nonEmpty)(new Error("Repository not found!"))
                    repoDir <- Sync[F].delay(
                        repo.map(repo =>
                            darcsConfig.repositoriesDirectory.toPath
                                .resolve(user.name.toString)
                                .resolve(repo.name.toString)
                        )
                    )
                    _ <- repoDir.traverse(directory => deleteRepositoryDirectory(repositoryOwnerName)(directory))
                    _ <- repo.traverse(repo =>
                        vcsMetadataRepo.deleteVcsRepository(repo) *> ticketsProjectRepo.deleteProject(repo.convert)
                    )
                    resp <- SeeOther(Location(redirectUri))
                } yield resp
            }
    }

    private val editRepository: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "edit" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    _ <- Sync[F].raiseUnless(user.validatedEmail)(
                        new Error(
                            "An unvalidated account is not allowed to edit a repository!"
                        ) // FIXME: Proper error handling!
                    )
                    repoAndId <- loadRepo(user.some)(repositoryOwnerName, repositoryName)
                    repo = repoAndId.map(_._1)
                    branches <- repoAndId.map(_._2) match {
                        case Some(repoId) => vcsMetadataRepo.findVcsRepositoryBranches(repoId).compile.toList
                        case _            => Sync[F].delay(List.empty)
                    }
                    resp <- repo match {
                        case None => NotFound()
                        case Some(repo) =>
                            for {
                                editAction <- Sync[F].delay(
                                    linkConfig.createFullUri(
                                        Uri(path =
                                            Uri.Path(
                                                Vector(
                                                    Uri.Path.Segment(s"~$repositoryOwnerName"),
                                                    Uri.Path.Segment(repositoryName.toString),
                                                    Uri.Path.Segment("edit")
                                                )
                                            )
                                        )
                                    )
                                )
                                repoUri <- Sync[F].delay(
                                    linkConfig.createFullUri(
                                        Uri(path =
                                            Uri.Path(
                                                Vector(
                                                    Uri.Path.Segment(s"~$repositoryOwnerName"),
                                                    Uri.Path.Segment(repositoryName.toString)
                                                )
                                            )
                                        )
                                    )
                                )
                                formData <- Sync[F].delay {
                                    urlForm.values.map { t =>
                                        val (key, values) = t
                                        (
                                            key,
                                            values.headOption.getOrElse("")
                                        ) // Pick the first value (a field might get submitted multiple times)!
                                    }
                                }
                                form <- Sync[F].delay(EditVcsRepositoryForm.validate(formData))
                                resp <- form match {
                                    case Validated.Invalid(errors) =>
                                        BadRequest(
                                            views.html.editRepository(lang = language)(
                                                editAction,
                                                csrf,
                                                repoUri,
                                                Option(s"~$repositoryOwnerName/$repositoryName - edit"),
                                                user,
                                                repo,
                                                branches
                                            )(formData, FormErrors.fromNec(errors))
                                        )
                                    case Validated.Valid(updatedVcsRepositoryForm) =>
                                        val updatedRepo = repo.copy(
                                            isPrivate = updatedVcsRepositoryForm.isPrivate,
                                            description = updatedVcsRepositoryForm.description,
                                            ticketsEnabled = updatedVcsRepositoryForm.ticketsEnabled,
                                            website = updatedVcsRepositoryForm.website
                                        )
                                        // If the repo switched from tickets disabled to enabled, we create a ticket tracker.
                                        val createTicketTracker =
                                            updatedRepo.ticketsEnabled && updatedRepo.ticketsEnabled =!= repo.ticketsEnabled
                                        // If the repo switched from tickets enabled to disabled, we delete the ticket tracker.
                                        val deleteTicketTracker =
                                            !updatedRepo.ticketsEnabled && updatedRepo.ticketsEnabled =!= repo.ticketsEnabled
                                        for {
                                            _ <- vcsMetadataRepo.updateVcsRepository(updatedRepo)
                                            _ <-
                                                if (createTicketTracker)
                                                    ticketsProjectRepo.createProject(updatedRepo.convert)
                                                else
                                                    Sync[F].pure(0)
                                            _ <-
                                                if (deleteTicketTracker)
                                                    ticketsProjectRepo.deleteProject(updatedRepo.convert)
                                                else
                                                    Sync[F].pure(0)
                                            _ <-
                                                if (
                                                    !createTicketTracker && !deleteTicketTracker && updatedRepo.ticketsEnabled
                                                )
                                                    ticketsProjectRepo.updateProject(updatedRepo.convert)
                                                else
                                                    Sync[F].pure(0)
                                            resp <- SeeOther(Location(repoUri))
                                        } yield resp
                                }
                            } yield resp
                    }
                } yield resp
            }
    }

    private val showAllRepositories: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "projects" as user =>
            for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                resp <- doShowAllRepositories(csrf)(user.some)
            } yield resp
    }

    private val showAllRepositoriesForGuests: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root / "projects" =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            resp <- doShowAllRepositories(csrf)(None)
        } yield resp
    }

    private val showCreateRepositoryForm: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "repo" / "create" as user =>
            for {
                csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                resp <- user.validatedEmail match {
                    case false =>
                        Forbidden(
                            views.html.errors
                                .unvalidatedAccount(lang = language)(
                                    csrf,
                                    "Smederee - Account not validated!".some,
                                    user
                                )
                        )
                    case true =>
                        Ok(
                            views.html
                                .createRepository(lang = language)(
                                    createRepoPath,
                                    csrf,
                                    "Smederee - Create a new repository".some,
                                    user
                                )()
                        )
                }
            } yield resp
    }

    private val showDeleteRepositoryForm: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "delete" as user =>
            for {
                csrf      <- Sync[F].delay(ar.req.getCsrfToken)
                repoAndId <- loadRepo(user.some)(repositoryOwnerName, repositoryName)
                repo = repoAndId.map(_._1)
                branches <- repoAndId.map(_._2) match {
                    case Some(repoId) => vcsMetadataRepo.findVcsRepositoryBranches(repoId).compile.toList
                    case _            => Sync[F].delay(List.empty)
                }
                deleteAction <- Sync[F].delay(
                    linkConfig.createFullUri(
                        Uri(path =
                            Uri.Path(
                                Vector(
                                    Uri.Path.Segment(s"~$repositoryOwnerName"),
                                    Uri.Path.Segment(repositoryName.toString),
                                    Uri.Path.Segment("delete")
                                )
                            )
                        )
                    )
                )
                repoUri <- Sync[F].delay(
                    linkConfig.createFullUri(
                        Uri(path =
                            Uri.Path(
                                Vector(
                                    Uri.Path.Segment(s"~$repositoryOwnerName"),
                                    Uri.Path.Segment(repositoryName.toString)
                                )
                            )
                        )
                    )
                )
                resp <- repo match {
                    case None => NotFound()
                    case Some(repo) =>
                        Ok(
                            views.html.deleteRepository()(
                                deleteAction,
                                repoUri,
                                csrf,
                                Option(s"~$repositoryOwnerName/$repositoryName - delete"),
                                user,
                                repo,
                                branches
                            )
                        )
                }
            } yield resp
    }

    private val showEditRepositoryForm: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "edit" as user =>
            for {
                csrf      <- Sync[F].delay(ar.req.getCsrfToken)
                repoAndId <- loadRepo(user.some)(repositoryOwnerName, repositoryName)
                repo = repoAndId.map(_._1)
                branches <- repoAndId.map(_._2) match {
                    case Some(repoId) => vcsMetadataRepo.findVcsRepositoryBranches(repoId).compile.toList
                    case _            => Sync[F].delay(List.empty)
                }
                editAction <- Sync[F].delay(
                    linkConfig.createFullUri(
                        Uri(path =
                            Uri.Path(
                                Vector(
                                    Uri.Path.Segment(s"~$repositoryOwnerName"),
                                    Uri.Path.Segment(repositoryName.toString),
                                    Uri.Path.Segment("edit")
                                )
                            )
                        )
                    )
                )
                repoUri <- Sync[F].delay(
                    linkConfig.createFullUri(
                        Uri(path =
                            Uri.Path(
                                Vector(
                                    Uri.Path.Segment(s"~$repositoryOwnerName"),
                                    Uri.Path.Segment(repositoryName.toString)
                                )
                            )
                        )
                    )
                )
                formData <- Sync[F].delay(repo.map(EditVcsRepositoryForm.fromVcsRepository).map(_.toMap))
                resp <- (formData, repo) match {
                    case (Some(formData), Some(repo)) =>
                        Ok(
                            views.html.editRepository()(
                                editAction,
                                csrf,
                                repoUri,
                                Option(s"~$repositoryOwnerName/$repositoryName - edit"),
                                user,
                                repo,
                                branches
                            )(formData)
                        )
                    case _ => NotFound()
                }
            } yield resp
    }

    private val showRepositories: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(repositoriesOwnerName) as user =>
            for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                resp <- doShowRepositories(csrf)(repositoriesOwnerName)(user.some)
            } yield resp
    }

    private val showRepositoriesForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(repositoriesOwnerName) =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowRepositories(csrf)(repositoriesOwnerName)(None)
            } yield resp
    }

    private val showRepositoryBranches: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "branches" as user =>
            for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                resp <- doShowRepositoryBranches(csrf)(user.some)(repositoryOwnerName, repositoryName)
            } yield resp
    }

    private val showRepositoryBranchesForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "branches" =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowRepositoryBranches(csrf)(None)(repositoryOwnerName, repositoryName)
            } yield resp
    }

    private val showRepositoryOverview: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) as user =>
            for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                resp <- doShowRepositoryOverview(csrf)(user.some)(repositoryOwnerName, repositoryName)
            } yield resp
    }

    private val showRepositoryOverviewForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowRepositoryOverview(csrf)(None)(repositoryOwnerName, repositoryName)
            } yield resp
    }

    private val showRepositoryFiles: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> UsernamePathParameter(repositoryOwnerName) /: VcsRepositoryNamePathParameter(
                repositoryName
            ) /: "files" /: filePath as user =>
            for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                resp <- doShowRepositoryFiles(csrf)(user.some)(repositoryOwnerName, repositoryName)(filePath)
            } yield resp
    }

    private val showRepositoryFilesForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> UsernamePathParameter(repositoryOwnerName) /: VcsRepositoryNamePathParameter(
                repositoryName
            ) /: "files" /: filePath =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowRepositoryFiles(csrf)(None)(repositoryOwnerName, repositoryName)(filePath)
            } yield resp
    }

    private val showRepositoryHistory: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "history" :? HistoryFromQueryParameter(fromCount) as user =>
            for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                resp <- doShowRepositoryHistory(csrf)(user.some)(repositoryOwnerName, repositoryName)(fromCount)
            } yield resp
    }

    private val showRepositoryHistoryForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "history" :? HistoryFromQueryParameter(fromCount) =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowRepositoryHistory(csrf)(None)(repositoryOwnerName, repositoryName)(fromCount)
            } yield resp
    }

    private val showRepositoryRssFeed: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "feed.rss" =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowRepositoryRssFeed(csrf)(None)(repositoryOwnerName, repositoryName)
            } yield resp
    }

    private val showRepositoryPatchDetails: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "patch" / DarcsHashPathParameter(hash) as user =>
            for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                resp <- doShowRepositoryPatchDetails(csrf)(user.some)(repositoryOwnerName, repositoryName)(hash)
            } yield resp
    }

    private val showRepositoryPatchDetailsForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(repositoryOwnerName) / VcsRepositoryNamePathParameter(
                repositoryName
            ) / "patch" / DarcsHashPathParameter(hash) =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowRepositoryPatchDetails(csrf)(None)(repositoryOwnerName, repositoryName)(hash)
            } yield resp
    }

    val protectedRoutes =
        downloadDistribution <+> forkRepository <+> showAllRepositories <+> showRepositories <+>
            createRepository <+> deleteRepository <+> editRepository <+>
            showCreateRepositoryForm <+> showDeleteRepositoryForm <+> showEditRepositoryForm <+>
            showRepositoryOverview <+> showRepositoryBranches <+> showRepositoryHistory <+>
            showRepositoryPatchDetails <+> showRepositoryFiles

    val routes =
        cloneRepository <+> downloadDistributionForGuests <+> showAllRepositoriesForGuests <+>
            showRepositoriesForGuests <+> showRepositoryOverviewForGuests <+> showRepositoryBranchesForGuests <+>
            showRepositoryHistoryForGuests <+> showRepositoryPatchDetailsForGuests <+> showRepositoryFilesForGuests <+>
            showRepositoryRssFeed

}

/** A path parameter extractor to get the hash of a darcs patch for showing patch details.
  */
object DarcsHashPathParameter {
    def unapply(str: String): Option[DarcsHash] = Option(str).flatMap(DarcsHash.from)
}

/** Extractor for an optional query parameter we use in our history route.
  */
object HistoryFromQueryParameter extends OptionalQueryParamDecoderMatcher[Int]("from")

/** A path parameter extractor to get the vcs repository name for a clone operation.
  */
object VcsRepositoryClonePathParameter {
    def unapply(str: String): Option[VcsRepositoryName] =
        Option(str).flatMap { string =>
            if (string.endsWith(".darcs"))
                VcsRepositoryName.from(string.reverse.drop(6).reverse)
            else
                None
        }
}
