/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.security.SecureRandom
import java.util.Base64

import cats.Eq
import cats.syntax.all.*

import scala.util.matching.Regex

opaque type ErrorMessageKey = String
object ErrorMessageKey {

    /** Create an instance of ErrorMessageKey from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a ErrorMessageKey.
      * @return
      *   The appropriate instance of ErrorMessageKey.
      */
    def apply(source: String): ErrorMessageKey = source

    /** Try to create an instance of ErrorMessageKey from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a ErrorMessageKey.
      * @return
      *   An option to the successfully converted ErrorMessageKey.
      */
    def from(source: String): Option[ErrorMessageKey] =
        if (Option(source).map(_.startsWith("error.")).getOrElse(false))
            Option(source)
        else
            None

}

opaque type SessionId = String
object SessionId {
    val Format: Regex = "^([A-Za-z0-9+/\\-\\_]{4})+$".r
    val Length: Int   = 24 // 24 Bytes => 192 Bits

    given Eq[SessionId] = Eq.fromUniversalEquals

    private val base64Decoder   = Base64.getUrlDecoder
    private val base64Encoder   = Base64.getUrlEncoder.withoutPadding
    private val randomGenerator = new SecureRandom()

    /** Create an instance of SessionId from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a SessionId.
      * @return
      *   The appropriate instance of SessionId.
      */
    def apply(source: String): SessionId = source

    /** Try to create an instance of SessionId from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a SessionId.
      * @return
      *   An option to the successfully converted SessionId.
      */
    def from(source: String): Option[SessionId] =
        Option(source).filter(Format.matches).map(base64Decoder.decode) match {
            case Some(bytes) =>
                if (bytes.length === Length)
                    Option(source)
                else
                    None
            case _ => None
        }

    /** Generate a new globally unique session id using a secure random generator.
      *
      * @return
      *   A string containing a base 64 encoded url safe version of the generated random bytes.
      */
    def generate: SessionId = {
        var buffer = new Array[Byte](Length) // scalafix:ok; Mutability is needed here for the random generator.
        randomGenerator.nextBytes(buffer)
        base64Encoder.encodeToString(buffer)
    }
}
