/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.ssh

import java.io.InputStream
import java.io.OutputStream
import java.nio.file.Paths

import cats.*
import cats.effect.*
import cats.effect.std.Dispatcher
import cats.effect.unsafe.implicits.global
import cats.syntax.all.*
import de.smederee.hub.VcsRepositoryName
import de.smederee.hub.config.*
import de.smederee.security.UserId
import de.smederee.security.Username
import org.apache.sshd.scp.common.ScpHelper
import org.apache.sshd.scp.server.*
import org.apache.sshd.server.Environment
import org.apache.sshd.server.ExitCallback
import org.apache.sshd.server.channel.ChannelSession
import org.apache.sshd.server.command.Command
import org.apache.sshd.server.command.CommandFactory
import org.apache.sshd.server.session.ServerSession
import org.apache.sshd.server.session.ServerSessionAware
import org.apache.sshd.server.shell.UnknownCommand
import org.slf4j.LoggerFactory

import scala.util.matching.Regex

/** A base class for the specific darcs commands that we implement.
  */
@SuppressWarnings(Array("scalafix:DisableSyntax.null", "scalafix:DisableSyntax.var"))
abstract class DarcsSshCommand extends Command with ServerSessionAware {
    private val log = LoggerFactory.getLogger(classOf[DarcsSshCommand])

    @volatile protected var session: ServerSession = null
    @volatile protected var stdin: InputStream     = null
    @volatile protected var stdout: OutputStream   = null
    @volatile protected var stderr: OutputStream   = null
    @volatile protected var callback: ExitCallback = null

    /** This method SHALL implement the custom logic for the ssh command.
      *
      * The `start` method MUST execute this within a [[java.lang.Thread]]!
      *
      * {{{
      * val thread = new Thread {
      *   override def run(): Unit = {
      *     val result = commandLogic(cmd)
      *     callback.onExit(result.exitCode)
      *   }
      * }
      * thread.start()
      * }}}
      *
      * @param cmd
      *   A command that will be executed.
      * @return
      *   The result of the given command.
      * @todo
      *   De-couple this from the os-lib sometime in the future.
      */
    protected def commandLogic(cmd: os.proc): os.CommandResult

    override def destroy(channel: ChannelSession): Unit = {
        log.debug("destroy: {}", channel)
        ()
    }
    override def setErrorStream(errorStream: OutputStream): Unit = {
        log.debug("setErrorStream")
        this.stderr = errorStream
    }
    override def setExitCallback(callback: ExitCallback): Unit = {
        log.debug("setExitCallback: {}", callback)
        this.callback = callback
    }
    override def setInputStream(inputStream: InputStream): Unit = {
        log.debug("setInputStream")
        this.stdin = inputStream
    }
    override def setOutputStream(outputStream: OutputStream): Unit = {
        log.debug("setOutputStream")
        this.stdout = outputStream
    }
    override def setSession(session: ServerSession): Unit = {
        log.debug("setSession: {}", session)
        this.session = session
    }

}

/** Implementation of a ssh command to perform a darcs apply command.
  *
  * @param darcsConfiguration
  *   The configuration needed to properly execute underlying darcs commands and access repository data on the
  *   filesystem.
  * @param owner
  *   The username of the owner of the requested repository.
  * @param repository
  *   The name of the vcs repository.
  * @param useDebugFlag
  *   A flag indicating if we should use the `--debug` flag.
  */
final class DarcsApply(
    darcsConfiguration: DarcsConfiguration,
    owner: Username,
    repository: VcsRepositoryName,
    useDebugFlag: Boolean
) extends DarcsSshCommand {
    private val log = LoggerFactory.getLogger(classOf[DarcsApply])

    override protected def commandLogic(cmd: os.proc): os.CommandResult = {
        val workingDirectory = darcsConfiguration.repositoriesDirectory.toPath
        log.debug(s"DarcsApply: calling command: $cmd in directory $workingDirectory.")
        cmd.call(
            cwd = os.Path(workingDirectory),
            stdin = this.stdin,
            stdout = os.ProcessOutput { (bytes, _) =>
                log.trace("DarcsApply: STDOUT: {}", new String(bytes, "UTF-8"))
                this.stdout.write(bytes)
            },
            stderr = os.ProcessOutput { (bytes, _) =>
                log.trace("DarcsApply: STDERR: {}", new String(bytes, "UTF-8"))
                this.stderr.write(bytes)
            }
        )
    }

    override def start(channel: ChannelSession, env: Environment): Unit = {
        log.debug(s"DarcsApply for $owner/$repository")
        val repoDir = Paths.get(owner.toString, repository.toString)
        val options =
            if (useDebugFlag)
                List("apply", "--all", "--debug", "--repodir", repoDir.toString)
            else
                List("apply", "--all", "--repodir", repoDir.toString)
        val cmd = os.proc(darcsConfiguration.executable.toString, options)
        val thread = new Thread {
            override def run(): Unit = {
                val result = commandLogic(cmd)
                callback.onExit(result.exitCode)
            }
        }
        thread.start()
    }
}

/** Implementation of a ssh command to perform a darcs transfer-mode command.
  *
  * @param darcsConfiguration
  *   The configuration needed to properly execute underlying darcs commands and access repository data on the
  *   filesystem.
  * @param owner
  *   The username of the owner of the requested repository.
  * @param repository
  *   The name of the vcs repository.
  */
final class DarcsTransferMode(
    darcsConfiguration: DarcsConfiguration,
    owner: Username,
    repository: VcsRepositoryName
) extends DarcsSshCommand {
    private val log = LoggerFactory.getLogger(classOf[DarcsTransferMode])

    override protected def commandLogic(cmd: os.proc): os.CommandResult = {
        val workingDirectory = darcsConfiguration.repositoriesDirectory.toPath
        log.debug(s"DarcsTransferMode: calling command: $cmd in directory $workingDirectory.")
        cmd.call(
            cwd = os.Path(workingDirectory),
            stdin = this.stdin,
            stdout = os.ProcessOutput { (bytes, _) =>
                log.trace("DarcsTransferMode: STDOUT: {}", new String(bytes, "UTF-8"))
                this.stdout.write(bytes)
            },
            stderr = os.ProcessOutput { (bytes, _) =>
                log.trace("DarcsTransferMode: STDERR: {}", new String(bytes, "UTF-8"))
                this.stderr.write(bytes)
            }
        )
    }

    override def start(channel: ChannelSession, env: Environment): Unit = {
        log.debug(s"DarcsTransferMode for $owner/$repository")
        val repoDir = Paths.get(owner.toString, repository.toString)
        val cmd = os.proc(
            darcsConfiguration.executable.toString,
            List("transfer-mode", "--repodir", repoDir.toString)
        )
        val thread = new Thread {
            override def run(): Unit = {
                val result = commandLogic(cmd)
                callback.onExit(result.exitCode)
            }
        }
        thread.start()
    }
}

/** The command factory is appended to the apache ssh server and responsible for parsing requested commands and
  * delegating to the appropriate specific command implementation.
  *
  * @param darcsConfiguration
  *   The configuration needed to properly execute underlying darcs commands and access repository data on the
  *   filesystem.
  * @param repository
  *   A repository providing the needed functionality like getting ssh keys and user/repo information.
  */
final class DarcsSshCommandFactory(darcsConfiguration: DarcsConfiguration, repository: SshAuthenticationRepository[IO])
    extends CommandFactory {
    private val log = LoggerFactory.getLogger(classOf[DarcsSshCommandFactory])

    /** Check if the given repository of the given owner is readable by the user with the given id.
      *
      * @param ownerName
      *   The unique name (user name) of the owner of the repository.
      * @param repoName
      *   The name of the repository which is unique within the context of the owner.
      * @param userId
      *   The unique id of the account that is requesting access.
      * @return
      *   Either `true` if the repository is readable by the user or `false` otherwise.
      */
    protected def repositoryIsReadableBy(ownerName: Username, repoName: VcsRepositoryName, userId: UserId): Boolean =
        Dispatcher
            .sequential[IO]
            .use { dispatcher =>
                val checkPermissions =
                    for {
                        _ <- IO.delay(
                            log.debug(s"Checking if vcs repository $ownerName/$repoName is readable by $userId.")
                        )
                        vcsOwner <- repository.findVcsRepositoryOwner(ownerName)
                        _        <- IO.delay(log.debug(s"VCS repository owner name maps to $vcsOwner."))
                        userIsOwner = vcsOwner.exists(_.uid === userId)
                    } yield userIsOwner
                checkPermissions.recoverWith { error =>
                    log.error("Internal Server Error", error)
                    false.pure[IO]
                }
            }
            .unsafeRunSync()

    @SuppressWarnings(Array("scalafix:DisableSyntax.null"))
    override def createCommand(channel: ChannelSession, command: String): Command = {
        log.debug(s"Requested SSH command: $command")
        val sshKeyOwnerId = channel.getSession().getAttribute(SshServerConfiguration.SshKeyOwnerIdAttribute)

        command match {
            case DarcsSshCommandFactory.FilterDarcsApplyCommand(
                    "apply",
                    _,
                    _,
                    _,
                    debugFlag,
                    "--repodir",
                    owner,
                    repository
                ) =>
                (
                    SshUsername.from(owner),
                    VcsRepositoryName.from(repository)
                )
                    .mapN { case (owner, repository) =>
                        if (repositoryIsReadableBy(owner.toUsername, repository, sshKeyOwnerId))
                            new DarcsApply(
                                darcsConfiguration,
                                owner.toUsername,
                                repository,
                                debugFlag === "--debug"
                            )
                        else
                            new UnknownCommand("You are only allowed to access your own repositories!")
                    }
                    .getOrElse(new UnknownCommand(command))
            case DarcsSshCommandFactory.FilterDarcsTransferModeCommand(
                    "transfer-mode",
                    "--repodir",
                    owner,
                    repository
                ) =>
                (
                    SshUsername.from(owner),
                    VcsRepositoryName.from(repository)
                )
                    .mapN { case (owner, repository) =>
                        if (repositoryIsReadableBy(owner.toUsername, repository, sshKeyOwnerId))
                            // new DarcsTransferMode(darcsConfiguration, owner.toUsername, repository)
                            new UnknownCommand(command) // FIXME: Make transfer-mode work (stalls currently).
                        else
                            new UnknownCommand("You are only allowed to access your own repositories!")
                    }
                    .getOrElse(new UnknownCommand(command))
            case DarcsSshCommandFactory.FilterScpCommand("-f", _, "--", owner, repository, path) =>
                (
                    SshUsername.from(owner),
                    VcsRepositoryName.from(repository)
                )
                    .mapN { case (owner, repository) =>
                        if (repositoryIsReadableBy(owner.toUsername, repository, sshKeyOwnerId) && !path.contains(".."))
                            new ScpCommand(
                                channel,
                                command,
                                null,
                                ScpHelper.DEFAULT_SEND_BUFFER_SIZE,
                                ScpHelper.DEFAULT_RECEIVE_BUFFER_SIZE,
                                null,
                                null
                            )
                        else
                            new UnknownCommand("You are only allowed to access your own repositories!")
                    }
                    .getOrElse(new UnknownCommand(command))
            case DarcsSshCommandFactory.FilterScpCommand("-f", null, null, owner, repository, path) =>
                (
                    SshUsername.from(owner),
                    VcsRepositoryName.from(repository)
                )
                    .mapN { case (owner, repository) =>
                        if (repositoryIsReadableBy(owner.toUsername, repository, sshKeyOwnerId) && !path.contains(".."))
                            new ScpCommand(
                                channel,
                                command,
                                null,
                                ScpHelper.DEFAULT_SEND_BUFFER_SIZE,
                                ScpHelper.DEFAULT_RECEIVE_BUFFER_SIZE,
                                null,
                                null
                            )
                        else
                            new UnknownCommand("You are only allowed to access your own repositories!")
                    }
                    .getOrElse(new UnknownCommand(command))
            case _ => new UnknownCommand(command)
        }
    }
}

object DarcsSshCommandFactory {
    // A regular expression which should match on our allowed darcs commands and their arguments.
    val FilterDarcsApplyCommand: Regex =
        "^darcs (apply) ((--all)\\s)?((--debug)\\s)?\\s*(--repodir) '([a-z][a-z0-9]{1,31})/([a-zA-Z0-9][a-zA-Z0-9\\-_]{1,63})'$".r
    val FilterDarcsTransferModeCommand: Regex =
        "^darcs (transfer-mode) (--repodir) ([a-z][a-z0-9]{1,31})/([a-zA-Z0-9][a-zA-Z0-9\\-_]{1,63})/?$".r
    // After the initial transfer-mode call darcs will fallback to using scp.
    // val FilterScpCommand: Regex = "^scp (-f) ((--)\\s)?([^\u0000]+)$".r
    val FilterScpCommand: Regex =
        "^scp (-f) ((--)\\s)?([a-z][a-z0-9]{1,31})/([a-zA-Z0-9][a-zA-Z0-9\\-_]{1,63})/([^\u0000]+)$".r
    // TODO: Add support for sftp because that could be enforced from the client side.
}
