/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.ssh

import java.security.MessageDigest
import java.time.OffsetDateTime
import java.util.Base64
import java.util.UUID

import cats.*
import cats.syntax.all.*
import de.smederee.security.UserId
import org.apache.sshd.common.config.keys.AuthorizedKeyEntry
import org.bouncycastle.crypto.util.OpenSSHPublicKeyUtil

import scala.util.Try
import scala.util.matching.Regex

opaque type EncodedKeyBytes = String
object EncodedKeyBytes {
    val Format: Regex = "^[a-zA-z0-9+/]+=*$".r

    /** Create an instance of EncodedKeyBytes from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a EncodedKeyBytes.
      * @return
      *   The appropriate instance of EncodedKeyBytes.
      */
    def apply(source: String): EncodedKeyBytes = source

    /** Try to create an instance of EncodedKeyBytes from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a EncodedKeyBytes.
      * @return
      *   An option to the successfully converted EncodedKeyBytes.
      */
    def from(source: String): Option[EncodedKeyBytes] = Option(source).filter(string => Format.matches(string))

    /** An unsafe method to create an instance of EncodedKeyBytes from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a EncodedKeyBytes.
      * @return
      *   The converted EncodedKeyBytes instance.
      */
    @SuppressWarnings(Array("scalafix:DisableSyntax.throw"))
    @throws[IllegalArgumentException]("if the given string is not syntactically valid")
    def unsafeFrom(source: String): EncodedKeyBytes =
        from(source) match {
            case None           => throw new IllegalArgumentException(s"Illegal format for EncodedKeyBytes: $source")
            case Some(keyBytes) => keyBytes
        }

    extension (keyBytes: EncodedKeyBytes) {

        /** Convert the key bytes into an array of bytes by decoding the base64 string.
          *
          * @return
          *   The decoded bytes from the actual key bytes that were base64 encoded.
          */
        def toByteArray: Array[Byte] = Base64.getDecoder().decode(keyBytes)
    }

}

opaque type KeyComment = String
object KeyComment {
    val MaxLength: Int = 256

    /** Create an instance of KeyComment from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a KeyComment.
      * @return
      *   The appropriate instance of KeyComment.
      */
    def apply(source: String): KeyComment = source

    /** Try to create an instance of KeyComment from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a KeyComment.
      * @return
      *   An option to the successfully converted KeyComment.
      */
    def from(source: String): Option[KeyComment] = Option(source).map(_.take(MaxLength)).filter(_.nonEmpty)

}

opaque type KeyFingerprint = String
object KeyFingerprint {

    /** Create an instance of KeyFingerprint from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a KeyFingerprint.
      * @return
      *   The appropriate instance of KeyFingerprint.
      */
    def apply(source: String): KeyFingerprint = source

    /** Try to create an instance of KeyFingerprint from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a KeyFingerprint.
      * @return
      *   An option to the successfully converted KeyFingerprint.
      */
    def from(source: String): Option[KeyFingerprint] = Option(source).filter(_.nonEmpty)

}

opaque type SshPublicKeyString = String
object SshPublicKeyString {
    val Format: Regex = "^([\\w-]+)\\s(.+)(\\s(.+))?$".r

    /** Create an instance of SshPublicKeyString from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a SshPublicKeyString.
      * @return
      *   The appropriate instance of SshPublicKeyString.
      */
    def apply(source: String): SshPublicKeyString = source

    /** Try to create an instance of SshPublicKeyString from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a SshPublicKeyString.
      * @return
      *   An option to the successfully converted SshPublicKeyString.
      */
    def from(source: String): Option[SshPublicKeyString] = Option(source).filter(string => Format.matches(string))
}

/** Possible ssh key types.
  *
  * @param identifier
  *   A unique identifier (the code written into openssh key files).
  */
enum SshKeyType(val identifier: String) {
    case SshDsa extends SshKeyType("ssh-dss")

    case SshEcDsa extends SshKeyType("ecdsa-sha2-nistp256")

    // case SshEcDsaSk

    /** Recommended key type because smaller payload and more secure but some servers still don't support them.
      */
    case SshEd25519 extends SshKeyType("ssh-ed25519")

    // case SshEd25519Sk

    /** RSA keys are (still) the most common ones and the most compatible.
      */
    case SshRsa extends SshKeyType("ssh-rsa")
}

object SshKeyType {
    val Mappings: Map[String, SshKeyType] = Map(
        SshDsa.identifier     -> SshDsa,
        SshEcDsa.identifier   -> SshEcDsa,
        SshEd25519.identifier -> SshEd25519,
        SshRsa.identifier     -> SshRsa
    )

    given Eq[SshKeyType] = Eq.fromUniversalEquals

    /** Try to extract a supported [[SshKeyType]] from the given string that should contain a valid ssh public key.
      * According to RFC 4253 the key format is as follows:
      * {{{
      * [type-name] [base-64-encoded-public-key] [comment]
      * }}}
      * While the comment is optional the other parts are required.
      *
      * @param sshKeyString
      *   A string containing a correctly formatted public ssh key.
      * @return
      *   An option to the extracted key type.
      */
    def from(sshKeyString: String): Option[SshKeyType] = {
        val typeString = sshKeyString.takeWhile(char => !char.isWhitespace)
        Mappings.get(typeString)
    }
}

/** Actual ssh key and related metadata describing the public ssh key of a user.
  *
  * @param id
  *   The globally unique id of the ssh key.
  * @param ownerId
  *   The unique id of the user account associated with this key.
  * @param keyType
  *   The type of the key (rsa, ed25519, ...).
  * @param keyBytes
  *   The actual key (base 64 encoded).
  * @param fingerprint
  *   The fingerprint of the key using SHA-256.
  * @param comment
  *   An optional comment for the key, quite often this is an email address. Currently this is limited to 256
  *   characters.
  * @param createdAt
  *   The timestamp of when the ssh key was created.
  * @param lastUsedAt
  *   The timestamp of when the ssh key was last used by the user.
  */
final case class PublicSshKey(
    id: UUID,
    ownerId: UserId,
    keyType: SshKeyType,
    keyBytes: EncodedKeyBytes,
    fingerprint: KeyFingerprint,
    comment: Option[KeyComment],
    createdAt: OffsetDateTime,
    lastUsedAt: Option[OffsetDateTime]
)

object PublicSshKey {
    private val digest = MessageDigest.getInstance("SHA256")

    /** Create a [[PublicSshKey]] from the given parameters and [[SshPublicKeyString]].
      *
      * @param id
      *   The globally unique id of the ssh key.
      * @param ownerId
      *   The unique id of the user account associated with this key.
      * @param createdAt
      *   The timestamp of when the ssh key (this [[PublicSshKey]]) was created.
      * @param sshKey
      *   A string containing a valid OpenSSH public key.
      * @return
      *   An option to the sucessfully created [[PublicSshKey]].
      */
    def from(id: UUID)(ownerId: UserId)(createdAt: OffsetDateTime)(sshKey: SshPublicKeyString): Option[PublicSshKey] = {
        val keyType = SshKeyType.from(sshKey.toString)
        val base64Key = EncodedKeyBytes.from(
            sshKey.toString.dropWhile(char => !char.isWhitespace).trim.takeWhile(char => !char.isWhitespace)
        )
        val comment = sshKey.toString.split("\\s").drop(2).toList match {
            case Nil          => None
            case commentParts => KeyComment.from(commentParts.mkString(" "))
        }
        val fingerprint = base64Key.flatMap { base64Key =>
            val rawFingerprint = Try {
                val publicKey   = OpenSSHPublicKeyUtil.parsePublicKey(base64Key.toByteArray)
                val digestedKey = digest.digest(OpenSSHPublicKeyUtil.encodePublicKey(publicKey))
                Base64.getEncoder().withoutPadding().encodeToString(digestedKey)
            }.toOption
            rawFingerprint.flatMap(KeyFingerprint.from)
        }
        (keyType, base64Key, fingerprint).mapN { case (keyType, base64Key, fingerprint) =>
            PublicSshKey(id, ownerId, keyType, base64Key, fingerprint, comment, createdAt, None)
        }
    }

    extension (sshKey: PublicSshKey) {

        /** Convert this key into an instance of a [[SshPublicKeyStringhPublicKeyString]].
          *
          * @return
          *   A well formatted ssh public key string.
          */
        def toSshPublicKeyString: SshPublicKeyString =
            SshPublicKeyString(
                s"""${sshKey.keyType.identifier} ${sshKey.keyBytes.toString}${sshKey.comment
                        .map(c => s" ${c.toString}")
                        .getOrElse("")}"""
            )

        /** Convert this key into an instance of a AuthorizedKeyEntry useable by the Apache Mina SSHD library.
          *
          * @return
          *   Either the converted AuthorizedKeyEntry or an error message.
          */
        def toAuthorizedKeyEntry: Either[String, AuthorizedKeyEntry] =
            Try {
                AuthorizedKeyEntry.parseAuthorizedKeyEntry(sshKey.toSshPublicKeyString.toString)
            }.toEither.leftMap(_.getMessage())
    }
}
