/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.ssh

import java.util.UUID

import de.smederee.hub.*
import de.smederee.security.Username

/** The base class for needed repository functionality releated to ssh authentication like loading/providing keys.
  *
  * ### General notes ###
  *
  * {{{
  * 1. An account is considered *locked* **NOT** by the presence of an unlock token **BUT** by the presence
  *    of the `locked_at` date!
  * 2. All functions MUST work only on *unlocked* accounts!
  * }}}
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class SshAuthenticationRepository[F[_]] {

    /** Search for the ssh key with the given fingerprint for the specified owner who must not be locked.
      *
      * @param fingerprint
      *   The unique fingerprint of the ssh key.
      * @return
      *   An option to the ssh key if it exists.
      */
    def findSshKey(fingerprint: KeyFingerprint): F[Option[PublicSshKey]]

    /** Search for a repository owner of whom we only know the name.
      *
      * @param name
      *   The name of the repository owner which is the username of the actual owners account.
      * @return
      *   An option to successfully found repository owner.
      */
    def findVcsRepositoryOwner(name: Username): F[Option[VcsRepositoryOwner]]

    /** Update the last used column for the key in the database.
      *
      * @param keyId
      *   The unique id of the public ssh key.
      * @return
      *   The number of affected database rows.
      */
    def updateLastUsed(keyId: UUID): F[Int]

}
