/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.util.UUID

import cats.effect.*
import de.smederee.email.EmailAddress
import doobie.*
import doobie.implicits.*
import doobie.postgres.implicits.*
import org.slf4j.LoggerFactory

final class DoobieProjectRepository[F[_]: Sync](tx: Transactor[F]) extends ProjectRepository[F] {
    private val log = LoggerFactory.getLogger(getClass)

    given LogHandler[F] = Slf4jLogHandler.createLogHandler[F](log)

    given Meta[EmailAddress]       = Meta[String].timap(EmailAddress.apply)(_.toString)
    given Meta[ProjectDescription] = Meta[String].timap(ProjectDescription.apply)(_.toString)
    given Meta[ProjectId]          = Meta[Long].timap(ProjectId.apply)(_.toLong)
    given Meta[ProjectName]        = Meta[String].timap(ProjectName.apply)(_.toString)
    given Meta[ProjectOwnerId]     = Meta[UUID].timap(ProjectOwnerId.apply)(_.toUUID)
    given Meta[ProjectOwnerName]   = Meta[String].timap(ProjectOwnerName.apply)(_.toString)
    given Meta[TicketNumber]       = Meta[Int].timap(TicketNumber.apply)(_.toInt)

    override def createProject(project: Project): F[Int] =
        sql"""INSERT INTO tickets.projects (name, owner, is_private, description, created_at, updated_at)
          VALUES (
            ${project.name},
            ${project.owner.uid},
            ${project.isPrivate},
            ${project.description},
            NOW(),
            NOW()
          )""".update.run.transact(tx)

    override def deleteProject(project: Project): F[Int] =
        sql"""DELETE FROM tickets.projects WHERE owner = ${project.owner.uid} AND name = ${project.name}""".update.run
            .transact(tx)

    override def findProject(owner: ProjectOwner, name: ProjectName): F[Option[Project]] =
        sql"""SELECT
            users.uid AS owner_id,
            users.name AS owner_name,
            users.email AS owner_email,
            projects.name,
            projects.description,
            projects.is_private
          FROM tickets.projects AS projects
          JOIN tickets.users AS users
            ON projects.owner = users.uid
          WHERE
            projects.owner = ${owner.uid}
          AND
            projects.name = $name""".query[Project].option.transact(tx)

    override def findProjectId(owner: ProjectOwner, name: ProjectName): F[Option[ProjectId]] =
        sql"""SELECT
            projects.id
          FROM tickets.projects AS projects
          JOIN tickets.users AS users
            ON projects.owner = users.uid
          WHERE
            projects.owner = ${owner.uid}
          AND
            projects.name = $name""".query[ProjectId].option.transact(tx)

    override def findProjectOwner(name: ProjectOwnerName): F[Option[ProjectOwner]] =
        sql"""SELECT
            users.uid,
            users.name,
            users.email
          FROM tickets.users
          WHERE name = $name""".query[ProjectOwner].option.transact(tx)

    override def incrementNextTicketNumber(projectId: ProjectId): F[TicketNumber] = {
        // TODO: Find out which of the queries is more reliable and more performant.
        /*
        val sqlQuery1 = sql"""UPDATE tickets.projects AS alias1
                            SET next_ticket_number = alias2.next_ticket_number + 1
                          FROM (
                            SELECT
                              id,
                              next_ticket_number
                            FROM tickets.projects
                            WHERE id = $projectId
                          ) AS alias2
                          WHERE alias1.id = alias2.id
                          RETURNING alias2.next_ticket_number AS next_ticket_number"""
         */
        val sqlQuery2 = sql"""WITH old_number AS (
                            SELECT next_ticket_number FROM tickets.projects WHERE id = $projectId
                          )
                          UPDATE tickets.projects
                            SET next_ticket_number = next_ticket_number + 1
                          WHERE id = $projectId
                          RETURNING (
                            SELECT next_ticket_number FROM old_number
                          )"""
        sqlQuery2.query[TicketNumber].unique.transact(tx)
    }

    override def updateProject(project: Project): F[Int] =
        sql"""UPDATE tickets.projects SET
            is_private = ${project.isPrivate},
            description = ${project.description},
            updated_at = NOW()
          WHERE
            owner = ${project.owner.uid}
          AND name = ${project.name}""".update.run.transact(tx)
}
