/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.util.UUID

import cats.*
import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import doobie.*
import doobie.Fragments.*
import doobie.implicits.*
import doobie.postgres.implicits.*
import fs2.Stream
import org.slf4j.LoggerFactory

final class DoobieTicketRepository[F[_]: Sync](tx: Transactor[F]) extends TicketRepository[F] {
    private val log = LoggerFactory.getLogger(getClass)

    given LogHandler[F] = Slf4jLogHandler.createLogHandler[F](log)

    given Meta[AssigneeId]           = Meta[UUID].timap(AssigneeId.apply)(_.toUUID)
    given Meta[AssigneeName]         = Meta[String].timap(AssigneeName.apply)(_.toString)
    given Meta[ColourCode]           = Meta[String].timap(ColourCode.apply)(_.toString)
    given Meta[LabelDescription]     = Meta[String].timap(LabelDescription.apply)(_.toString)
    given Meta[LabelId]              = Meta[Long].timap(LabelId.apply)(_.toLong)
    given Meta[LabelName]            = Meta[String].timap(LabelName.apply)(_.toString)
    given Meta[MilestoneDescription] = Meta[String].timap(MilestoneDescription.apply)(_.toString)
    given Meta[MilestoneId]          = Meta[Long].timap(MilestoneId.apply)(_.toLong)
    given Meta[MilestoneTitle]       = Meta[String].timap(MilestoneTitle.apply)(_.toString)
    given Meta[ProjectId]            = Meta[Long].timap(ProjectId.apply)(_.toLong)
    given Meta[SubmitterId]          = Meta[UUID].timap(SubmitterId.apply)(_.toUUID)
    given Meta[SubmitterName]        = Meta[String].timap(SubmitterName.apply)(_.toString)
    given Meta[TicketContent]        = Meta[String].timap(TicketContent.apply)(_.toString)
    given Meta[TicketId]             = Meta[Long].timap(TicketId.apply)(_.toLong)
    given Meta[TicketNumber]         = Meta[Int].timap(TicketNumber.apply)(_.toInt)
    given Meta[TicketResolution]     = Meta[String].timap(TicketResolution.valueOf)(_.toString)
    given Meta[TicketStatus]         = Meta[String].timap(TicketStatus.valueOf)(_.toString)
    given Meta[TicketTitle]          = Meta[String].timap(TicketTitle.apply)(_.toString)

    private val selectTicketColumns =
        fr"""SELECT
            tickets.number AS number,
            tickets.title AS title,
            tickets.content AS content,
            tickets.status AS status,
            tickets.resolution AS resolution,
            submitters.uid AS submitter_uid,
            submitters.name AS submitter_name,
            tickets.created_at AS created_at,
            tickets.updated_at AS updated_at
          FROM tickets.tickets AS tickets
          LEFT OUTER JOIN tickets.users AS submitters
          ON tickets.submitter = submitters.uid"""

    /** Construct a query fragment that fetches the internal unique ticket id from the tickets table via the given
      * project id and ticket number. The fetched id can be referenced like this `SELECT id FROM ticket_id`.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @return
      *   A query fragment useable within other queries which defines a common table expression using the `WITH` clause.
      */
    private def withTicketId(projectId: ProjectId, ticketNumber: TicketNumber): Fragment =
        fr"""WITH ticket_id AS (
           SELECT id AS id
           FROM tickets.tickets AS tickets
           WHERE tickets.project = $projectId
           AND tickets.number = $ticketNumber)"""

    override def addAssignee(projectId: ProjectId)(ticketNumber: TicketNumber)(assignee: Assignee): F[Int] =
        sql"""INSERT INTO tickets.ticket_assignees (
            ticket,
            assignee
          ) SELECT 
            id,
            ${assignee.id}
          FROM tickets.tickets
          WHERE project = $projectId
          AND number = $ticketNumber""".update.run.transact(tx)

    override def addLabel(projectId: ProjectId)(ticketNumber: TicketNumber)(label: Label): F[Int] =
        label.id match {
            case None => Sync[F].pure(0)
            case Some(labelId) =>
                sql"""INSERT INTO tickets.ticket_labels (
            ticket,
            label
          ) SELECT
            id,
            $labelId
          FROM tickets.tickets
          WHERE project = $projectId
          AND number = $ticketNumber""".update.run.transact(tx)
        }

    override def addMilestone(projectId: ProjectId)(ticketNumber: TicketNumber)(milestone: Milestone): F[Int] =
        milestone.id match {
            case None => Sync[F].pure(0)
            case Some(milestoneId) =>
                sql"""INSERT INTO tickets.milestone_tickets (
                ticket,
                milestone
              ) SELECT
                id,
                $milestoneId
          FROM tickets.tickets
          WHERE project = $projectId
          AND number = $ticketNumber""".update.run.transact(tx)
        }

    override def allTickets(filter: Option[TicketFilter])(projectId: ProjectId): Stream[F, Ticket] = {
        val projectFilter = fr"""tickets.project = $projectId"""
        val tickets = filter match {
            case None => selectTicketColumns ++ whereAnd(projectFilter)
            case Some(filter) =>
                val numberFilter = filter.number.toNel.map(numbers => Fragments.in(fr"""tickets.number""", numbers))
                val statusFilter = filter.status.toNel.map(status => Fragments.in(fr"""tickets.status""", status))
                val resolutionFilter =
                    filter.resolution.toNel.map(resolutions => Fragments.in(fr"""tickets.resolution""", resolutions))
                val submitterFilter =
                    filter.submitter.toNel.map(submitters => Fragments.in(fr"""submitters.name""", submitters))
                selectTicketColumns ++ whereAndOpt(
                    projectFilter.some,
                    numberFilter,
                    statusFilter,
                    resolutionFilter,
                    submitterFilter
                )
        }
        tickets.query[Ticket].stream.transact(tx)
    }

    override def createTicket(projectId: ProjectId)(ticket: Ticket): F[Int] =
        sql"""INSERT INTO tickets.tickets (
            project,
            number,
            title,
            content,
            status,
            resolution,
            submitter,
            created_at,
            updated_at
          ) VALUES (
            $projectId,
            ${ticket.number},
            ${ticket.title},
            ${ticket.content},
            ${ticket.status},
            ${ticket.resolution},
            ${ticket.submitter.map(_.id)},
            NOW(),
            NOW()
          )""".update.run.transact(tx)

    override def deleteTicket(projectId: ProjectId)(ticket: Ticket): F[Int] =
        sql"""DELETE FROM tickets.tickets
          WHERE project = $projectId
          AND number = ${ticket.number}""".update.run.transact(tx)

    override def findTicket(projectId: ProjectId)(ticketNumber: TicketNumber): F[Option[Ticket]] = {
        val projectFilter = fr"""project = $projectId"""
        val numberFilter  = fr"""number = $ticketNumber"""
        val ticket        = selectTicketColumns ++ whereAnd(projectFilter, numberFilter) ++ fr"""LIMIT 1"""
        ticket.query[Ticket].option.transact(tx)
    }

    override def findTicketId(projectId: ProjectId)(ticketNumber: TicketNumber): F[Option[TicketId]] =
        sql"""SELECT id FROM tickets.tickets WHERE project = $projectId AND number = $ticketNumber"""
            .query[TicketId]
            .option
            .transact(tx)

    override def loadAssignees(projectId: ProjectId)(ticketNumber: TicketNumber): Stream[F, Assignee] = {
        val sqlQuery = withTicketId(projectId, ticketNumber) ++
            fr"""SELECT
             users.uid AS uid,
             users.name AS name
           FROM tickets.ticket_assignees AS assignees
           JOIN tickets.users            AS users
             ON assignees.assignee = users.uid
           WHERE assignees.ticket = (SELECT id FROM ticket_id)"""
        sqlQuery.query[Assignee].stream.transact(tx)
    }

    override def loadLabels(projectId: ProjectId)(ticketNumber: TicketNumber): Stream[F, Label] = {
        val sqlQuery = withTicketId(projectId, ticketNumber) ++
            fr"""SELECT
             labels.id          AS id,
             labels.name        AS name,
             labels.description AS description,
             labels.colour      AS colour
           FROM tickets.labels        AS labels
           JOIN tickets.ticket_labels AS ticket_labels
             ON labels.id = ticket_labels.label
           WHERE ticket_labels.ticket = (SELECT id FROM ticket_id)"""
        sqlQuery.query[Label].stream.transact(tx)
    }

    override def loadMilestones(projectId: ProjectId)(ticketNumber: TicketNumber): Stream[F, Milestone] = {
        val sqlQuery = withTicketId(projectId, ticketNumber) ++
            fr"""SELECT
             milestones.id AS id,
             milestones.title AS title,
             milestones.description AS description,
             milestones.due_date AS due_date,
             milestones.closed AS closed
           FROM tickets.milestones        AS milestones
           JOIN tickets.milestone_tickets AS milestone_tickets
             ON milestones.id = milestone_tickets.milestone
           WHERE milestone_tickets.ticket = (SELECT id FROM ticket_id)
           ORDER BY milestones.due_date ASC, milestones.title ASC"""
        sqlQuery.query[Milestone].stream.transact(tx)
    }

    override def removeAssignee(projectId: ProjectId)(ticket: Ticket)(assignee: Assignee): F[Int] = {
        val sqlQuery = withTicketId(projectId, ticket.number) ++
            fr"""DELETE FROM tickets.ticket_assignees AS ticket_assignees
           WHERE ticket_assignees.ticket = (SELECT id FROM ticket_id)
           AND ticket_assignees.assignee = ${assignee.id}"""
        sqlQuery.update.run.transact(tx)
    }

    override def removeLabel(projectId: ProjectId)(ticket: Ticket)(label: Label): F[Int] =
        label.id match {
            case None => Sync[F].pure(0)
            case Some(labelId) =>
                val sqlQuery = withTicketId(projectId, ticket.number) ++
                    fr"""DELETE FROM tickets.ticket_labels AS labels
               WHERE labels.ticket = (SELECT id FROM ticket_id)
               AND labels.label = $labelId"""
                sqlQuery.update.run.transact(tx)
        }

    override def removeMilestone(projectId: ProjectId)(ticket: Ticket)(milestone: Milestone): F[Int] =
        milestone.id match {
            case None => Sync[F].pure(0)
            case Some(milestoneId) =>
                val sqlQuery = withTicketId(projectId, ticket.number) ++
                    fr"""DELETE FROM tickets.milestone_tickets AS milestone_tickets
               WHERE milestone_tickets.ticket = (SELECT id FROM ticket_id)
               AND milestone_tickets.milestone = $milestoneId"""
                sqlQuery.update.run.transact(tx)
        }

    override def updateTicket(projectId: ProjectId)(ticket: Ticket): F[Int] =
        sql"""UPDATE tickets.tickets SET
            title = ${ticket.title},
            content = ${ticket.content},
            status = ${ticket.status},
            resolution = ${ticket.resolution},
            submitter = ${ticket.submitter.map(_.id)},
            updated_at = NOW()
          WHERE project = $projectId
          AND number = ${ticket.number}""".update.run.transact(tx)
}
