/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import cats.*
import cats.data.*
import cats.syntax.all.*
import de.smederee.tickets.forms.FormValidator
import de.smederee.tickets.forms.types.*

/** Data container to edit a label.
  *
  * @param id
  *   An optional attribute containing the unique internal database ID for the label.
  * @param name
  *   A short descriptive name for the label which is supposed to be unique in a project context.
  * @param description
  *   An optional description if needed.
  * @param colour
  *   A hexadecimal HTML colour code which can be used to mark the label on a rendered website.
  */
final case class LabelForm(
    id: Option[LabelId],
    name: LabelName,
    description: Option[LabelDescription],
    colour: ColourCode
)

object LabelForm extends FormValidator[LabelForm] {
    val fieldColour: FormField      = FormField("colour")
    val fieldDescription: FormField = FormField("description")
    val fieldId: FormField          = FormField("id")
    val fieldName: FormField        = FormField("name")

    /** Create a form for editing a label from the given label data.
      *
      * @param label
      *   The label which provides the data for the edit form.
      * @return
      *   A label form filled with the data from the given label.
      */
    def fromLabel(label: Label): LabelForm =
        LabelForm(id = label.id, name = label.name, description = label.description, colour = label.colour)

    override def validate(data: Map[String, Chain[String]]): ValidatedNec[FormErrors, LabelForm] = {
        val id = data
            .get(fieldId)
            .fold(Option.empty[LabelId].validNec)(
                _.headOption
                    .flatMap(LabelId.fromString)
                    .fold(FormFieldError("Invalid label id!").invalidNec)(id => Option(id).validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldId -> es.toList)))
        val name = data
            .get(fieldName)
            .fold(FormFieldError("No label name given!").invalidNec)(
                _.headOption
                    .map(_.trim) // We strip leading and trailing whitespace!
                    .flatMap(LabelName.from)
                    .fold(FormFieldError("Invalid label name!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldName -> es.toList)))
        val description = data
            .get(fieldDescription)
            .fold(Option.empty[LabelDescription].validNec)(
                _.headOption
                    .map(_.trim)
                    .filter(_.nonEmpty)
                    .fold(none[LabelDescription].validNec)(s =>
                        LabelDescription
                            .from(s)
                            .fold(FormFieldError("Invalid label description!").invalidNec)(descr =>
                                Option(descr).validNec
                            )
                    )
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldDescription -> es.toList)))
        val colour = data
            .get(fieldColour)
            .fold(FormFieldError("No label colour given!").invalidNec)(
                _.headOption
                    .flatMap(ColourCode.from)
                    .fold(FormFieldError("Invalid label colour!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldColour -> es.toList)))
        (id, name, description, colour).mapN { case (id, name, description, colour) =>
            LabelForm(id, name, description, colour)
        }
    }

    extension (form: LabelForm) {

        /** Convert the form class into a stringified map which is used as underlying data type for form handling in the
          * twirl templating library.
          *
          * @return
          *   A stringified map containing the data of the form.
          */
        def toMap: Map[String, Chain[String]] =
            Map(
                LabelForm.fieldId.toString   -> form.id.map(_.toString).fold(Chain.empty)(id => Chain(id)),
                LabelForm.fieldName.toString -> Chain(form.name.toString),
                LabelForm.fieldDescription.toString -> form.description
                    .map(_.toString)
                    .fold(Chain.empty)(d => Chain(d)),
                LabelForm.fieldColour.toString -> Chain(form.colour.toString)
            )
    }

}
