/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.time.*

import cats.*
import cats.data.*
import cats.syntax.all.*
import de.smederee.tickets.forms.FormValidator
import de.smederee.tickets.forms.types.*

/** Data container to edit a milestone.
  *
  * @param id
  *   An optional attribute containing the unique internal database ID for the milestone.
  * @param title
  *   A title for the milestone, usually a version number, a word or a short phrase that is supposed to be unique within
  *   a project context.
  * @param description
  *   An optional longer description of the milestone.
  * @param dueDate
  *   An optional date on which the milestone is supposed to be reached.
  */
final case class MilestoneForm(
    id: Option[MilestoneId],
    title: MilestoneTitle,
    description: Option[MilestoneDescription],
    dueDate: Option[LocalDate]
)

object MilestoneForm extends FormValidator[MilestoneForm] {
    val fieldDescription: FormField = FormField("description")
    val fieldDueDate: FormField     = FormField("due_date")
    val fieldId: FormField          = FormField("id")
    val fieldTitle: FormField       = FormField("title")

    /** Create a form for editing a milestone from the given milestone data.
      *
      * @param milestone
      *   The milestone which provides the data for the edit form.
      * @return
      *   A milestone form filled with the data from the given milestone.
      */
    def fromMilestone(milestone: Milestone): MilestoneForm =
        MilestoneForm(
            id = milestone.id,
            title = milestone.title,
            description = milestone.description,
            dueDate = milestone.dueDate
        )

    override def validate(data: Map[String, Chain[String]]): ValidatedNec[FormErrors, MilestoneForm] = {
        val id = data
            .get(fieldId)
            .fold(Option.empty[MilestoneId].validNec)(
                _.headOption
                    .flatMap(MilestoneId.fromString)
                    .fold(FormFieldError("Invalid milestone id!").invalidNec)(id => Option(id).validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldId -> es.toList)))
        val title = data
            .get(fieldTitle)
            .fold(FormFieldError("No milestone title given!").invalidNec)(
                _.headOption
                    .map(_.trim) // We strip leading and trailing whitespace!
                    .flatMap(MilestoneTitle.from)
                    .fold(FormFieldError("Invalid milestone title!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldTitle -> es.toList)))
        val description = data
            .get(fieldDescription)
            .fold(Option.empty[MilestoneDescription].validNec)(
                _.headOption
                    .map(_.trim)
                    .filter(_.nonEmpty)
                    .fold(none[MilestoneDescription].validNec)(s =>
                        MilestoneDescription
                            .from(s)
                            .fold(FormFieldError("Invalid milestone description!").invalidNec)(descr =>
                                Option(descr).validNec
                            )
                    )
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldDescription -> es.toList)))
        val dueDate = data
            .get(fieldDueDate)
            .fold(Option.empty[LocalDate].validNec)(
                _.headOption
                    .map(_.trim)
                    .filter(_.nonEmpty)
                    .fold(none[LocalDate].validNec)(s =>
                        Validated.catchNonFatal(LocalDate.parse(s)).map(date => date.some)
                    )
            )
            .leftMap(_ => NonEmptyChain.of(Map(fieldDueDate -> List(FormFieldError("Invalid milestone due date!")))))
        (id, title, description, dueDate).mapN { case (id, title, description, dueDate) =>
            MilestoneForm(id, title, description, dueDate)
        }
    }

    extension (form: MilestoneForm) {

        /** Convert the form class into a stringified map which is used as underlying data type for form handling in the
          * twirl templating library.
          *
          * @return
          *   A stringified map containing the data of the form.
          */
        def toMap: Map[String, Chain[String]] =
            Map(
                MilestoneForm.fieldId.toString    -> form.id.map(_.toString).fold(Chain.empty)(id => Chain(id)),
                MilestoneForm.fieldTitle.toString -> Chain(form.title.toString),
                MilestoneForm.fieldDescription.toString -> form.description
                    .map(_.toString)
                    .fold(Chain.empty)(d => Chain(d)),
                MilestoneForm.fieldDueDate.toString -> form.dueDate
                    .map(_.toString)
                    .fold(Chain.empty)(date => Chain(date))
            )
    }

}
