/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.util.UUID

import cats.*
import cats.data.*
import cats.syntax.all.*
import de.smederee.email.EmailAddress
import de.smederee.security.UserId
import de.smederee.security.Username

import scala.util.Try
import scala.util.matching.Regex

opaque type ProjectDescription = String
object ProjectDescription {
    val MaximumLength: Int = 8192

    /** Create an instance of ProjectDescription from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a ProjectDescription.
      * @return
      *   The appropriate instance of ProjectDescription.
      */
    def apply(source: String): ProjectDescription = source

    /** Try to create an instance of ProjectDescription from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a ProjectDescription.
      * @return
      *   An option to the successfully converted ProjectDescription.
      */
    def from(source: String): Option[ProjectDescription] = Option(source).map(_.take(MaximumLength))

}

opaque type ProjectId = Long
object ProjectId {
    given Eq[ProjectId] = Eq.fromUniversalEquals

    val Format: Regex = "^-?\\d+$".r

    /** Create an instance of ProjectId from the given Long type.
      *
      * @param source
      *   An instance of type Long which will be returned as a ProjectId.
      * @return
      *   The appropriate instance of ProjectId.
      */
    def apply(source: Long): ProjectId = source

    /** Try to create an instance of ProjectId from the given Long.
      *
      * @param source
      *   A Long that should fulfil the requirements to be converted into a ProjectId.
      * @return
      *   An option to the successfully converted ProjectId.
      */
    def from(source: Long): Option[ProjectId] = Option(source)

    /** Try to create an instance of ProjectId from the given String.
      *
      * @param source
      *   A string that should fulfil the requirements to be converted into a ProjectId.
      * @return
      *   An option to the successfully converted ProjectId.
      */
    def fromString(source: String): Option[ProjectId] =
        Option(source).filter(Format.matches).flatMap(string => Try(string.toLong).toOption).flatMap(from)

    extension (id: ProjectId) {
        def toLong: Long = id
    }
}

opaque type ProjectName = String
object ProjectName {

    given Eq[ProjectName] = Eq.fromUniversalEquals

    given Order[ProjectName] = Order.from((a, b) => a.toString.compareTo(b.toString))

    // TODO: Can we rewrite this in a Scala-3 way (i.e. without implicitly)?
    given Ordering[ProjectName] = implicitly[Order[ProjectName]].toOrdering

    val Format: Regex = "^[a-zA-Z0-9][a-zA-Z0-9\\-_]{1,63}$".r

    /** Create an instance of ProjectName from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a ProjectName.
      * @return
      *   The appropriate instance of ProjectName.
      */
    def apply(source: String): ProjectName = source

    /** Try to create an instance of ProjectName from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a ProjectName.
      * @return
      *   An option to the successfully converted ProjectName.
      */
    def from(source: String): Option[ProjectName] = validate(source).toOption

    /** Validate the given string and return either the validated repository name or a list of errors.
      *
      * @param s
      *   An arbitrary string which should be a repository name.
      * @return
      *   Either a list of errors or the validated repository name.
      */
    def validate(s: String): ValidatedNec[String, ProjectName] =
        Option(s).map(_.trim.nonEmpty) match {
            case Some(true) =>
                val input = s.trim
                val miniumLength =
                    if (input.length > 1)
                        input.validNec
                    else
                        "Repository name too short (min. 2 characters)!".invalidNec
                val maximumLength =
                    if (input.length < 65)
                        input.validNec
                    else
                        "Repository name too long (max. 64 characters)!".invalidNec
                val validFormat =
                    if (Format.matches(input))
                        input.validNec
                    else
                        "Repository name must start with a letter or number and is only allowed to contain alphanumeric characters plus .".invalidNec
                (miniumLength, maximumLength, validFormat).mapN { case (_, _, name) =>
                    name
                }
            case _ => "Repository name must not be empty!".invalidNec
        }
}

/** Extractor to retrieve a ProjectName from a path parameter.
  */
object ProjectNamePathParameter {
    def unapply(str: String): Option[ProjectName] = Option(str).flatMap(ProjectName.from)
}

/** A project owner id is supposed to be globally unique and maps to the [[java.util.UUID]] type beneath.
  */
opaque type ProjectOwnerId = UUID
object ProjectOwnerId {
    val Format = "^[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}$".r

    given Eq[ProjectOwnerId] = Eq.fromUniversalEquals

    // given Conversion[UserId, ProjectOwnerId] = ProjectOwnerId.fromUserId

    /** Create an instance of ProjectOwnerId from the given UUID type.
      *
      * @param source
      *   An instance of type UUID which will be returned as a ProjectOwnerId.
      * @return
      *   The appropriate instance of ProjectOwnerId.
      */
    def apply(source: UUID): ProjectOwnerId = source

    /** Try to create an instance of ProjectOwnerId from the given UUID.
      *
      * @param source
      *   A UUID that should fulfil the requirements to be converted into a ProjectOwnerId.
      * @return
      *   An option to the successfully converted ProjectOwnerId.
      */
    def from(source: UUID): Option[ProjectOwnerId] = Option(source)

    /** Try to create an instance of ProjectOwnerId from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a ProjectOwnerId.
      * @return
      *   An option to the successfully converted ProjectOwnerId.
      */
    def fromString(source: String): Either[String, ProjectOwnerId] =
        Option(source)
            .filter(s => Format.matches(s))
            .flatMap { uuidString =>
                Either.catchNonFatal(UUID.fromString(uuidString)).toOption
            }
            .toRight("Illegal value for ProjectOwnerId!")

    /** Create an instance of ProjectOwnerId from the given UserId type.
      *
      * @param uid
      *   An instance of type UserId which will be returned as a ProjectOwnerId.
      * @return
      *   The appropriate instance of ProjectOwnerId.
      */
    def fromUserId(uid: UserId): ProjectOwnerId = uid.toUUID

    /** Generate a new random project owner id.
      *
      * @return
      *   A project owner id which is pseudo randomly generated.
      */
    def randomProjectOwnerId: ProjectOwnerId = UUID.randomUUID

    extension (uid: ProjectOwnerId) {
        def toUUID: UUID = uid
    }
}

/** A project owner name for an account has to obey several restrictions which are similiar to the ones found for Unix
  * usernames. It must start with a letter, contain only alphanumeric characters, be at least 2 and at most 31
  * characters long and be all lowercase.
  */
opaque type ProjectOwnerName = String
object ProjectOwnerName {
    given Eq[ProjectOwnerName] = Eq.fromUniversalEquals

    given Conversion[Username, ProjectOwnerName] = ProjectOwnerName.fromUsername

    val isAlphanumeric = "^[a-z][a-z0-9]+$".r

    /** Create an instance of ProjectOwnerName from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a ProjectOwnerName.
      * @return
      *   The appropriate instance of ProjectOwnerName.
      */
    def apply(source: String): ProjectOwnerName = source

    /** Try to create an instance of ProjectOwnerName from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a ProjectOwnerName.
      * @return
      *   An option to the successfully converted ProjectOwnerName.
      */
    def from(s: String): Option[ProjectOwnerName] = validate(s).toOption

    /** Create an instance of ProjectOwnerName from the given Username type.
      *
      * @param username
      *   An instance of the type Username which will be returned as a ProjectOwnerName.
      * @return
      *   The appropriate instance of ProjectOwnerName.
      */
    def fromUsername(username: Username): ProjectOwnerName = username.toString

    /** Validate the given string and return either the validated project owner name or a list of errors.
      *
      * @param s
      *   An arbitrary string which should be a project owner name.
      * @return
      *   Either a list of errors or the validated project owner name.
      */
    def validate(s: String): ValidatedNec[String, ProjectOwnerName] =
        Option(s).map(_.trim.nonEmpty) match {
            case Some(true) =>
                val input = s.trim
                val miniumLength =
                    if (input.length >= 2)
                        input.validNec
                    else
                        "ProjectOwnerName too short (min. 2 characters)!".invalidNec
                val maximumLength =
                    if (input.length < 32)
                        input.validNec
                    else
                        "ProjectOwnerName too long (max. 31 characters)!".invalidNec
                val alphanumeric =
                    if (isAlphanumeric.matches(input))
                        input.validNec
                    else
                        "ProjectOwnerName must be all lowercase alphanumeric characters and start with a character.".invalidNec
                (miniumLength, maximumLength, alphanumeric).mapN { case (_, _, name) =>
                    name
                }
            case _ => "ProjectOwnerName must not be empty!".invalidNec
        }

    extension (ownername: ProjectOwnerName) {

        /** Convert this project owner name into a username.
          *
          * @return
          *   A syntactically valid username.
          */
        def toUsername: Username = Username(ownername.toString)
    }
}

/** Extractor to retrieve an ProjectOwnerName from a path parameter.
  */
object ProjectOwnerNamePathParameter {
    def unapply(str: String): Option[ProjectOwnerName] =
        Option(str).flatMap { string =>
            if (string.startsWith("~"))
                ProjectOwnerName.from(string.drop(1))
            else
                None
        }
}

/** Descriptive information about the owner of a project.
  *
  * @param owner
  *   The unique ID of the project owner.
  * @param name
  *   The name of the project owner which is supposed to be unique.
  * @param email
  *   The email address of the project owner.
  */
final case class ProjectOwner(uid: ProjectOwnerId, name: ProjectOwnerName, email: EmailAddress)

object ProjectOwner {
    given Eq[ProjectOwner] = Eq.fromUniversalEquals
}

/** A project is the base entity for tracking tickets.
  *
  * @param owner
  *   The owner of the project.
  * @param name
  *   The name of the project. A project name must start with a letter or number and must contain only alphanumeric
  *   ASCII characters as well as minus or underscore signs. It must be between 2 and 64 characters long.
  * @param description
  *   An optional short text description of the project.
  * @param isPrivate
  *   A flag indicating if this project is private i.e. only visible / accessible for accounts with appropriate
  *   permissions.
  */
final case class Project(
    owner: ProjectOwner,
    name: ProjectName,
    description: Option[ProjectDescription],
    isPrivate: Boolean
)
