/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import cats.effect.*
import doobie.util.log.*
import org.slf4j.Logger

object Slf4jLogHandler {
    private val RedactArguments: Boolean = true // This SHALL only be set to `false` when debugging issues!

    private val sqlArgumentsToLogString: List[Any] => String = arguments =>
        if (RedactArguments)
            arguments.map(_ => "redacted").mkString(", ")
        else
            arguments.mkString(", ")

    private val sqlQueryToLogString: String => String = _.linesIterator.dropWhile(_.trim.isEmpty).mkString("\n  ")

    /** Create a [[doobie.util.log.LogHandler]] for logging doobie queries and errors. For convenience it is best to
      * simply return the return value of this method to a given (implicit) instance.
      *
      * @param log
      *   A logger which provides an slf4j interface.
      * @return
      *   A log handler as expected by doobie.
      */
    def createLogHandler[F[_]: Sync](log: Logger): LogHandler[F] =
        new LogHandler[F] {
            def run(logEvent: LogEvent): F[Unit] =
                Sync[F].delay {
                    logEvent match {
                        case Success(sqlQuery, arguments, label, executionTime, processingTime) =>
                            log.debug(s"""SQL command successful:
                           |
                           | ${sqlQueryToLogString(sqlQuery)}
                           |
                           | arguments: [${sqlArgumentsToLogString(arguments)}]
                           | label: $label
                           |
                           | execution time : ${executionTime.toMillis} ms
                           | processing time: ${processingTime.toMillis} ms
                           | total time     : ${(executionTime + processingTime).toMillis} ms
                           |""".stripMargin)
                        case ProcessingFailure(sqlQuery, arguments, label, executionTime, processingTime, failure) =>
                            log.error(
                                s"""SQL PROCESSING FAILURE:
                   |
                   | ${sqlQueryToLogString(sqlQuery)}
                   |
                   | arguments: [${sqlArgumentsToLogString(arguments)}]
                   | label: $label
                   |
                   | execution time : ${executionTime.toMillis} ms
                   | processing time: ${processingTime.toMillis} ms
                   | total time     : ${(executionTime + processingTime).toMillis} ms
                   |""".stripMargin,
                                failure
                            )
                        case ExecFailure(sqlQuery, arguments, label, executionTime, failure) =>
                            log.error(
                                s"""SQL EXECUTION FAILURE:
                   |
                   | ${sqlQueryToLogString(sqlQuery)}
                   |
                   | arguments: [${sqlArgumentsToLogString(arguments)}]
                   | label: $label
                   |
                   | execution time : ${executionTime.toMillis} ms
                   |""".stripMargin,
                                failure
                            )
                    }
                }
        }
}
