/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import fs2.Stream

/** The base class that defines the needed functionality to handle tickets and related data types within a database.
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class TicketRepository[F[_]] {

    /** Add the given assignee to the ticket of the given repository id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @param assignee
      *   The assignee to be added to the ticket.
      * @return
      *   The number of affected database rows.
      */
    def addAssignee(projectId: ProjectId)(ticketNumber: TicketNumber)(assignee: Assignee): F[Int]

    /** Add the given label to the ticket of the given repository id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @param label
      *   The label to be added to the ticket.
      * @return
      *   The number of affected database rows.
      */
    def addLabel(projectId: ProjectId)(ticketNumber: TicketNumber)(label: Label): F[Int]

    /** Add the given milestone to the ticket of the given repository id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @param milestone
      *   The milestone to be added to the ticket.
      * @return
      *   The number of affected database rows.
      */
    def addMilestone(projectId: ProjectId)(ticketNumber: TicketNumber)(milestone: Milestone): F[Int]

    /** Return all tickets associated with the given repository.
      *
      * @param filter
      *   A ticket filter containing possible values which will be used to filter the list of tickets.
      * @param projectId
      *   The unique internal ID of a ticket tracking project for which all tickets shall be returned.
      * @return
      *   A stream of tickets associated with the vcs repository which may be empty.
      */
    def allTickets(filter: Option[TicketFilter])(projectId: ProjectId): Stream[F, Ticket]

    /** Create a database entry for the given ticket definition within the scope of the repository with the given id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticket
      *   The ticket definition that shall be written to the database.
      * @return
      *   The number of affected database rows.
      */
    def createTicket(projectId: ProjectId)(ticket: Ticket): F[Int]

    /** Delete the ticket of the repository with the given id from the database.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticket
      *   The ticket definition that shall be deleted from the database.
      * @return
      *   The number of affected database rows.
      */
    def deleteTicket(projectId: ProjectId)(ticket: Ticket): F[Int]

    /** Find the ticket with the given number of the repository with the given id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @return
      *   An option to the found ticket.
      */
    def findTicket(projectId: ProjectId)(ticketNumber: TicketNumber): F[Option[Ticket]]

    /** Find the ticket with the given number of the project with the given id and return the internal unique id of the
      * ticket.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @return
      *   An option to the found ticket.
      */
    def findTicketId(projectId: ProjectId)(ticketNumber: TicketNumber): F[Option[TicketId]]

    /** Load all assignees that are assigned to the ticket with the given number and repository id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @return
      *   A stream of assigness that may be empty.
      */
    def loadAssignees(projectId: ProjectId)(ticketNumber: TicketNumber): Stream[F, Assignee]

    /** Load all labels that are attached to the ticket with the given number and repository id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @return
      *   A stream of labels that may be empty.
      */
    def loadLabels(projectId: ProjectId)(ticketNumber: TicketNumber): Stream[F, Label]

    /** Load all milestones that are attached to the ticket with the given number and repository id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @return
      *   A stream of milestones that may be empty.
      */
    def loadMilestones(projectId: ProjectId)(ticketNumber: TicketNumber): Stream[F, Milestone]

    /** Remove the given assignee from the ticket of the given repository id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @param assignee
      *   The assignee to be removed from the ticket.
      * @return
      *   The number of affected database rows.
      */
    def removeAssignee(projectId: ProjectId)(ticket: Ticket)(assignee: Assignee): F[Int]

    /** Remove the given label from the ticket of the given repository id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @param label
      *   The label to be removed from the ticket.
      * @return
      *   The number of affected database rows.
      */
    def removeLabel(projectId: ProjectId)(ticket: Ticket)(label: Label): F[Int]

    /** Remove the given milestone from the ticket of the given repository id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticketNumber
      *   The unique identifier of a ticket within the project scope is its number.
      * @param milestone
      *   The milestone to be removed from the ticket.
      * @return
      *   The number of affected database rows.
      */
    def removeMilestone(projectId: ProjectId)(ticket: Ticket)(milestone: Milestone): F[Int]

    /** Update the database entry for the given ticket within the scope of the repository with the given id.
      *
      * @param projectId
      *   The unique internal ID of a ticket tracking project.
      * @param ticket
      *   The ticket definition that shall be updated within the database.
      * @return
      *   The number of affected database rows.
      */
    def updateTicket(projectId: ProjectId)(ticket: Ticket): F[Int]

}
