/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.time.OffsetDateTime
import java.time.ZoneOffset

import cats.*
import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.html.*
import de.smederee.html.LinkTools.*
import de.smederee.hub.Account
import de.smederee.hub.RequestHelpers.instances.given
import de.smederee.i18n.LanguageCode
import de.smederee.security.CsrfToken
import de.smederee.security.Username
import de.smederee.tickets.config.*
import de.smederee.tickets.forms.types.*
import org.http4s.*
import org.http4s.dsl.Http4sDsl
import org.http4s.headers.Location
import org.http4s.twirl.TwirlInstances.*
import org.slf4j.LoggerFactory

/** Routes for managing tickets.
  *
  * @param configuration
  *   The ticket service configuration.
  * @param labelRepo
  *   A repository for handling database operations for labels.
  * @param milestoneRepo
  *   A repository for handling database operations for milestones.
  * @param projectRepo
  *   A repository for handling database operations regarding our projects and their metadata.
  * @param ticketRepo
  *   A repository for handling database operations regarding tickets and related metadata.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class TicketRoutes[F[_]: Async](
    configuration: SmedereeTicketsConfiguration,
    labelRepo: LabelRepository[F],
    milestoneRepo: MilestoneRepository[F],
    projectRepo: ProjectRepository[F],
    ticketRepo: TicketRepository[F]
) extends Http4sDsl[F] {
    private val log = LoggerFactory.getLogger(getClass)

    private val linkToHubService = configuration.hub.baseUri
    private val linkConfig       = configuration.external

    /** Load the project metadata with the given owner and name from the database and return it and its primary key id
      * if the project exists and is readable by the given user account.
      *
      * @param currentUser
      *   The user account that is requesting access to the project or None for a guest user.
      * @param projectOwnerName
      *   The name of the account that owns the project.
      * @param projectName
      *   The name of the project. A project name must start with a letter or number and must contain only alphanumeric
      *   ASCII characters as well as minus or underscore signs. It must be between 2 and 64 characters long.
      * @return
      *   An option to a tuple holding the [[Project]] and its primary key id.
      */
    private def loadProject(
        currentUser: Option[Account]
    )(projectOwnerName: ProjectOwnerName, projectName: ProjectName): F[Option[(Project, ProjectId)]] =
        for {
            owner <- projectRepo.findProjectOwner(projectOwnerName)
            loadedRepo <- owner match {
                case None => Sync[F].pure(None)
                case Some(owner) =>
                    (
                        projectRepo.findProject(owner, projectName),
                        projectRepo.findProjectId(owner, projectName)
                    ).mapN {
                        case (Some(project), Some(projectId)) => (project, projectId).some
                        case _                                => None
                    }
            }
            // TODO: Replace with whatever we implement as proper permission model. ;-)
            projectAndId = currentUser match {
                case None => loadedRepo.filter(tuple => tuple._1.isPrivate === false)
                case Some(user) =>
                    loadedRepo.filter(tuple =>
                        tuple._1.isPrivate === false || tuple._1.owner.uid === ProjectOwnerId.fromUserId(user.uid)
                    )
            }
        } yield projectAndId

    /** Logic for rendering a detail page for a single ticket.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of tickets shall be rendered.
      * @param projectOwnerName
      *   The username of the account who owns the project.
      * @param projectName
      *   The name of the project.
      * @param ticketNumber
      *   The number of the ticket that shall be rendered.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    private def doShowTicket(csrf: Option[CsrfToken])(user: Option[Account])(projectOwnerName: Username)(
        projectName: ProjectName
    )(ticketNumber: TicketNumber): F[Response[F]] =
        for {
            _ <- Sync[F].delay(log.debug(s"doShowTicket: $csrf, $user, $projectOwnerName, $projectName, $ticketNumber"))
            language     <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            projectAndId <- loadProject(user)(projectOwnerName, projectName)
            ticket       <- projectAndId.traverse(tuple => ticketRepo.findTicket(tuple._2)(ticketNumber))
            resp <- (projectAndId, ticket.getOrElse(None)) match {
                case (Some((project, projectId)), Some(ticket)) =>
                    for {
                        labels     <- ticketRepo.loadLabels(projectId)(ticket.number).compile.toList
                        milestones <- ticketRepo.loadMilestones(projectId)(ticket.number).compile.toList
                        projectBaseUri <- Sync[F].delay(
                            linkConfig.createFullUri(
                                Uri(path =
                                    Uri.Path(
                                        Vector(
                                            Uri.Path.Segment(s"~$projectOwnerName"),
                                            Uri.Path.Segment(projectName.toString)
                                        )
                                    )
                                )
                            )
                        )
                        renderedTicketContent <- Sync[F].delay(
                            ticket.content.map(_.toString).map(MarkdownRenderer.render)
                        )
                        resp <- Ok(
                            views.html.showTicket(lang = language)(
                                projectBaseUri.addSegment("tickets"),
                                csrf,
                                linkToHubService,
                                labels,
                                milestones,
                                ticket,
                                renderedTicketContent,
                                projectBaseUri,
                                ticket.title.toString.some,
                                user,
                                project
                            )
                        )
                    } yield resp
                case _ => NotFound()
            }
        } yield resp

    /** Logic for rendering a list of all tickets for a project and optionally management functionality.
      *
      * @param filter
      *   An optional ticket filter containing possible values which will be used to filter the list of tickets.
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of tickets shall be rendered.
      * @param projectOwnerName
      *   The username of the account who owns the project.
      * @param projectName
      *   The name of the project.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    private def doShowTickets(filter: Option[TicketFilter])(csrf: Option[CsrfToken])(user: Option[Account])(
        projectOwnerName: Username
    )(projectName: ProjectName): F[Response[F]] =
        for {
            _ <- Sync[F].delay(log.debug(s"doShowTickets: $csrf, $user, $projectOwnerName, $projectName, $filter"))
            language     <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            projectAndId <- loadProject(user)(projectOwnerName, projectName)
            resp <- projectAndId match {
                case Some((project, projectId)) =>
                    for {
                        tickets <- ticketRepo.allTickets(filter)(projectId).compile.toList
                        projectBaseUri <- Sync[F].delay(
                            linkConfig.createFullUri(
                                Uri(path =
                                    Uri.Path(
                                        Vector(
                                            Uri.Path.Segment(s"~$projectOwnerName"),
                                            Uri.Path.Segment(projectName.toString)
                                        )
                                    )
                                )
                            )
                        )
                        resp <- Ok(
                            views.html.showTickets(lang = language)(
                                projectBaseUri.addSegment("tickets"),
                                csrf,
                                linkToHubService,
                                tickets,
                                filter,
                                projectBaseUri,
                                "Manage your project tickets.".some,
                                user,
                                project
                            )
                        )
                    } yield resp
                case _ => NotFound("Ticket project not found!")
            }
        } yield resp

    private val addTicket: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "tickets" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                val response =
                    for {
                        csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                        language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                        projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                        resp <- projectAndId match {
                            case Some((project, projectId)) =>
                                for {
                                    _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                        new Error("Only maintainers may add tickets!")
                                    )
                                    labels     <- labelRepo.allLabels(projectId).compile.toList
                                    milestones <- milestoneRepo.allMilestones(projectId).compile.toList
                                    formData   <- Sync[F].delay(urlForm.values)
                                    form       <- Sync[F].delay(TicketForm.validate(formData))
                                    projectBaseUri <- Sync[F].delay(
                                        linkConfig.createFullUri(
                                            Uri(path =
                                                Uri.Path(
                                                    Vector(
                                                        Uri.Path.Segment(s"~$projectOwnerName"),
                                                        Uri.Path.Segment(projectName.toString)
                                                    )
                                                )
                                            )
                                        )
                                    )
                                    resp <- form match {
                                        case Validated.Invalid(errors) =>
                                            BadRequest(
                                                views.html.createTicket(lang = language)(
                                                    projectBaseUri.addSegment("tickets"),
                                                    csrf,
                                                    linkToHubService,
                                                    labels,
                                                    milestones,
                                                    projectBaseUri,
                                                    "Create a new ticket.".some,
                                                    user.some,
                                                    project
                                                )(formData.withDefaultValue(Chain.empty), FormErrors.fromNec(errors))
                                            )
                                        case Validated.Valid(ticketData) =>
                                            for {
                                                timestamp <- Sync[F].delay(OffsetDateTime.now(ZoneOffset.UTC))
                                                ticketLabels <- ticketData.labels.traverse(name =>
                                                    labelRepo.findLabel(projectId)(name)
                                                )
                                                ticketMilestones <- ticketData.milestones.traverse(title =>
                                                    milestoneRepo.findMilestone(projectId)(title)
                                                )
                                                number <- projectRepo.incrementNextTicketNumber(projectId)
                                                ticket <- Sync[F].delay(
                                                    Ticket(
                                                        number = number,
                                                        title = ticketData.title,
                                                        content = ticketData.content,
                                                        status = ticketData.status,
                                                        resolution = ticketData.resolution,
                                                        submitter = Submitter(
                                                            SubmitterId(user.uid.toUUID),
                                                            SubmitterName(user.name.toString)
                                                        ).some,
                                                        createdAt = timestamp,
                                                        updatedAt = timestamp
                                                    )
                                                )
                                                _ <- ticketRepo.createTicket(projectId)(ticket)
                                                _ <- ticketLabels.flatten.traverse(
                                                    ticketRepo.addLabel(projectId)(number)
                                                )
                                                _ <- ticketMilestones.flatten.traverse(
                                                    ticketRepo.addMilestone(projectId)(number)
                                                )
                                                resp <- SeeOther(Location(projectBaseUri.addSegment("tickets")))
                                            } yield resp
                                    }
                                } yield resp
                            case _ => NotFound()
                        }
                    } yield resp
            response.recoverWith { error =>
                log.error("Internal Server Error", error)
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    resp <- InternalServerError(views.html.errors.internalServerError(lang = language)(csrf, user.some))
                } yield resp
            }
            }
    }

    private val editTicket: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "tickets" / TicketNumberPathParameter(ticketNumber) as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                val response =
                    for {
                        csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                        language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                        projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                        ticket       <- projectAndId.traverse(tuple => ticketRepo.findTicket(tuple._2)(ticketNumber))
                        resp <- (projectAndId, ticket.getOrElse(None)) match {
                            case (Some((project, projectId)), Some(ticket)) =>
                                for {
                                    _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                        new Error("Only maintainers may edit tickets!")
                                    )
                                    labels     <- labelRepo.allLabels(projectId).compile.toList
                                    milestones <- milestoneRepo.allMilestones(projectId).compile.toList
                                    formData   <- Sync[F].delay(urlForm.values)
                                    form       <- Sync[F].delay(TicketForm.validate(formData))
                                    projectBaseUri <- Sync[F].delay(
                                        linkConfig.createFullUri(
                                            Uri(path =
                                                Uri.Path(
                                                    Vector(
                                                        Uri.Path.Segment(s"~$projectOwnerName"),
                                                        Uri.Path.Segment(projectName.toString)
                                                    )
                                                )
                                            )
                                        )
                                    )
                                    resp <- form match {
                                        case Validated.Invalid(errors) =>
                                            BadRequest(
                                                views.html.createTicket(lang = language)(
                                                    projectBaseUri.addSegment("tickets"),
                                                    csrf,
                                                    linkToHubService,
                                                    labels,
                                                    milestones,
                                                    projectBaseUri,
                                                    "Create a new ticket.".some,
                                                    user.some,
                                                    project
                                                )(formData.withDefaultValue(Chain.empty), FormErrors.fromNec(errors))
                                            )
                                        case Validated.Valid(ticketData) =>
                                            for {
                                                timestamp <- Sync[F].delay(OffsetDateTime.now(ZoneOffset.UTC))
                                                oldLabels <- ticketRepo
                                                    .loadLabels(projectId)(ticket.number)
                                                    .compile
                                                    .toList
                                                oldMilestones <- ticketRepo
                                                    .loadMilestones(projectId)(ticket.number)
                                                    .compile
                                                    .toList
                                                ticketLabels <- ticketData.labels.traverse(name =>
                                                    labelRepo.findLabel(projectId)(name)
                                                )
                                                ticketMilestones <- ticketData.milestones.traverse(title =>
                                                    milestoneRepo.findMilestone(projectId)(title)
                                                )
                                                labelsToCreate     = ticketLabels.flatten.diff(oldLabels)
                                                labelsToRemove     = oldLabels.diff(ticketLabels.flatten)
                                                milestonesToCreate = ticketMilestones.flatten.diff(oldMilestones)
                                                milestonesToRemove = oldMilestones.diff(ticketMilestones.flatten)
                                                updatedTicket =
                                                    ticket.copy(
                                                        title = ticketData.title,
                                                        content = ticketData.content,
                                                        status = ticketData.status,
                                                        resolution = ticketData.resolution,
                                                        createdAt = ticket.createdAt,
                                                        updatedAt = timestamp
                                                    )
                                                _ <- ticketRepo.updateTicket(projectId)(updatedTicket)
                                                _ <- labelsToRemove.traverse(
                                                    ticketRepo.removeLabel(projectId)(updatedTicket)
                                                )
                                                _ <- milestonesToRemove.traverse(
                                                    ticketRepo.removeMilestone(projectId)(updatedTicket)
                                                )
                                                _ <- labelsToCreate.traverse(
                                                    ticketRepo.addLabel(projectId)(ticket.number)
                                                )
                                                _ <- milestonesToCreate.traverse(
                                                    ticketRepo.addMilestone(projectId)(ticket.number)
                                                )
                                                resp <- SeeOther(Location(projectBaseUri.addSegment("tickets")))
                                            } yield resp
                                    }
                                } yield resp
                            case _ => NotFound()
                        }
                    } yield resp
            response.recoverWith { error =>
                log.error("Internal Server Error", error)
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    resp <- InternalServerError(views.html.errors.internalServerError(lang = language)(csrf, user.some))
                } yield resp
            }
            }
    }

    private val showCreateTicketPage: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "newTicket" as user =>
            val response =
                for {
                    csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                    language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                    resp <- projectAndId match {
                        case Some((project, projectId)) =>
                            for {
                                labels     <- labelRepo.allLabels(projectId).compile.toList
                                milestones <- milestoneRepo.allMilestones(projectId).compile.toList
                                projectBaseUri <- Sync[F].delay(
                                    linkConfig.createFullUri(
                                        Uri(path =
                                            Uri.Path(
                                                Vector(
                                                    Uri.Path.Segment(s"~$projectOwnerName"),
                                                    Uri.Path.Segment(projectName.toString)
                                                )
                                            )
                                        )
                                    )
                                )
                                resp <- Ok(
                                    views.html.createTicket(lang = language)(
                                        projectBaseUri.addSegment("tickets"),
                                        csrf,
                                        linkToHubService,
                                        labels,
                                        milestones,
                                        projectBaseUri,
                                        "Create a new ticket.".some,
                                        user.some,
                                        project
                                    )()
                                )
                            } yield resp
                        case _ => NotFound()
                    }
                } yield resp
            response.recoverWith { error =>
                log.error("Internal Server Error", error)
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    resp <- InternalServerError(views.html.errors.internalServerError(lang = language)(csrf, user.some))
                } yield resp
            }
    }

    private val showEditTicketPage: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "tickets" / TicketNumberPathParameter(ticketNumber) / "edit" as user =>
            val response =
                for {
                    csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                    language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                    ticket       <- projectAndId.traverse(tuple => ticketRepo.findTicket(tuple._2)(ticketNumber))
                    resp <- (projectAndId, ticket.getOrElse(None)) match {
                        case (Some((project, projectId)), Some(ticket)) =>
                            for {
                                ticketLabels <- ticketRepo
                                    .loadLabels(projectId)(ticket.number)
                                    .map(_.name)
                                    .compile
                                    .toList
                                ticketMilestones <- ticketRepo
                                    .loadMilestones(projectId)(ticket.number)
                                    .map(_.title)
                                    .compile
                                    .toList
                                form     <- Sync[F].delay(TicketForm.fromTicket(ticketLabels)(ticketMilestones)(ticket))
                                formData <- Sync[F].delay(form.toMap)
                                labels   <- labelRepo.allLabels(projectId).compile.toList
                                milestones <- milestoneRepo.allMilestones(projectId).compile.toList
                                projectBaseUri <- Sync[F].delay(
                                    linkConfig.createFullUri(
                                        Uri(path =
                                            Uri.Path(
                                                Vector(
                                                    Uri.Path.Segment(s"~$projectOwnerName"),
                                                    Uri.Path.Segment(projectName.toString)
                                                )
                                            )
                                        )
                                    )
                                )
                                resp <- Ok(
                                    views.html.editTicket(lang = language)(
                                        projectBaseUri.addSegment("tickets"),
                                        csrf,
                                        linkToHubService,
                                        labels,
                                        milestones,
                                        projectBaseUri,
                                        ticket.number,
                                        s"Edit ticket ${ticket.number}".some,
                                        user.some,
                                        project
                                    )(formData.withDefaultValue(Chain.empty), FormErrors.empty)
                                )
                            } yield resp
                        case _ => NotFound()
                    }
                } yield resp
            response.recoverWith { error =>
                log.error("Internal Server Error", error)
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    resp <- InternalServerError(views.html.errors.internalServerError(lang = language)(csrf, user.some))
                } yield resp
            }
    }

    private val showTicketPage: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "tickets" / TicketNumberPathParameter(ticketNumber) as user =>
            val response =
                for {
                    csrf <- Sync[F].delay(ar.req.getCsrfToken)
                    resp <- doShowTicket(csrf)(user.some)(projectOwnerName)(projectName)(ticketNumber)
                } yield resp
            response.recoverWith { error =>
                log.error("Internal Server Error", error)
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    resp <- InternalServerError(views.html.errors.internalServerError(lang = language)(csrf, user.some))
                } yield resp
            }
    }

    private val showTicketPageForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "tickets" / TicketNumberPathParameter(ticketNumber) =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowTicket(csrf)(None)(projectOwnerName)(projectName)(ticketNumber)
            } yield resp
    }

    private val showTicketsPage: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "tickets" :? TicketFilter.OptionalUrlParameter(maybeFilter) as user =>
            val response =
                for {
                    csrf <- Sync[F].delay(ar.req.getCsrfToken)
                    resp <- doShowTickets(maybeFilter)(csrf)(user.some)(projectOwnerName)(projectName)
                } yield resp
            response.recoverWith { error =>
                log.error("Internal Server Error", error)
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    resp <- InternalServerError(views.html.errors.internalServerError(lang = language)(csrf, user.some))
                } yield resp
            }
    }

    private val showTicketsForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "tickets" :? TicketFilter.OptionalUrlParameter(maybeFilter) =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowTickets(maybeFilter)(csrf)(None)(projectOwnerName)(projectName)
            } yield resp
    }

    val protectedRoutes =
        addTicket <+> editTicket <+> showCreateTicketPage <+> showEditTicketPage <+> showTicketPage <+> showTicketsPage

    val routes = showTicketPageForGuests <+> showTicketsForGuests

}
