/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets.config

import com.comcast.ip4s.*
import de.smederee.html.ExternalUrlConfiguration
import org.http4s.Uri
import pureconfig.*

/** Configuration regarding the integration with the hub service.
  *
  * @param baseUri
  *   The base URI used to build links to the hub service.
  */
final case class HubIntegrationConfiguration(baseUri: Uri)

object HubIntegrationConfiguration {
    given ConfigReader[Uri] = ConfigReader.fromStringOpt(s => Uri.fromString(s).toOption)
    given ConfigReader[HubIntegrationConfiguration] =
        ConfigReader.forProduct1("base-uri")(HubIntegrationConfiguration.apply)
}

/** Wrapper class for the confiuration of the Smederee tickets module.
  *
  * @param database
  *   The configuration needed to access the database.
  * @param external
  *   Configuration regarding support for generating "external urls" which is usually needed if the service runs behind
  *   a reverse proxy.
  * @param hub
  *   Configuration regarding the integration with the hub service.
  */
final case class SmedereeTicketsConfiguration(
    database: DatabaseConfig,
    external: ExternalUrlConfiguration,
    hub: HubIntegrationConfiguration
)

object SmedereeTicketsConfiguration {
    val location: ConfigurationPath = ConfigurationPath("tickets")

    given ConfigReader[Host]       = ConfigReader.fromStringOpt[Host](Host.fromString)
    given ConfigReader[Port]       = ConfigReader.fromStringOpt[Port](Port.fromString)
    given ConfigReader[Uri]        = ConfigReader.fromStringOpt(s => Uri.fromString(s).toOption)
    given ConfigReader[Uri.Scheme] = ConfigReader.fromStringOpt(s => Uri.Scheme.fromString(s).toOption)

    given ConfigReader[ExternalUrlConfiguration] =
        ConfigReader.forProduct4("host", "path", "port", "scheme")(ExternalUrlConfiguration.apply)

    given ConfigReader[SmedereeTicketsConfiguration] =
        ConfigReader.forProduct3("database", "external", "hub-integration")(
            SmedereeTicketsConfiguration.apply
        )
}
