/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.*

import cats.effect.*
import cats.syntax.all.*
import de.smederee.TestTags.*
import de.smederee.hub.Generators.*
import de.smederee.hub.config.*
import de.smederee.security.*
import doobie.*
import org.http4s.*

final class AuthenticationMiddlewareTest extends BaseSpec with AuthenticationMiddleware {
    test("extractSessionId must return the session id".tag(NeedsDatabase)) {
        (genSignAndValidate.sample, genSessionId.sample) match {
            case (Some(signAndValidate), Some(sessionId)) =>
                val clock = java.time.Clock.systemUTC
                val token = signAndValidate.signToken(sessionId.toString)(clock.millis.toString)
                val request = Request[IO](method = Method.GET)
                    .addCookie(RequestCookie(Constants.authenticationCookieName.toString, token.toString))
                val test = for {
                    id <- extractSessionId[IO](request, signAndValidate)
                } yield id
                test.map { result =>
                    assert(result === sessionId)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test(
        "resolveUser must return the account if session id and account exist and the session has not reached absolute timeout"
            .tag(NeedsDatabase)
    ) {
        (genFiniteDuration.sample, genValidSession.sample, genValidAccount.sample) match {
            case (Some(duration), Some(s), Some(account)) =>
                val createdAt = OffsetDateTime.now(ZoneOffset.UTC)
                val session   = s.copy(uid = account.uid, createdAt = createdAt)
                val timeouts  = AuthenticationTimeouts(duration, duration, duration)
                val dbConfig  = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieAuthenticationRepository[IO](tx)
                val test = for {
                    _    <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    _    <- createUserSession(session)
                    user <- resolveUser[IO](repo)(timeouts).run(session.id)
                } yield user
                test.map { maybeUser =>
                    assert(clue(maybeUser) === clue(Option(account)))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test(
        "resolveUser must return None if session id and account exist but the session has reached absolute timeout".tag(
            NeedsDatabase
        )
    ) {
        (genFiniteDuration.sample, genValidSession.sample, genValidAccount.sample) match {
            case (Some(duration), Some(s), Some(account)) =>
                val createdAt = OffsetDateTime.now(ZoneOffset.UTC).minusSeconds(duration.toSeconds + 5L)
                val session   = s.copy(uid = account.uid, createdAt = createdAt)
                val timeouts  = AuthenticationTimeouts(duration, duration, duration)
                val dbConfig  = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieAuthenticationRepository[IO](tx)
                val test = for {
                    _    <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    _    <- createUserSession(session)
                    user <- resolveUser[IO](repo)(timeouts).run(session.id)
                } yield user
                test.map { maybeUser =>
                    assert(clue(maybeUser) === clue(Option(account)))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("resolveUser must return None if no session exists".tag(NeedsDatabase)) {
        (genFiniteDuration.sample, genValidSession.sample, genValidAccount.sample) match {
            case (Some(duration), Some(s), Some(account)) =>
                val session  = s.copy(uid = account.uid)
                val timeouts = AuthenticationTimeouts(duration, duration, duration)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieAuthenticationRepository[IO](tx)
                val test = for {
                    _    <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    user <- resolveUser[IO](repo)(timeouts).run(session.id)
                } yield user
                test.map { maybeUser =>
                    assert(clue(maybeUser) === None)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

}
