/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.nio.charset.StandardCharsets

import cats.effect.*
import cats.syntax.all.*
import com.typesafe.config.ConfigFactory
import de.smederee.hub.Generators.*
import de.smederee.hub.config.*
import de.smederee.hub.forms.*
import de.smederee.hub.forms.types.*
import de.smederee.security.SignAndValidate
import de.smederee.security.SignedToken
import org.http4s.*
import org.http4s.implicits.*
import org.http4s.server.*
import pureconfig.ConfigSource

import munit.*

class AuthenticationRoutesTest extends CatsEffectSuite {
    val loginPath         = uri"/login"
    private val resetPath = uri"/forgot-password"

    protected final val configuration: SmedereeHubConfig =
        ConfigSource
            .fromConfig(ConfigFactory.load(getClass.getClassLoader))
            .at(SmedereeHubConfig.location)
            .loadOrThrow[SmedereeHubConfig]

    test("GET /login must return the login form for guest users") {
        val expectedHtml =
            views.html.login()(loginPath, None, resetPath, title = "Smederee - Login to your account".some)()

        val clock                = java.time.Clock.systemUTC
        val authenticationConfig = configuration.service.authentication
        val externalConfig       = configuration.service.external
        val repo                 = new TestAuthenticationRepository[IO](List.empty, List.empty)
        val signAndValidate      = SignAndValidate(configuration.service.authentication.cookieSecret)

        def service: HttpRoutes[IO] =
            Router(
                "/" -> new AuthenticationRoutes[IO](
                    clock,
                    authenticationConfig,
                    externalConfig,
                    repo,
                    signAndValidate
                ).routes
            )

        def request = Request[IO](method = Method.GET, uri = loginPath)

        def response: IO[Response[IO]] = service.orNotFound.run(request)

        val test = for {
            result <- response
            body   <- result.as[String]
        } yield (result, body)

        test.map { output =>
            val (result, body) = output
            assertEquals(result.status, Status.Ok)
            assertEquals(body, expectedHtml.toString)
        }
    }

    test("POST /login must return 415 - Unsupported Media Type if the request is malformed") {
        val clock                = java.time.Clock.systemUTC
        val authenticationConfig = configuration.service.authentication
        val externalConfig       = configuration.service.external
        val repo                 = new TestAuthenticationRepository[IO](List.empty, List.empty)
        val signAndValidate      = SignAndValidate(configuration.service.authentication.cookieSecret)

        def service: HttpRoutes[IO] =
            Router(
                "/" -> new AuthenticationRoutes[IO](
                    clock,
                    authenticationConfig,
                    externalConfig,
                    repo,
                    signAndValidate
                ).routes
            )

        val payload = "This is not the request body you are looking for.".getBytes(StandardCharsets.UTF_8)
        def request = Request[IO](method = Method.POST, uri = loginPath).withEntity(payload)

        def response: IO[Response[IO]] = service.orNotFound.run(request)

        val test = for {
            result <- response
            body   <- result.as[String]
        } yield (result, body)

        test.map { output =>
            val (result, body) = output
            assertEquals(result.status, Status.UnsupportedMediaType)
            assertEquals(
                body,
                "Media type supplied in Content-Type header is not supported. Expected one of the following media ranges: application/*"
            )
        }
    }

    test("POST /login must return 400 - Bad Request if the form data is invalid") {
        val expectedErrors = LoginForm.validate(Map.empty).leftMap(FormErrors.fromNec).swap.toOption.get
        val expectedHtml =
            views.html.login()(loginPath, None, resetPath, title = "Smederee - Login to your account".some)(formErrors =
                expectedErrors
            )

        val clock                = java.time.Clock.systemUTC
        val authenticationConfig = configuration.service.authentication
        val externalConfig       = configuration.service.external
        val repo                 = new TestAuthenticationRepository[IO](List.empty, List.empty)
        val signAndValidate      = SignAndValidate(configuration.service.authentication.cookieSecret)

        def service: HttpRoutes[IO] =
            Router(
                "/" -> new AuthenticationRoutes[IO](
                    clock,
                    authenticationConfig,
                    externalConfig,
                    repo,
                    signAndValidate
                ).routes
            )

        val payload = UrlForm.empty
        def request = Request[IO](method = Method.POST, uri = loginPath).withEntity(payload)

        def response: IO[Response[IO]] = service.orNotFound.run(request)

        val test = for {
            result <- response
            body   <- result.as[String]
        } yield (result, body)

        test.map { output =>
            val (result, body) = output
            assertEquals(result.status, Status.BadRequest)
            assertEquals(body, expectedHtml.toString)
        }
    }

    test("POST /login must return 400 - Bad Request AND increase failed attempts if the credentials are invalid") {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val clock                = java.time.Clock.systemUTC
                val authenticationConfig = configuration.service.authentication
                val externalConfig       = configuration.service.external
                val repo                 = new TestAuthenticationRepository[IO](List(account), List.empty)
                val signAndValidate      = SignAndValidate(configuration.service.authentication.cookieSecret)

                def service: HttpRoutes[IO] =
                    Router(
                        "/" -> new AuthenticationRoutes[IO](
                            clock,
                            authenticationConfig,
                            externalConfig,
                            repo,
                            signAndValidate
                        ).routes
                    )

                val payload = UrlForm.empty
                    .updateFormField(LoginForm.fieldName.toString, account.name.toString)
                    .updateFormField(LoginForm.fieldPassword.toString, "doesn't matter")
                def request = Request[IO](method = Method.POST, uri = loginPath).withEntity(payload)

                val expectedHtml =
                    views.html.login()(loginPath, None, resetPath, title = "Smederee - Login to your account".some)(
                        formData = Map(LoginForm.fieldName.toString -> account.name.toString),
                        Map(LoginForm.fieldGlobal -> List(FormFieldError("Invalid credentials!")))
                    )

                def response: IO[Response[IO]] = service.orNotFound.run(request)

                val test = for {
                    result <- response
                    body   <- result.as[String]
                    failed <- repo.getFailed
                } yield (result, body, failed)

                test.map { output =>
                    val (result, body, failed) = output
                    assertEquals(result.status, Status.BadRequest)
                    assertEquals(failed.find(_._1 === account.uid), Option((account.uid, 1)))
                    assertEquals(body, expectedHtml.toString)
                }
        }
    }

    test(
        "POST /login must return 400 - Bad Request AND lock the account if the credentials are invalid and failed attempts are too high"
    ) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val clock                = java.time.Clock.systemUTC
                val authenticationConfig = configuration.service.authentication
                val externalConfig       = configuration.service.external
                val repo                 = new TestAuthenticationRepository[IO](List(account), List.empty)
                val signAndValidate      = SignAndValidate(configuration.service.authentication.cookieSecret)

                for (_ <- 0 to authenticationConfig.lockAfter.toInt) yield repo.incrementFailedAttempts(account.uid)

                def service: HttpRoutes[IO] =
                    Router(
                        "/" -> new AuthenticationRoutes[IO](
                            clock,
                            authenticationConfig,
                            externalConfig,
                            repo,
                            signAndValidate
                        ).routes
                    )

                val payload = UrlForm.empty
                    .updateFormField(LoginForm.fieldName.toString, account.name.toString)
                    .updateFormField(LoginForm.fieldPassword.toString, "doesn't matter")
                def request = Request[IO](method = Method.POST, uri = loginPath).withEntity(payload)

                val expectedHtml =
                    views.html.login()(loginPath, None, resetPath, title = "Smederee - Login to your account".some)(
                        formData = Map(LoginForm.fieldName.toString -> account.name.toString),
                        Map(LoginForm.fieldGlobal -> List(FormFieldError("Invalid credentials!")))
                    )

                def response: IO[Response[IO]] = service.orNotFound.run(request)

                val test = for {
                    result <- response
                    body   <- result.as[String]
                    locked <- repo.getLocked
                } yield (result, body, locked)

                test.map { output =>
                    val (result, body, locked) = output
                    assertEquals(result.status, Status.BadRequest)
                    assert(locked.exists(_ === account.uid), s"Account was not locked! ($locked)")
                    assertEquals(body, expectedHtml.toString)
                }
        }
    }

    test("POST /login must return 400 - Bad Request if the credentials are valid but the account is locked") {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val clock                = java.time.Clock.systemUTC
                val authenticationConfig = configuration.service.authentication
                val externalConfig       = configuration.service.external
                // A locked account is not supposed to be returned by findAccount, so we leave the account list empty.
                val repo            = new TestAuthenticationRepository[IO](List.empty, List.empty)
                val signAndValidate = SignAndValidate(configuration.service.authentication.cookieSecret)

                def service: HttpRoutes[IO] =
                    Router(
                        "/" -> new AuthenticationRoutes[IO](
                            clock,
                            authenticationConfig,
                            externalConfig,
                            repo,
                            signAndValidate
                        ).routes
                    )

                val payload = UrlForm.empty
                    .updateFormField(LoginForm.fieldName.toString, account.name.toString)
                    .updateFormField(LoginForm.fieldPassword.toString, "doesn't matter")
                def request = Request[IO](method = Method.POST, uri = loginPath).withEntity(payload)

                val expectedHtml =
                    views.html.login()(loginPath, None, resetPath, title = "Smederee - Login to your account".some)(
                        formData = Map(LoginForm.fieldName.toString -> account.name.toString),
                        Map(LoginForm.fieldGlobal -> List(FormFieldError("Invalid credentials!")))
                    )

                def response: IO[Response[IO]] = service.orNotFound.run(request)

                val test = for {
                    result <- response
                    body   <- result.as[String]
                } yield (result, body)

                test.map { output =>
                    val (result, body) = output
                    assertEquals(result.status, Status.BadRequest)
                    assertEquals(body, expectedHtml.toString)
                }
        }
    }

    test("POST /login must return 303 - See Other AND set the authentication cookie if the credentials are valid") {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val clock                = java.time.Clock.systemUTC
                val authenticationConfig = configuration.service.authentication
                val externalConfig       = configuration.service.external
                val repo                 = new TestAuthenticationRepository[IO](List(account), List.empty)
                val signAndValidate      = SignAndValidate(configuration.service.authentication.cookieSecret)

                def service: HttpRoutes[IO] =
                    Router(
                        "/" -> new AuthenticationRoutes[IO](
                            clock,
                            authenticationConfig,
                            externalConfig,
                            repo,
                            signAndValidate
                        ).routes
                    )

                val payload = UrlForm.empty
                    .updateFormField(LoginForm.fieldName.toString, account.name.toString)
                    .updateFormField(LoginForm.fieldPassword.toString, repo.DefaultPassword)
                def request = Request[IO](method = Method.POST, uri = loginPath).withEntity(payload)

                def response: IO[Response[IO]] = service.orNotFound.run(request)

                val test = for {
                    result <- response
                    body   <- result.as[String]
                } yield (result, body)

                test.map { output =>
                    val (result, body) = output
                    assertEquals(result.status, Status.SeeOther)
                    assert(body.isEmpty, "Response body must be empty!")
                    result.cookies.find(_.name === Constants.authenticationCookieName.toString).map(_.content) match {
                        case None => fail("Authentication cookie not set!")
                        case Some(content) =>
                            assert(SignedToken.from(content).flatMap(signAndValidate.validate).nonEmpty)
                    }
                }
        }
    }
}
