/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.effect.*
import cats.syntax.all.*
import de.smederee.TestTags.*
import de.smederee.hub.Generators.*
import de.smederee.security.*
import doobie.*

final class DoobieSignupRepositoryTest extends BaseSpec {
    test("createAccount must create a new account".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieSignupRepository[IO](tx)
                val test = for {
                    w <- repo.createAccount(account, PasswordHash("I am not a password hash!"))
                    o <- loadAccount(account.uid)
                } yield (w, o)
                test.map { result =>
                    val (written, loadedAccount) = result
                    assert(written === 1, "1 database row must have been written!")
                    assert(loadedAccount === Option(account), "Saved account differs from expected one!")
                }
        }
    }

    test("createAccount must fail if the email already exists".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidAccount.sample) match {
            case (Some(a), Some(b)) =>
                val existingAccount = a
                val newAccount      = b.copy(email = a.email)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieSignupRepository[IO](tx)
                val test = for {
                    c <- createAccount(existingAccount, PasswordHash("I am not a password hash!"), None, None)
                    w <- repo.createAccount(newAccount, PasswordHash("I am not a password hash!"))
                } yield (c, w)
                test.attempt.map {
                    case Left(error) =>
                        assert(
                            error.getMessage.contains("accounts_unique_email"),
                            "Error must be triggered by accounts_unique_email constraint!"
                        )
                    case Right(_) => fail("Creating accounts with already used emails must fail!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("createAccount must fail if the username already exists".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidAccount.sample) match {
            case (Some(a), Some(b)) =>
                val existingAccount = a
                val newAccount      = b.copy(name = a.name)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieSignupRepository[IO](tx)
                val test = for {
                    c <- createAccount(existingAccount, PasswordHash("I am not a password hash!"), None, None)
                    w <- repo.createAccount(newAccount, PasswordHash("I am not a password hash!"))
                } yield (c, w)
                test.attempt.map {
                    case Left(error) =>
                        assert(
                            error.getMessage.contains("accounts_unique_name"),
                            "Error must be triggered by accounts_unique_name constraint!"
                        )
                    case Right(_) => fail("Creating accounts with already used names must fail!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findEmail must return an existing email".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieSignupRepository[IO](tx)
                val test = for {
                    c <- repo.createAccount(account, PasswordHash("I am not a password hash!"))
                    e <- repo.findEmail(account.email)
                } yield (c, e)
                test.map { result =>
                    val (created, email) = result
                    assert(created === 1, "Test account not created!")
                    assert(email === Option(account.email), "Expected email not found!")
                }
        }
    }

    test("findUsername must return an existing name".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieSignupRepository[IO](tx)
                val test = for {
                    c <- repo.createAccount(account, PasswordHash("I am not a password hash!"))
                    n <- repo.findUsername(account.name)
                } yield (c, n)
                test.map { result =>
                    val (created, name) = result
                    assert(created === 1, "Test account not created!")
                    assert(name === Option(account.name), "Expected name not found!")
                }
        }
    }
}
