/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.effect.*
import cats.syntax.all.*
import de.smederee.TestTags.*
import de.smederee.email.EmailAddress
import de.smederee.hub.Generators.*
import de.smederee.hub.VcsMetadataRepositoriesOrdering.*
import de.smederee.security.*
import doobie.*
import org.http4s.implicits.*

import scala.collection.immutable.Queue

final class DoobieVcsMetadataRepositoryTest extends BaseSpec {

    /** Find all forks of the original repository with the given ID.
      *
      * @param originalRepoId
      *   The unique ID of the original repo from which was forked.
      * @return
      *   A list of ID pairs (original repository id, forked repository id) which may be empty.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    @SuppressWarnings(Array("DisableSyntax.var", "DisableSyntax.while"))
    protected def findForks(originalRepoId: VcsRepositoryId): IO[Seq[(VcsRepositoryId, VcsRepositoryId)]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT original_repo, forked_repo FROM "hub"."forks" WHERE original_repo = ?"""
                    )
                )
                _      <- IO.delay(statement.setLong(1, originalRepoId.toLong))
                result <- IO.delay(statement.executeQuery)
                forks <- IO.delay {
                    var queue = Queue.empty[(VcsRepositoryId, VcsRepositoryId)]
                    while (result.next())
                        queue = queue :+ (VcsRepositoryId(result.getLong("original_repo")), VcsRepositoryId(
                            result.getLong("forked_repo")
                        ))
                    queue
                }
                _ <- IO.delay(statement.close())
            } yield forks.toList
        }

    test("createFork must work correctly".tag(NeedsDatabase)) {
        (genValidAccounts.suchThat(_.size > 4).sample, genValidVcsRepositories.suchThat(_.size > 4).sample) match {
            case (Some(accounts), Some(repositories)) =>
                val vcsRepositories = accounts.zip(repositories).map { tuple =>
                    val (account, repo) = tuple
                    repo.copy(owner = account.toVcsRepositoryOwner)
                }
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _ <- accounts.traverse(account =>
                        createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    )
                    written <- vcsRepositories.traverse(repo.createVcsRepository)
                    original <- vcsRepositories.headOption.traverse(vcsRepository =>
                        loadVcsRepositoryId(vcsRepository.owner.uid, vcsRepository.name)
                    )
                    toFork <- vcsRepositories.drop(1).take(5).traverse { vcsRepository =>
                        loadVcsRepositoryId(vcsRepository.owner.uid, vcsRepository.name)
                    }
                    forked <- (original, toFork) match {
                        case (Some(Some(originalId)), forkIds) =>
                            forkIds.flatten.traverse(id => repo.createFork(originalId, id))
                        case _ => IO.pure(List.empty)
                    }
                    foundForks <- original match {
                        case Some(Some(originalId)) => findForks(originalId)
                        case _                      => IO.pure(List.empty)
                    }
                } yield (written, forked, foundForks)
                test.map { result =>
                    val (written, forked, foundForks) = result
                    assert(written.sum === vcsRepositories.size, "Test repository data was not written to database!")
                    assert(
                        forked.sum === vcsRepositories.drop(1).take(5).size,
                        "Number of created forks does not match!"
                    )
                    assert(
                        foundForks.size === vcsRepositories.drop(1).take(5).size,
                        "Number of found forks does not match!"
                    )
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("createVcsRepository must create a repository entry".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepository.sample) match {
            case (Some(account), Some(repository)) =>
                val vcsRepository = repository.copy(owner = account.toVcsRepositoryOwner)
                val dbConfig      = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _       <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written <- repo.createVcsRepository(vcsRepository)
                } yield written
                test.map { written =>
                    assert(written === 1, "Creating a vcs repository must modify one database row!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("createVcsRepository must fail if the user does not exist".tag(NeedsDatabase)) {
        genValidVcsRepository.sample match {
            case Some(repository) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    written <- repo.createVcsRepository(repository)
                } yield written
                test.attempt.map(result => assert(result.isLeft))
            case _ => fail("Could not generate data samples!")
        }
    }

    test("createVcsRepository must fail if a repository with the same name exists".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepository.sample) match {
            case (Some(account), Some(repository)) =>
                val vcsRepository = repository.copy(owner = account.toVcsRepositoryOwner)
                val dbConfig      = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _       <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    _       <- repo.createVcsRepository(vcsRepository)
                    written <- repo.createVcsRepository(vcsRepository)
                } yield written
                test.attempt.map(result => assert(result.isLeft))
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findVcsRepository must return an existing repository".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepository.sample) match {
            case (Some(account), Some(repository)) =>
                val vcsRepository = repository.copy(owner = account.toVcsRepositoryOwner)
                val dbConfig      = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _         <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written   <- repo.createVcsRepository(vcsRepository)
                    foundRepo <- repo.findVcsRepository(vcsRepository.owner, vcsRepository.name)
                } yield (written, foundRepo)
                test.map { result =>
                    val (written, foundRepo) = result
                    assert(written === 1, "Test repository data was not written to database!")
                    foundRepo match {
                        case None       => fail("Repository was not found!")
                        case Some(repo) => assert(repo === vcsRepository)
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findVcsRepositoryBranches must return all branches".tag(NeedsDatabase)) {
        (genValidAccounts.suchThat(_.size > 4).sample, genValidVcsRepositories.suchThat(_.size > 4).sample) match {
            case (Some(accounts), Some(repositories)) =>
                val vcsRepositories = accounts.zip(repositories).map { tuple =>
                    val (account, repo) = tuple
                    repo.copy(owner = account.toVcsRepositoryOwner)
                }
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _ <- accounts.traverse(account =>
                        createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    )
                    written <- vcsRepositories.traverse(repo.createVcsRepository)
                    original <- vcsRepositories.headOption.traverse(vcsRepository =>
                        loadVcsRepositoryId(vcsRepository.owner.uid, vcsRepository.name)
                    )
                    toFork <- vcsRepositories.drop(1).take(5).traverse { vcsRepository =>
                        loadVcsRepositoryId(vcsRepository.owner.uid, vcsRepository.name)
                    }
                    forked <- (original, toFork) match {
                        case (Some(Some(originalId)), forkIds) =>
                            forkIds.flatten.traverse(id => repo.createFork(originalId, id))
                        case _ => IO.pure(List.empty)
                    }
                    foundForks <- original match {
                        case Some(Some(originalId)) => repo.findVcsRepositoryBranches(originalId).compile.toList
                        case _                      => IO.pure(List.empty)
                    }
                } yield (written, forked, foundForks)
                test.map { result =>
                    val (written, forked, foundForks) = result
                    assert(written.sum === vcsRepositories.size, "Test repository data was not written to database!")
                    assert(
                        forked.sum === vcsRepositories.drop(1).take(5).size,
                        "Number of created forks does not match!"
                    )
                    assert(
                        foundForks.size === vcsRepositories.drop(1).take(5).size,
                        "Number of found forks does not match!"
                    )
                    foundForks.zip(vcsRepositories.drop(1).take(5)).map { tuple =>
                        val ((ownerName, repoName), repo) = tuple
                        assertEquals(ownerName, repo.owner.name)
                        assertEquals(repoName, repo.name)
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("loadVcsRepositoryId must return the id of an existing repository".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepository.sample) match {
            case (Some(account), Some(repository)) =>
                val vcsRepository = repository.copy(owner = account.toVcsRepositoryOwner)
                val dbConfig      = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _           <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    writtenRows <- repo.createVcsRepository(vcsRepository)
                    writtenId   <- loadVcsRepositoryId(vcsRepository.owner.uid, vcsRepository.name)
                    foundId     <- repo.findVcsRepositoryId(vcsRepository.owner, vcsRepository.name)
                } yield (writtenRows, writtenId, foundId)
                test.map { result =>
                    val (written, writtenId, foundId) = result
                    assert(written === 1, "Test repository data was not written to database!")
                    assert(writtenId.nonEmpty)
                    assert(foundId.nonEmpty)
                    assert(writtenId === foundId)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findVcsRepositoryOwner must return the correct account".tag(NeedsDatabase)) {
        genValidAccounts.sample match {
            case Some(accounts) =>
                val expectedOwner = accounts(scala.util.Random.nextInt(accounts.size)).toVcsRepositoryOwner
                val dbConfig      = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    written <- accounts.traverse(account =>
                        createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    )
                    foundOwner <- repo.findVcsRepositoryOwner(expectedOwner.name)
                } yield (written, foundOwner)
                test.map { result =>
                    val (written, foundOwner) = result
                    assert(
                        written.filter(_ === 1).size === written.size,
                        "Not all test repository data was not written to database!"
                    )
                    foundOwner match {
                        case None        => fail("Vcs repository owner not found!")
                        case Some(owner) => assertEquals(owner, expectedOwner)
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findVcsRepositoryParentFork must return the parent repository if it exists".tag(NeedsDatabase)) {
        (genValidAccounts.suchThat(_.size > 4).sample, genValidVcsRepositories.suchThat(_.size > 4).sample) match {
            case (Some(accounts), Some(repositories)) =>
                val vcsRepositories = accounts.zip(repositories).map { tuple =>
                    val (account, repo) = tuple
                    repo.copy(owner = account.toVcsRepositoryOwner)
                }
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _ <- accounts.traverse(account =>
                        createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    )
                    written <- vcsRepositories.traverse(repo.createVcsRepository)
                    original <- vcsRepositories.headOption.traverse(vcsRepository =>
                        loadVcsRepositoryId(vcsRepository.owner.uid, vcsRepository.name)
                    )
                    toFork <- vcsRepositories.drop(1).take(5).traverse { vcsRepository =>
                        loadVcsRepositoryId(vcsRepository.owner.uid, vcsRepository.name)
                    }
                    forked <- (original, toFork) match {
                        case (Some(Some(originalId)), forkIds) =>
                            forkIds.flatten.traverse(id => repo.createFork(originalId, id))
                        case _ => IO.pure(List.empty)
                    }
                    foundParents <- vcsRepositories.drop(1).take(5).traverse { vcsRepository =>
                        repo.findVcsRepositoryParentFork(vcsRepository.owner, vcsRepository.name)
                    }
                } yield (written, forked, vcsRepositories.headOption, foundParents)
                test.map { result =>
                    val (written, forked, original, foundParents) = result
                    assert(written.sum === vcsRepositories.size, "Test repository data was not written to database!")
                    assert(
                        forked.sum === vcsRepositories.drop(1).take(5).size,
                        "Number of created forks does not match!"
                    )
                    assert(foundParents.forall(_ === original), "Parent vcs repository not matching!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("listAllRepositories must return only public repositories for guest users".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepositories.sample) match {
            case (Some(account), Some(repositories)) =>
                val vcsRepositories = repositories
                val accounts = vcsRepositories.map(repo =>
                    Account(
                        repo.owner.uid,
                        repo.owner.name,
                        EmailAddress(s"${repo.owner.name}@example.com"),
                        account.fullName,
                        validatedEmail = true,
                        None
                    )
                )
                val expectedRepoList = vcsRepositories.filter(_.isPrivate === false)
                val dbConfig         = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _ <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    _ <- accounts.traverse(account =>
                        createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    )
                    written    <- vcsRepositories.traverse(vcsRepository => repo.createVcsRepository(vcsRepository))
                    foundRepos <- repo.listAllRepositories(None)(NameAscending).compile.toList
                } yield (written, foundRepos)
                test.map { result =>
                    val (written, foundRepos) = result
                    assert(
                        written.filter(_ === 1).size === written.size,
                        "Not all test repository data was written to database!"
                    )
                    assertEquals(foundRepos.size, expectedRepoList.size)
                }

            case _ => fail("Could not generate data samples!")
        }
    }

    test("listAllRepositories must return only public repositories of others for any user".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepositories.sample) match {
            case (Some(account), Some(repositories)) =>
                val vcsRepositories = repositories
                val accounts = vcsRepositories.map(repo =>
                    Account(
                        repo.owner.uid,
                        repo.owner.name,
                        EmailAddress(s"${repo.owner.name}@example.com"),
                        account.fullName,
                        validatedEmail = true,
                        None
                    )
                )
                val expectedRepoList = vcsRepositories.filter(_.isPrivate === false)
                val dbConfig         = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _ <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    _ <- accounts.traverse(account =>
                        createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    )
                    written    <- vcsRepositories.traverse(vcsRepository => repo.createVcsRepository(vcsRepository))
                    foundRepos <- repo.listAllRepositories(account.some)(NameDescending).compile.toList
                } yield (written, foundRepos)
                test.map { result =>
                    val (written, foundRepos) = result
                    assert(
                        written.filter(_ === 1).size === written.size,
                        "Not all test repository data was written to database!"
                    )
                    assertEquals(foundRepos.size, expectedRepoList.size)
                }

            case _ => fail("Could not generate data samples!")
        }
    }

    test("listAllRepositories must include all private repositories of the user".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepositories.sample) match {
            case (Some(account), Some(repositories)) =>
                val privateRepos =
                    repositories.filter(_.isPrivate === true).map(_.copy(owner = account.toVcsRepositoryOwner))
                val publicRepos     = repositories.filter(_.isPrivate === false)
                val vcsRepositories = privateRepos ::: publicRepos
                val accounts = publicRepos.map(repo =>
                    Account(
                        repo.owner.uid,
                        repo.owner.name,
                        EmailAddress(s"${repo.owner.name}@example.com"),
                        account.fullName,
                        validatedEmail = true,
                        None
                    )
                )
                val expectedRepoList = vcsRepositories
                val dbConfig         = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _ <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    _ <- accounts.traverse(account =>
                        createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    )
                    written    <- vcsRepositories.traverse(vcsRepository => repo.createVcsRepository(vcsRepository))
                    foundRepos <- repo.listAllRepositories(account.some)(NameDescending).compile.toList
                } yield (written, foundRepos)
                test.map { result =>
                    val (written, foundRepos) = result
                    assert(
                        written.filter(_ === 1).size === written.size,
                        "Not all test repository data was written to database!"
                    )
                    assertEquals(foundRepos.size, expectedRepoList.size)
                }

            case _ => fail("Could not generate data samples!")
        }
    }

    test("listRepositories must return only public repositories for guest users".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepositories.sample) match {
            case (Some(account), Some(repositories)) =>
                val vcsRepositories  = repositories.map(_.copy(owner = account.toVcsRepositoryOwner))
                val expectedRepoList = vcsRepositories.filter(_.isPrivate === false).sortBy(_.name)
                val dbConfig         = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _          <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written    <- vcsRepositories.traverse(vcsRepository => repo.createVcsRepository(vcsRepository))
                    foundRepos <- repo.listRepositories(None)(account.toVcsRepositoryOwner).compile.toList
                } yield (written, foundRepos)
                test.map { result =>
                    val (written, foundRepos) = result
                    assert(
                        written.filter(_ === 1).size === written.size,
                        "Not all test repository data was written to database!"
                    )
                    assertEquals(foundRepos.size, expectedRepoList.size)
                    // We sort again because database sorting might differ slightly from code sorting.
                    assertEquals(foundRepos.sortBy(_.name), expectedRepoList)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("listRepositories must return all repositories for the owner".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepositories.sample) match {
            case (Some(account), Some(repositories)) =>
                val vcsRepositories  = repositories.map(_.copy(owner = account.toVcsRepositoryOwner))
                val expectedRepoList = vcsRepositories.sortBy(_.name)
                val dbConfig         = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _          <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written    <- vcsRepositories.traverse(vcsRepository => repo.createVcsRepository(vcsRepository))
                    foundRepos <- repo.listRepositories(account.some)(account.toVcsRepositoryOwner).compile.toList
                } yield (written, foundRepos)
                test.map { result =>
                    val (written, foundRepos) = result
                    assert(
                        written.filter(_ === 1).size === written.size,
                        "Not all test repository data was written to database!"
                    )
                    assertEquals(foundRepos.size, expectedRepoList.size)
                    // We sort again because database sorting might differ slightly from code sorting.
                    assertEquals(foundRepos.sortBy(_.name), expectedRepoList)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("listRepositories must return only public repositories for any user".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepositories.sample, genValidAccount.sample) match {
            case (Some(account), Some(repositories), Some(otherAccount)) =>
                val vcsRepositories  = repositories.map(_.copy(owner = account.toVcsRepositoryOwner))
                val expectedRepoList = vcsRepositories.filter(_.isPrivate === false).sortBy(_.name)
                val dbConfig         = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _          <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written    <- vcsRepositories.traverse(vcsRepository => repo.createVcsRepository(vcsRepository))
                    foundRepos <- repo.listRepositories(otherAccount.some)(account.toVcsRepositoryOwner).compile.toList
                } yield (written, foundRepos)
                test.map { result =>
                    val (written, foundRepos) = result
                    assert(
                        written.filter(_ === 1).size === written.size,
                        "Not all test repository data was written to database!"
                    )
                    assertEquals(foundRepos.size, expectedRepoList.size)
                    // We sort again because database sorting might differ slightly from code sorting.
                    assertEquals(foundRepos.sortBy(_.name), expectedRepoList)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("updateVcsRepository must update all columns correctly".tag(NeedsDatabase)) {
        (genValidAccount.sample, genValidVcsRepositories.sample) match {
            case (Some(account), Some(repository :: repositories)) =>
                val vcsRepositories = repository.copy(owner = account.toVcsRepositoryOwner) :: repositories.map(
                    _.copy(owner = account.toVcsRepositoryOwner)
                )
                val updatedRepo = repository
                    .copy(owner = account.toVcsRepositoryOwner)
                    .copy(
                        isPrivate = !repository.isPrivate,
                        description = Option(VcsRepositoryDescription("I am a description...")),
                        website = Option(uri"https://updated.example.com")
                    )
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieVcsMetadataRepository[IO](tx)
                val test = for {
                    _         <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written   <- vcsRepositories.traverse(vcsRepository => repo.createVcsRepository(vcsRepository))
                    updated   <- repo.updateVcsRepository(updatedRepo)
                    persisted <- repo.findVcsRepository(account.toVcsRepositoryOwner, repository.name)
                } yield (written, updated, persisted)
                test.map { result =>
                    val (written, updated, persisted) = result
                    assert(
                        written.filter(_ === 1).size === written.size,
                        "Not all test repository data was written to database!"
                    )
                    assert(updated === 1, "Repository was not updated in database!")
                    assert(persisted === Some(updatedRepo))
                }
            case _ => fail("Could not generate data samples!")
        }
    }
}
