/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.effect.*
import cats.syntax.all.*
import com.typesafe.config.ConfigFactory
import de.smederee.hub.Generators.*
import de.smederee.hub.config.*
import org.http4s.*
import org.http4s.implicits.*
import org.http4s.server.*
import pureconfig.ConfigSource

import munit.*

final class ResetPasswordRoutesTest extends CatsEffectSuite {
    private val resetChangePasswordPath = uri"/forgot-password/change-password"

    protected final val configuration: SmedereeHubConfig =
        ConfigSource
            .fromConfig(ConfigFactory.load(getClass.getClassLoader))
            .at(SmedereeHubConfig.location)
            .loadOrThrow[SmedereeHubConfig]

    test("GET forgot-password/change-password/TOKEN must return 200 and the change password form") {
        genValidAccount.sample match {
            case Some(account) =>
                val token             = ResetToken.generate
                val changePasswordUri = resetChangePasswordPath.addSegment(token.toString)
                val expectedHtml = views.html
                    .changePassword()(changePasswordUri, None, title = "Smederee - Change your password.".some, token)()

                val authenticationConfig = configuration.service.authentication
                val authenticationRepo   = new TestAuthenticationRepository[IO](List.empty, List.empty)
                val emailMiddleware      = new TestEmailMiddleware[IO]
                val externalConfig       = configuration.service.external
                val resetPasswordRepo    = new TestResetPasswordRepository[IO](account, token.some)

                def service: HttpRoutes[IO] =
                    Router(
                        "/" -> new ResetPasswordRoutes[IO](
                            authenticationConfig,
                            authenticationRepo,
                            emailMiddleware,
                            externalConfig,
                            resetPasswordRepo
                        ).routes
                    )

                def request                    = Request[IO](method = Method.GET, uri = changePasswordUri)
                def response: IO[Response[IO]] = service.orNotFound.run(request)

                val test = for {
                    result <- response
                    body   <- result.as[String]
                } yield (result, body)

                test.map { output =>
                    val (result, body) = output
                    assertEquals(result.status, Status.Ok)
                    assertEquals(body, expectedHtml.toString)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("GET forgot-password/change-password/TOKEN must return 404 for non existing token") {
        genValidAccount.sample match {
            case Some(account) =>
                val token             = ResetToken.generate
                val changePasswordUri = resetChangePasswordPath.addSegment(token.toString)

                val authenticationConfig = configuration.service.authentication
                val authenticationRepo   = new TestAuthenticationRepository[IO](List.empty, List.empty)
                val emailMiddleware      = new TestEmailMiddleware[IO]
                val externalConfig       = configuration.service.external
                val resetPasswordRepo    = new TestResetPasswordRepository[IO](account, None)

                def service: HttpRoutes[IO] =
                    Router(
                        "/" -> new ResetPasswordRoutes[IO](
                            authenticationConfig,
                            authenticationRepo,
                            emailMiddleware,
                            externalConfig,
                            resetPasswordRepo
                        ).routes
                    )

                def request                    = Request[IO](method = Method.GET, uri = changePasswordUri)
                def response: IO[Response[IO]] = service.orNotFound.run(request)

                val test = for {
                    result <- response
                    body   <- result.as[String]
                } yield (result, body)

                test.map { output =>
                    val (result, _) = output
                    assertEquals(result.status, Status.NotFound)
                }
            case _ => fail("Could not generate data samples!")
        }
    }
}
