/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.OffsetDateTime
import java.time.ZoneOffset
import java.time.ZonedDateTime

import de.smederee.hub.Generators.given
import de.smederee.hub.SessionHelpers.instances.*
import de.smederee.hub.config.AuthenticationTimeouts
import de.smederee.hub.config.Constants
import de.smederee.security.SignAndValidate
import org.http4s.HttpDate
import org.http4s.SameSite

import munit.*

import org.scalacheck.*
import org.scalacheck.Prop.*

import scala.concurrent.duration.*

final class SessionHelpersTest extends ScalaCheckSuite {

    property("Session.hasReachedAbsoluteTimeout must work correctly") {
        forAll { (absoluteTimeout: FiniteDuration, currentTime: OffsetDateTime, session: Session) =>
            val timeouts = AuthenticationTimeouts(absoluteTimeout, absoluteTimeout, absoluteTimeout)
            val expected = session.createdAt.compareTo(currentTime.minusSeconds(absoluteTimeout.toSeconds)) < 0
            assertEquals(session.hasReachedAbsoluteTimeout(timeouts)(currentTime), expected)
        }
    }

    property("SignedToken.toAuthenticationCookie must work correctly") {
        forAll {
            (
                expires: Option[FiniteDuration],
                nonce: String,
                secure: Boolean,
                signAndValidate: SignAndValidate,
                token: String
            ) =>
                val domain      = Option("example.com")
                val signedToken = signAndValidate.signToken(token)(nonce)
                val obtained    = signedToken.toAuthenticationCookie(domain)(expires)(secure)
                assertEquals(obtained.name, Constants.authenticationCookieName.toString)
                assertEquals(obtained.content, signedToken.toString)
                assertEquals(obtained.domain, domain)
                assertEquals(obtained.sameSite, Option(SameSite.Strict))
                assertEquals(obtained.secure, secure)
                assertEquals(obtained.httpOnly, true)
                // If an expires value was passed then the expires field in the cookie must be set!
                assertEquals(obtained.expires.nonEmpty, expires.nonEmpty)
                // The function uses `ZonedDateTime.now(ZoneOffset.UTC)` inside, therefore we can only check
                // if the expiration date stored in the cookie is at least "later" than the current datetime.
                (expires, obtained.expires) match {
                    case (Some(_), Some(expiration)) =>
                        HttpDate.fromZonedDateTime(ZonedDateTime.now(ZoneOffset.UTC)).foreach { httpDate =>
                            assert(expiration.compareTo(httpDate) > 0)
                        }
                    case _ => // We only check if the values are set.
                }
        }
    }

}
