/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.nio.charset.StandardCharsets

import cats.effect.*
import cats.syntax.all.*
import de.smederee.email.EmailAddress
import de.smederee.hub.*
import de.smederee.security.*
import fs2.Stream

/** An implementation of a [[AuthenticationRepository]] for testing purposes.
  *
  * @param accounts
  *   A list of accounts from which data might be returned.
  * @param sessions
  *   A list of sessions from which data might be returned.
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
@SuppressWarnings(Array("DisableSyntax.var"))
class TestAuthenticationRepository[F[_]: Sync](accounts: List[Account], sessions: List[Session])
    extends AuthenticationRepository[F] {
    val DefaultPassword = "My voice is my passport. Verify me."

    private var failed = Map.empty[UserId, Int]
    private var locked = List.empty[UserId]

    /** Return the map of user ids and related failed login attempts.
      *
      * @return
      *   A map of user ids pointing to an integer holding the number of failed login attempts.
      */
    def getFailed: F[Map[UserId, Int]] = Sync[F].delay(failed)

    /** Return the list of locked user ids.
      *
      * @return
      *   A list with user ids that are locked.
      */
    def getLocked: F[List[UserId]] = Sync[F].delay(locked)

    override def allAccounts(): Stream[F, Account] = Stream.emits(accounts)

    override def lockAccount(uid: UserId)(token: Option[UnlockToken]): F[Int] = {
        locked = uid :: locked
        Sync[F].pure(1)
    }

    override def findAccountByName(name: Username): F[Option[Account]] = Sync[F].pure(accounts.headOption)

    override def resetFailedAttempts(uid: UserId): F[Int] = {
        failed = failed.filterNot(_._1 === uid)
        Sync[F].pure(1)
    }

    override def deleteAllUserSessions(uid: UserId): F[Int] = Sync[F].pure(1)

    override def createUserSession(session: Session): F[Int] = Sync[F].pure(1)

    override def incrementFailedAttempts(uid: UserId): F[Int] = {
        failed = failed |+| Map(uid -> 1)
        Sync[F].pure(1)
    }

    override def findUserSession(id: SessionId): F[Option[Session]] = Sync[F].pure(sessions.headOption)

    override def findLockedAccount(name: Username)(token: Option[UnlockToken]): F[Option[Account]] =
        Sync[F].pure(accounts.headOption)

    override def findPasswordHashAndAttempts(uid: UserId): F[Option[(PasswordHash, Int)]] =
        Sync[F].delay(
            Option((Password(DefaultPassword.getBytes(StandardCharsets.UTF_8)).encode, failed.withDefaultValue(0)(uid)))
        )

    override def findAccount(uid: UserId): F[Option[Account]] = Sync[F].pure(accounts.headOption)

    override def findAccountByEmail(email: EmailAddress): F[Option[Account]] = Sync[F].pure(accounts.headOption)

    override def unlockAccount(uid: UserId): F[Int] = {
        locked = locked.filterNot(_ === uid)
        Sync[F].pure(1)
    }

    override def deleteUserSession(id: SessionId): F[Int] = Sync[F].pure(1)

}
