/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.OffsetDateTime

import cats.effect.*
import cats.syntax.all.*
import de.smederee.hub.*
import de.smederee.security.*

/** An implementation of a [[ResetPasswordRepository]] for testing purposes.
  *
  * @param account
  *   An account on which shall be operated.
  * @param token
  *   A reset token on which shall be operated.
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
@SuppressWarnings(Array("DisableSyntax.var"))
class TestResetPasswordRepository[F[_]: Sync](private val account: Account, token: Option[ResetToken])
    extends ResetPasswordRepository[F] {

    override def findByNameAndResetPasswordToken(name: Username, token: ResetToken): F[Option[Account]] =
        Sync[F].delay(
            this.token.find(_ === token).map(_ => account).filter(_.name === name)
        ) // Return account if token and name matches.

    override def setResetPasswordToken(uid: UserId)(token: ResetToken, tokenExpiration: OffsetDateTime): F[Int] =
        Sync[F].pure(1)

    override def findByResetPasswordToken(token: ResetToken): F[Option[Account]] =
        Sync[F].delay(this.token.find(_ === token).map(_ => account)) // Return account if token matches.

    override def removeResetPasswordExpirationDate(uid: UserId): F[Int] = Sync[F].pure(1)

    override def removeResetPasswordToken(uid: UserId): F[Int] = Sync[F].pure(1)

    override def setPassword(uid: UserId)(hash: PasswordHash): F[Int] = Sync[F].pure(1)
}
