/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.net.ServerSocket

import cats.effect.*
import cats.syntax.all.*
import com.comcast.ip4s.*
import com.typesafe.config.ConfigFactory
import de.smederee.email.EmailAddress
import de.smederee.i18n.LanguageCode
import de.smederee.security.UserId
import de.smederee.security.Username
import de.smederee.tickets.config.*
import org.flywaydb.core.Flyway
import pureconfig.*

import munit.*

import scala.annotation.nowarn

/** Base class for our integration test suites.
  *
  * It loads and provides the configuration as a resource suit fixture (loaded once for all tests within a suite) and
  * does initialise the test database for each suite. The latter means a possibly existing database with the name
  * configured **will be deleted**!
  */
abstract class BaseSpec extends CatsEffectSuite {
    protected final val configuration: SmedereeTicketsConfiguration =
        ConfigSource
            .fromConfig(ConfigFactory.load(getClass.getClassLoader))
            .at(SmedereeTicketsConfiguration.location)
            .loadOrThrow[SmedereeTicketsConfiguration]

    protected final val flyway: Flyway =
        DatabaseMigrator
            .configureFlyway(configuration.database.url, configuration.database.user, configuration.database.pass)
            .cleanDisabled(false)
            .load()

    /** Connect to the DBMS using the generic "template1" database which should always be present.
      *
      * @param dbConfig
      *   The database configuration.
      * @return
      *   The connection to the database ("template1").
      */
    private def connect(dbConfig: DatabaseConfig): IO[java.sql.Connection] =
        for {
            _        <- IO(Class.forName(dbConfig.driver))
            database <- IO(dbConfig.url.split("/").reverse.take(1).mkString)
            connection <- IO(
                java.sql.DriverManager
                    .getConnection(dbConfig.url.replace(database, "template1"), dbConfig.user, dbConfig.pass)
            )
        } yield connection

    @nowarn("msg=discarded non-Unit value.*")
    override def beforeAll(): Unit = {
        // Extract the database name from the URL.
        val database = configuration.database.url.split("/").reverse.take(1).mkString
        val db       = Resource.make(connect(configuration.database))(con => IO(con.close()))
        // Create the test database if it does not already exist.
        db.use { connection =>
            for {
                statement <- IO(connection.createStatement())
                exists <- IO(
                    statement.executeQuery(
                        s"""SELECT datname FROM pg_catalog.pg_database WHERE datname = '$database'"""
                    )
                )
                _ <- IO {
                    if (!exists.next())
                        statement.execute(s"""CREATE DATABASE "$database"""")
                }
                _ <- IO(exists.close)
                _ <- IO(statement.close)
            } yield ()
        }.unsafeRunSync()
    }

    override def afterAll(): Unit = {
        // Extract the database name from the URL.
        val database = configuration.database.url.split("/").reverse.take(1).mkString
        val db       = Resource.make(connect(configuration.database))(con => IO(con.close()))
        // Drop the test database after all tests have been run.
        db.use { connection =>
            for {
                statement <- IO(connection.createStatement())
                _         <- IO(statement.execute(s"""DROP DATABASE IF EXISTS $database"""))
                _         <- IO(statement.close)
            } yield ()
        }.unsafeRunSync()
    }

    override def beforeEach(context: BeforeEach): Unit = {
        val _ = flyway.migrate()
    }

    override def afterEach(context: AfterEach): Unit = {
        val _ = flyway.clean()
    }

    /** Find and return a free port on the local machine by starting a server socket and closing it. The port number
      * used by the socket is marked to allow reuse, considered free and returned.
      *
      * @return
      *   An optional port number if a free one can be found.
      */
    protected def findFreePort(): Option[Port] = {
        val socket = new ServerSocket(0)
        val port   = socket.getLocalPort
        socket.setReuseAddress(true) // Allow instant rebinding of the socket.
        socket.close()               // Free the socket for further use by closing it.
        Port.fromInt(port)
    }

    /** Provide a resource with a database connection to allow db operations and proper resource release later.
      *
      * @param cfg
      *   The application configuration.
      * @return
      *   A cats resource encapsulation a database connection as defined within the given configuration.
      */
    protected def connectToDb(cfg: SmedereeTicketsConfiguration): Resource[IO, java.sql.Connection] =
        Resource.make(
            IO.delay(java.sql.DriverManager.getConnection(cfg.database.url, cfg.database.user, cfg.database.pass))
        )(c => IO.delay(c.close()))

    /** Create a project for ticket tracking in the database.
      *
      * @param project
      *   The project to be created.
      * @return
      *   The number of affected database rows.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def createTicketsProject(project: Project): IO[Int] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay {
                    con.prepareStatement(
                        """INSERT INTO tickets.projects (name, owner, is_private, description, created_at, updated_at) VALUES(?, ?, ?, ?, NOW(), NOW())"""
                    )
                }
                _ <- IO.delay(statement.setString(1, project.name.toString))
                _ <- IO.delay(statement.setObject(2, project.owner.uid))
                _ <- IO.delay(statement.setBoolean(3, project.isPrivate))
                _ <- IO.delay(
                    project.description.fold(statement.setNull(4, java.sql.Types.VARCHAR))(descr =>
                        statement.setString(4, descr.toString)
                    )
                )
                r <- IO.delay(statement.executeUpdate())
                _ <- IO.delay(statement.close())
            } yield r
        }

    /** Create a user account from a ticket submitter in the database.
      *
      * @param submitter
      *   The submitter for which the account shall be created.
      * @return
      *   The number of affected database rows.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def createTicketsSubmitter(submitter: Submitter): IO[Int] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay {
                    con.prepareStatement(
                        """INSERT INTO tickets.users (uid, name, email, created_at, updated_at) VALUES(?, ?, ?, NOW(), NOW())"""
                    )
                }
                _ <- IO.delay(statement.setObject(1, submitter.id))
                _ <- IO.delay(statement.setString(2, submitter.name.toString))
                _ <- IO.delay(statement.setString(3, s"email-${submitter.id.toString}@example.com"))
                r <- IO.delay(statement.executeUpdate())
                _ <- IO.delay(statement.close())
            } yield r
        }

    /** Create a tickets user account in the database.
      *
      * @param owner
      *   The user to be created.
      * @return
      *   The number of affected database rows.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def createProjectOwner(owner: ProjectOwner): IO[Int] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay {
                    con.prepareStatement(
                        """INSERT INTO tickets.users (uid, name, email, created_at, updated_at) VALUES(?, ?, ?, NOW(), NOW())"""
                    )
                }
                _ <- IO.delay(statement.setObject(1, owner.uid))
                _ <- IO.delay(statement.setString(2, owner.name.toString))
                _ <- IO.delay(statement.setString(3, owner.email.toString))
                r <- IO.delay(statement.executeUpdate())
                _ <- IO.delay(statement.close())
            } yield r
        }

    /** Create a tickets user account in the database.
      *
      * @param user
      *   The user to be created.
      * @return
      *   The number of affected database rows.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def createTicketsUser(user: TicketsUser): IO[Int] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay {
                    con.prepareStatement(
                        """INSERT INTO tickets.users (uid, name, email, created_at, updated_at) VALUES(?, ?, ?, NOW(), NOW())"""
                    )
                }
                _ <- IO.delay(statement.setObject(1, user.uid))
                _ <- IO.delay(statement.setString(2, user.name.toString))
                _ <- IO.delay(statement.setString(3, user.email.toString))
                r <- IO.delay(statement.executeUpdate())
                _ <- IO.delay(statement.close())
            } yield r
        }

    /** Return the next ticket number for the given project.
      *
      * @param projectId
      *   The internal database ID of the project.
      * @return
      *   The next ticket number.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def loadNextTicketNumber(projectId: ProjectId): IO[Int] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT next_ticket_number FROM tickets.projects WHERE id = ?"""
                    )
                )
                _      <- IO.delay(statement.setLong(1, projectId.toLong))
                result <- IO.delay(statement.executeQuery)
                number <- IO.delay {
                    result.next()
                    result.getInt("next_ticket_number")
                }
                _ <- IO(statement.close())
            } yield number
        }

    /** Find the project ID for the given owner and project name.
      *
      * @param owner
      *   The unique ID of the user account that owns the project.
      * @param name
      *   The project name which must be unique in regard to the owner.
      * @return
      *   An option to the internal database ID.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def loadProjectId(owner: ProjectOwnerId, name: ProjectName): IO[Option[ProjectId]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT id FROM tickets.projects WHERE owner = ? AND name = ? LIMIT 1"""
                    )
                )
                _      <- IO.delay(statement.setObject(1, owner))
                _      <- IO.delay(statement.setString(2, name.toString))
                result <- IO.delay(statement.executeQuery)
                projectId <- IO.delay {
                    if (result.next()) {
                        ProjectId.from(result.getLong("id"))
                    } else {
                        None
                    }
                }
                _ <- IO(statement.close())
            } yield projectId
        }

    /** Find the ticket ID for the given project ID and ticket number.
      *
      * @param projectId
      *   The unique internal project id.
      * @param ticketNumber
      *   The ticket number.
      * @return
      *   An option to the internal database ID of the ticket.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def loadTicketId(project: ProjectId, number: TicketNumber): IO[Option[TicketId]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT id FROM tickets.tickets WHERE project = ? AND number = ? LIMIT 1"""
                    )
                )
                _      <- IO.delay(statement.setLong(1, project.toLong))
                _      <- IO.delay(statement.setInt(2, number.toInt))
                result <- IO.delay(statement.executeQuery)
                ticketId <- IO.delay {
                    if (result.next()) {
                        TicketId.from(result.getLong("id"))
                    } else {
                        None
                    }
                }
                _ <- IO(statement.close())
            } yield ticketId
        }

    /** Find the ticket service user with the given user id.
      *
      * @param uid
      *   The unique id of the user account.
      * @return
      *   An option to the loaded user.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def loadTicketsUser(uid: UserId): IO[Option[TicketsUser]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement("""SELECT uid, name, email, language FROM tickets.users WHERE uid = ?""")
                )
                _      <- IO.delay(statement.setObject(1, uid.toUUID))
                result <- IO.delay(statement.executeQuery())
                user <- IO.delay {
                    if (result.next()) {
                        val language = LanguageCode.from(result.getString("language"))
                        (
                            uid.some,
                            Username.from(result.getString("name")),
                            EmailAddress.from(result.getString("email"))
                        ).mapN { case (uid, name, email) =>
                            TicketsUser(uid, name, email, language)
                        }
                    } else {
                        None
                    }
                }
            } yield user
        }
}
