/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.time.*

import cats.effect.*
import cats.syntax.all.*
import de.smederee.TestTags.*
import de.smederee.tickets.Generators.*
import doobie.*

final class DoobieMilestoneRepositoryTest extends BaseSpec {

    /** Find the milestone ID for the given repository and milestone title.
      *
      * @param owner
      *   The unique ID of the user owner that owns the repository.
      * @param projectName
      *   The project name which must be unique in regard to the owner.
      * @param title
      *   The milestone title which must be unique in the repository context.
      * @return
      *   An option to the internal database ID.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def findMilestoneId(
        owner: ProjectOwnerId,
        projectName: ProjectName,
        title: MilestoneTitle
    ): IO[Option[Long]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT "milestones".id FROM "tickets"."milestones" AS "milestones" JOIN "tickets"."projects" AS "projects" ON "milestones".project = "projects".id WHERE "projects".owner = ? AND "projects".name = ? AND "milestones".title = ?"""
                    )
                )
                _      <- IO.delay(statement.setObject(1, owner))
                _      <- IO.delay(statement.setString(2, projectName.toString))
                _      <- IO.delay(statement.setString(3, title.toString))
                result <- IO.delay(statement.executeQuery)
                owner <- IO.delay {
                    if (result.next()) {
                        Option(result.getLong("id"))
                    } else {
                        None
                    }
                }
                _ <- IO(statement.close())
            } yield owner
        }

    test("allMilestones must return all milestones".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestones.sample) match {
            case (Some(owner), Some(generatedProject), Some(milestones)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val test = for {
                    _            <- createProjectOwner(owner)
                    createdRepos <- createTicketsProject(project)
                    repoId       <- loadProjectId(owner.uid, project.name)
                    createdMilestones <- repoId match {
                        case None => IO.pure(List.empty)
                        case Some(repoId) =>
                            milestones.traverse(milestone => milestoneRepo.createMilestone(repoId)(milestone))
                    }
                    foundMilestones <- repoId match {
                        case None         => IO.pure(List.empty)
                        case Some(repoId) => milestoneRepo.allMilestones(repoId).compile.toList
                    }
                } yield foundMilestones
                test.map { foundMilestones =>
                    assert(foundMilestones.size === milestones.size, "Different number of milestones!")
                    foundMilestones.sortBy(_.title).zip(milestones.sortBy(_.title)).map { (found, expected) =>
                        assertEquals(found.copy(id = expected.id), expected)
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("allTickets must return all tickets associated with the milestone".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestone.sample, genTickets.sample) match {
            case (Some(owner), Some(generatedProject), Some(milestone), Some(rawTickets)) =>
                val project  = generatedProject.copy(owner = owner)
                val tickets  = rawTickets.map(_.copy(submitter = None))
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val ticketRepo    = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _      <- createProjectOwner(owner)
                    _      <- createTicketsProject(project)
                    repoId <- loadProjectId(owner.uid, project.name)
                    foundTickets <- repoId match {
                        case None => IO.pure(List.empty)
                        case Some(projectId) =>
                            for {
                                _ <- milestoneRepo.createMilestone(projectId)(milestone)
                                _ <- tickets.traverse(ticket => ticketRepo.createTicket(projectId)(ticket))
                                createdMilestone <- milestoneRepo.findMilestone(projectId)(milestone.title)
                                _ <- createdMilestone match {
                                    case None => IO.pure(List.empty)
                                    case Some(milestone) =>
                                        tickets.traverse(ticket =>
                                            ticketRepo.addMilestone(projectId)(ticket.number)(milestone)
                                        )
                                }
                                foundTickets <- createdMilestone.map(_.id).getOrElse(None) match {
                                    case None              => IO.pure(List.empty)
                                    case Some(milestoneId) => milestoneRepo.allTickets(None)(milestoneId).compile.toList
                                }
                            } yield foundTickets
                    }
                } yield foundTickets
                test.map { foundTickets =>
                    assertEquals(foundTickets.size, tickets.size, "Different number of tickets!")
                    foundTickets.sortBy(_.number).zip(tickets.sortBy(_.number)).map { (found, expected) =>
                        assertEquals(
                            found.copy(createdAt = expected.createdAt, updatedAt = expected.updatedAt),
                            expected
                        )
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("closeMilestone must set the closed flag correctly".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestone.sample) match {
            case (Some(owner), Some(generatedProject), Some(generatedMilestone)) =>
                val milestone = generatedMilestone.copy(closed = false)
                val project   = generatedProject.copy(owner = owner)
                val dbConfig  = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val test = for {
                    _            <- createProjectOwner(owner)
                    createdRepos <- createTicketsProject(project)
                    repoId       <- loadProjectId(owner.uid, project.name)
                    milestones <- repoId match {
                        case None => IO.pure((None, None))
                        case Some(projectId) =>
                            for {
                                _      <- milestoneRepo.createMilestone(projectId)(milestone)
                                before <- milestoneRepo.findMilestone(projectId)(milestone.title)
                                _      <- before.flatMap(_.id).traverse(milestoneRepo.closeMilestone)
                                after  <- milestoneRepo.findMilestone(projectId)(milestone.title)
                            } yield (before, after)
                    }
                } yield milestones
                test.map { result =>
                    val (before, after) = result
                    val expected        = before.map(m => milestone.copy(id = m.id))
                    assertEquals(before, expected, "Test milestone not properly initialised!")
                    assertEquals(after, before.map(_.copy(closed = true)), "Test milestone not properly closed!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("createMilestone must create the milestone".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestone.sample) match {
            case (Some(owner), Some(generatedProject), Some(milestone)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val test = for {
                    _                 <- createProjectOwner(owner)
                    createdRepos      <- createTicketsProject(project)
                    repoId            <- loadProjectId(owner.uid, project.name)
                    createdMilestones <- repoId.traverse(id => milestoneRepo.createMilestone(id)(milestone))
                    foundMilestone    <- repoId.traverse(id => milestoneRepo.findMilestone(id)(milestone.title))
                } yield (createdRepos, repoId, createdMilestones, foundMilestone)
                test.map { tuple =>
                    val (createdRepos, repoId, createdMilestones, foundMilestone) = tuple
                    assert(createdRepos === 1, "Test vcs generatedProject was not created!")
                    assert(repoId.nonEmpty, "No vcs generatedProject id found!")
                    assert(createdMilestones.exists(_ === 1), "Test milestone was not created!")
                    foundMilestone.getOrElse(None) match {
                        case None => fail("Created milestone not found!")
                        case Some(foundMilestone) =>
                            assertEquals(foundMilestone, milestone.copy(id = foundMilestone.id))
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("createMilestone must fail if the milestone name already exists".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestone.sample) match {
            case (Some(owner), Some(generatedProject), Some(milestone)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val test = for {
                    _                 <- createProjectOwner(owner)
                    createdRepos      <- createTicketsProject(project)
                    repoId            <- loadProjectId(owner.uid, project.name)
                    createdMilestones <- repoId.traverse(id => milestoneRepo.createMilestone(id)(milestone))
                    _                 <- repoId.traverse(id => milestoneRepo.createMilestone(id)(milestone))
                } yield (createdRepos, repoId, createdMilestones)
                test.attempt.map(r => assert(r.isLeft, "Creating a milestone with a duplicate name must fail!"))
            case _ => fail("Could not generate data samples!")
        }
    }

    test("deleteMilestone must delete an existing milestone".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestone.sample) match {
            case (Some(owner), Some(generatedProject), Some(milestone)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val test = for {
                    _                 <- createProjectOwner(owner)
                    createdRepos      <- createTicketsProject(project)
                    repoId            <- loadProjectId(owner.uid, project.name)
                    createdMilestones <- repoId.traverse(id => milestoneRepo.createMilestone(id)(milestone))
                    milestoneId       <- findMilestoneId(owner.uid, project.name, milestone.title)
                    deletedMilestones <- milestoneRepo.deleteMilestone(
                        milestone.copy(id = milestoneId.flatMap(MilestoneId.from))
                    )
                    foundMilestone <- repoId.traverse(id => milestoneRepo.findMilestone(id)(milestone.title))
                } yield (createdRepos, repoId, createdMilestones, deletedMilestones, foundMilestone)
                test.map { tuple =>
                    val (createdRepos, repoId, createdMilestones, deletedMilestones, foundMilestone) = tuple
                    assert(createdRepos === 1, "Test vcs generatedProject was not created!")
                    assert(repoId.nonEmpty, "No vcs generatedProject id found!")
                    assert(createdMilestones.exists(_ === 1), "Test milestone was not created!")
                    assert(deletedMilestones === 1, "Test milestone was not deleted!")
                    assert(foundMilestone.getOrElse(None).isEmpty, "Test milestone was not deleted!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findMilestone must find existing milestones".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestones.sample) match {
            case (Some(owner), Some(generatedProject), Some(milestones)) =>
                val project           = generatedProject.copy(owner = owner)
                val dbConfig          = configuration.database
                val expectedMilestone = milestones(scala.util.Random.nextInt(milestones.size))
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val test = for {
                    _            <- createProjectOwner(owner)
                    createdRepos <- createTicketsProject(project)
                    repoId       <- loadProjectId(owner.uid, project.name)
                    createdMilestones <- repoId match {
                        case None => IO.pure(List.empty)
                        case Some(repoId) =>
                            milestones.traverse(milestone => milestoneRepo.createMilestone(repoId)(milestone))
                    }
                    foundMilestone <- repoId.traverse(id => milestoneRepo.findMilestone(id)(expectedMilestone.title))
                } yield foundMilestone.flatten
                test.map { foundMilestone =>
                    assertEquals(foundMilestone.map(_.copy(id = expectedMilestone.id)), Option(expectedMilestone))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("openMilestone must reset the closed flag correctly".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestone.sample) match {
            case (Some(owner), Some(generatedProject), Some(generatedMilestone)) =>
                val milestone = generatedMilestone.copy(closed = true)
                val project   = generatedProject.copy(owner = owner)
                val dbConfig  = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val test = for {
                    _            <- createProjectOwner(owner)
                    createdRepos <- createTicketsProject(project)
                    repoId       <- loadProjectId(owner.uid, project.name)
                    milestones <- repoId match {
                        case None => IO.pure((None, None))
                        case Some(projectId) =>
                            for {
                                _      <- milestoneRepo.createMilestone(projectId)(milestone)
                                before <- milestoneRepo.findMilestone(projectId)(milestone.title)
                                _      <- before.flatMap(_.id).traverse(milestoneRepo.openMilestone)
                                after  <- milestoneRepo.findMilestone(projectId)(milestone.title)
                            } yield (before, after)
                    }
                } yield milestones
                test.map { result =>
                    val (before, after) = result
                    val expected        = before.map(m => milestone.copy(id = m.id))
                    assertEquals(before, expected, "Test milestone not properly initialised!")
                    assertEquals(after, before.map(_.copy(closed = false)), "Test milestone not properly opened!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("updateMilestone must update an existing milestone".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestone.sample) match {
            case (Some(owner), Some(generatedProject), Some(milestone)) =>
                val updatedMilestone = milestone.copy(
                    title = MilestoneTitle("updated milestone"),
                    description = Option(MilestoneDescription("I am an updated milestone description...")),
                    dueDate = Option(LocalDate.of(1879, 3, 14))
                )
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val test = for {
                    _                 <- createProjectOwner(owner)
                    createdRepos      <- createTicketsProject(project)
                    repoId            <- loadProjectId(owner.uid, project.name)
                    createdMilestones <- repoId.traverse(id => milestoneRepo.createMilestone(id)(milestone))
                    milestoneId       <- findMilestoneId(owner.uid, project.name, milestone.title)
                    updatedMilestones <- milestoneRepo.updateMilestone(
                        updatedMilestone.copy(id = milestoneId.map(MilestoneId.apply))
                    )
                    foundMilestone <- repoId.traverse(id => milestoneRepo.findMilestone(id)(updatedMilestone.title))
                } yield (createdRepos, repoId, createdMilestones, updatedMilestones, foundMilestone.flatten)
                test.map { tuple =>
                    val (createdRepos, repoId, createdMilestones, updatedMilestones, foundMilestone) = tuple
                    assert(createdRepos === 1, "Test vcs generatedProject was not created!")
                    assert(repoId.nonEmpty, "No vcs generatedProject id found!")
                    assert(createdMilestones.exists(_ === 1), "Test milestone was not created!")
                    assert(updatedMilestones === 1, "Test milestone was not updated!")
                    assert(foundMilestone.nonEmpty, "Updated milestone not found!")
                    foundMilestone.map { milestone =>
                        assertEquals(milestone, updatedMilestone.copy(id = milestone.id))
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("updateMilestone must do nothing if id attribute is empty".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genMilestone.sample) match {
            case (Some(owner), Some(generatedProject), Some(milestone)) =>
                val updatedMilestone = milestone.copy(
                    id = None,
                    title = MilestoneTitle("updated milestone"),
                    description = Option(MilestoneDescription("I am an updated milestone description...")),
                    dueDate = Option(LocalDate.of(1879, 3, 14))
                )
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val test = for {
                    _                 <- createProjectOwner(owner)
                    createdRepos      <- createTicketsProject(project)
                    repoId            <- loadProjectId(owner.uid, project.name)
                    createdMilestones <- repoId.traverse(id => milestoneRepo.createMilestone(id)(milestone))
                    milestoneId       <- findMilestoneId(owner.uid, project.name, milestone.title)
                    updatedMilestones <- milestoneRepo.updateMilestone(updatedMilestone)
                } yield (createdRepos, repoId, createdMilestones, updatedMilestones)
                test.map { tuple =>
                    val (createdRepos, repoId, createdMilestones, updatedMilestones) = tuple
                    assert(createdRepos === 1, "Test vcs generatedProject was not created!")
                    assert(repoId.nonEmpty, "No vcs generatedProject id found!")
                    assert(createdMilestones.exists(_ === 1), "Test milestone was not created!")
                    assert(updatedMilestones === 0, "Milestone with empty id must not be updated!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }
}
