/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.time.OffsetDateTime

import cats.effect.*
import cats.syntax.all.*
import de.smederee.TestTags.*
import de.smederee.tickets.Generators.*
import doobie.*

import scala.collection.immutable.Queue

final class DoobieTicketRepositoryTest extends BaseSpec {

    /** Return the internal ids of all lables associated with the given ticket number and project id.
      *
      * @param projectId
      *   The unique internal project id.
      * @param ticketNumber
      *   The ticket number.
      * @return
      *   A list of label ids that may be empty.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    @SuppressWarnings(Array("DisableSyntax.var", "DisableSyntax.while"))
    protected def loadTicketLabelIds(projectId: ProjectId, ticketNumber: TicketNumber): IO[List[LabelId]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT label FROM "tickets"."ticket_labels" AS "ticket_labels" JOIN "tickets"."tickets" ON "ticket_labels".ticket = "tickets".id WHERE "tickets".project = ? AND "tickets".number = ?"""
                    )
                )
                _      <- IO.delay(statement.setLong(1, projectId.toLong))
                _      <- IO.delay(statement.setInt(2, ticketNumber.toInt))
                result <- IO.delay(statement.executeQuery)
                labelIds <- IO.delay {
                    var queue = Queue.empty[LabelId]
                    while (result.next())
                        queue = queue :+ LabelId(result.getLong("label"))
                    queue.toList
                }
                _ <- IO(statement.close())
            } yield labelIds
        }

    /** Return the internal ids of all milestones associated with the given ticket number and project id.
      *
      * @param projectId
      *   The unique internal project id.
      * @param ticketNumber
      *   The ticket number.
      * @return
      *   A list of milestone ids that may be empty.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    @SuppressWarnings(Array("DisableSyntax.var", "DisableSyntax.while"))
    protected def loadTicketMilestoneIds(projectId: ProjectId, ticketNumber: TicketNumber): IO[List[MilestoneId]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT milestone FROM "tickets"."milestone_tickets" AS "milestone_tickets" JOIN "tickets"."tickets" ON "milestone_tickets".ticket = "tickets".id WHERE "tickets".project = ? AND "tickets".number = ?"""
                    )
                )
                _      <- IO.delay(statement.setLong(1, projectId.toLong))
                _      <- IO.delay(statement.setInt(2, ticketNumber.toInt))
                result <- IO.delay(statement.executeQuery)
                milestoneIds <- IO.delay {
                    var queue = Queue.empty[MilestoneId]
                    while (result.next())
                        queue = queue :+ MilestoneId(result.getLong("milestone"))
                    queue.toList
                }
                _ <- IO(statement.close())
            } yield milestoneIds
        }

    test("addAssignee must save the assignee relation to the database".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genTicketsUser.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(user)) =>
                val assignee = Assignee(AssigneeId(user.uid.toUUID), AssigneeName(user.name.toString))
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- createTicketsUser(user)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    _         <- ticket.submitter.traverse(createTicketsSubmitter)
                    _         <- projectId.traverse(projectId => ticketRepo.createTicket(projectId)(ticket))
                    _ <- projectId.traverse(projectId => ticketRepo.addAssignee(projectId)(ticket.number)(assignee))
                    foundAssignees <- projectId.traverse(projectId =>
                        ticketRepo.loadAssignees(projectId)(ticket.number).compile.toList
                    )
                } yield foundAssignees.getOrElse(Nil)
                test.map { foundAssignees =>
                    assertEquals(foundAssignees, List(assignee))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("addLabel must save the label relation to the database".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genLabel.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(label)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo  = new DoobieLabelRepository[IO](tx)
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => fail("Project ID not found in database!")
                        case Some(projectId) =>
                            for {
                                _            <- ticket.submitter.traverse(createTicketsSubmitter)
                                _            <- labelRepo.createLabel(projectId)(label)
                                createdLabel <- labelRepo.findLabel(projectId)(label.name)
                                _            <- ticketRepo.createTicket(projectId)(ticket)
                                _ <- createdLabel.traverse(cl => ticketRepo.addLabel(projectId)(ticket.number)(cl))
                                foundLabels <- loadTicketLabelIds(projectId, ticket.number)
                            } yield (createdLabel, foundLabels)
                    }
                } yield result
                test.map { result =>
                    val (createdLabel, foundLabels) = result
                    assert(createdLabel.nonEmpty, "Test label not created!")
                    createdLabel.flatMap(_.id) match {
                        case None          => fail("Test label has no ID!")
                        case Some(labelId) => assert(foundLabels.exists(_ === labelId))
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("addMilestone must save the milestone relation to the database".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genMilestone.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(milestone)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val ticketRepo    = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => fail("Project ID not found in database!")
                        case Some(projectId) =>
                            for {
                                _                <- ticket.submitter.traverse(createTicketsSubmitter)
                                _                <- milestoneRepo.createMilestone(projectId)(milestone)
                                createdMilestone <- milestoneRepo.findMilestone(projectId)(milestone.title)
                                _                <- ticketRepo.createTicket(projectId)(ticket)
                                _ <- createdMilestone.traverse(cl =>
                                    ticketRepo.addMilestone(projectId)(ticket.number)(cl)
                                )
                                foundMilestones <- loadTicketMilestoneIds(projectId, ticket.number)
                            } yield (createdMilestone, foundMilestones)
                    }
                } yield result
                test.map { result =>
                    val (createdMilestone, foundMilestones) = result
                    assert(createdMilestone.nonEmpty, "Test milestone not created!")
                    createdMilestone.flatMap(_.id) match {
                        case None              => fail("Test milestone has no ID!")
                        case Some(milestoneId) => assert(foundMilestones.exists(_ === milestoneId))
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("allTickets must return all tickets for the project".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTickets.sample) match {
            case (Some(owner), Some(generatedProject), Some(generatedTickets)) =>
                val defaultTimestamp = OffsetDateTime.now()
                val tickets =
                    generatedTickets
                        .sortBy(_.number)
                        .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp))
                val submitters = tickets.map(_.submitter).flatten
                val project    = generatedProject.copy(owner = owner)
                val dbConfig   = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- submitters.traverse(createTicketsSubmitter)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => IO.pure((0, Nil))
                        case Some(projectId) =>
                            for {
                                writtenTickets <- tickets.traverse(ticket => ticketRepo.createTicket(projectId)(ticket))
                                foundTickets   <- ticketRepo.allTickets(filter = None)(projectId).compile.toList
                            } yield (writtenTickets.sum, foundTickets)
                    }
                } yield result
                test.map { result =>
                    val (writtenTickets, foundTickets) = result
                    assertEquals(writtenTickets, tickets.size, "Wrong number of tickets created in the database!")
                    assertEquals(
                        foundTickets.size,
                        writtenTickets,
                        "Number of returned tickets differs from number of created tickets!"
                    )
                    assertEquals(
                        foundTickets
                            .sortBy(_.number)
                            .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp)),
                        tickets.sortBy(_.number)
                    )
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("allTickets must respect given filters for numbers".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTickets.sample) match {
            case (Some(owner), Some(generatedProject), Some(generatedTickets)) =>
                val defaultTimestamp = OffsetDateTime.now()
                val tickets =
                    generatedTickets
                        .sortBy(_.number)
                        .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp))
                val expectedTickets = tickets.take(tickets.size / 2)
                val filter          = TicketFilter(number = expectedTickets.map(_.number), Nil, Nil, Nil)
                val submitters      = tickets.map(_.submitter).flatten
                val project         = generatedProject.copy(owner = owner)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- submitters.traverse(createTicketsSubmitter)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => IO.pure((0, Nil))
                        case Some(projectId) =>
                            for {
                                writtenTickets <- tickets.traverse(ticket => ticketRepo.createTicket(projectId)(ticket))
                                foundTickets   <- ticketRepo.allTickets(filter.some)(projectId).compile.toList
                            } yield (writtenTickets.sum, foundTickets)
                    }
                } yield result
                test.map { result =>
                    val (writtenTickets, foundTickets) = result
                    assertEquals(writtenTickets, tickets.size, "Wrong number of tickets created in the database!")
                    assertEquals(
                        foundTickets
                            .sortBy(_.number)
                            .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp)),
                        expectedTickets.sortBy(_.number)
                    )
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("allTickets must respect given filters for status".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTickets.sample) match {
            case (Some(owner), Some(generatedProject), Some(generatedTickets)) =>
                val defaultTimestamp = OffsetDateTime.now()
                val tickets =
                    generatedTickets
                        .sortBy(_.number)
                        .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp))
                val statusFlags     = tickets.map(_.status).distinct.take(2)
                val expectedTickets = tickets.filter(t => statusFlags.exists(_ === t.status))
                val filter          = TicketFilter(Nil, status = statusFlags, Nil, Nil)
                val submitters      = tickets.map(_.submitter).flatten
                val project         = generatedProject.copy(owner = owner)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- submitters.traverse(createTicketsSubmitter)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => IO.pure((0, Nil))
                        case Some(projectId) =>
                            for {
                                writtenTickets <- tickets.traverse(ticket => ticketRepo.createTicket(projectId)(ticket))
                                foundTickets   <- ticketRepo.allTickets(filter.some)(projectId).compile.toList
                            } yield (writtenTickets.sum, foundTickets)
                    }
                } yield result
                test.map { result =>
                    val (writtenTickets, foundTickets) = result
                    assertEquals(writtenTickets, tickets.size, "Wrong number of tickets created in the database!")
                    assertEquals(
                        foundTickets
                            .sortBy(_.number)
                            .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp)),
                        expectedTickets.sortBy(_.number)
                    )
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("allTickets must respect given filters for resolution".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTickets.sample) match {
            case (Some(owner), Some(generatedProject), Some(generatedTickets)) =>
                val defaultTimestamp = OffsetDateTime.now()
                val tickets =
                    generatedTickets
                        .sortBy(_.number)
                        .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp))
                val resolutions     = tickets.map(_.resolution).flatten.distinct.take(2)
                val expectedTickets = tickets.filter(t => t.resolution.exists(r => resolutions.exists(_ === r)))
                val filter          = TicketFilter(Nil, Nil, resolution = resolutions, Nil)
                val submitters      = tickets.map(_.submitter).flatten
                val project         = generatedProject.copy(owner = owner)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- submitters.traverse(createTicketsSubmitter)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => IO.pure((0, Nil))
                        case Some(projectId) =>
                            for {
                                writtenTickets <- tickets.traverse(ticket => ticketRepo.createTicket(projectId)(ticket))
                                foundTickets   <- ticketRepo.allTickets(filter.some)(projectId).compile.toList
                            } yield (writtenTickets.sum, foundTickets)
                    }
                } yield result
                test.map { result =>
                    val (writtenTickets, foundTickets) = result
                    assertEquals(writtenTickets, tickets.size, "Wrong number of tickets created in the database!")
                    assertEquals(
                        foundTickets
                            .sortBy(_.number)
                            .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp)),
                        expectedTickets.sortBy(_.number)
                    )
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("allTickets must respect given filters for submitter".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTickets.sample) match {
            case (Some(owner), Some(generatedProject), Some(generatedTickets)) =>
                val defaultTimestamp = OffsetDateTime.now()
                val tickets =
                    generatedTickets
                        .sortBy(_.number)
                        .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp))
                val submitters       = tickets.map(_.submitter).flatten
                val wantedSubmitters = submitters.take(submitters.size / 2)
                val expectedTickets  = tickets.filter(t => t.submitter.exists(s => wantedSubmitters.exists(_ === s)))
                val filter           = TicketFilter(Nil, Nil, Nil, submitter = wantedSubmitters.map(_.name))
                val project          = generatedProject.copy(owner = owner)
                val dbConfig         = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- submitters.traverse(createTicketsSubmitter)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => IO.pure((0, Nil))
                        case Some(projectId) =>
                            for {
                                writtenTickets <- tickets.traverse(ticket => ticketRepo.createTicket(projectId)(ticket))
                                foundTickets   <- ticketRepo.allTickets(filter.some)(projectId).compile.toList
                            } yield (writtenTickets.sum, foundTickets)
                    }
                } yield result
                test.map { result =>
                    val (writtenTickets, foundTickets) = result
                    assertEquals(writtenTickets, tickets.size, "Wrong number of tickets created in the database!")
                    assertEquals(
                        foundTickets
                            .sortBy(_.number)
                            .map(_.copy(createdAt = defaultTimestamp, updatedAt = defaultTimestamp)),
                        expectedTickets.sortBy(_.number)
                    )
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("createTicket must save the ticket to the database".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _           <- createProjectOwner(owner)
                    _           <- ticket.submitter.traverse(createTicketsSubmitter)
                    _           <- createTicketsProject(project)
                    projectId   <- loadProjectId(owner.uid, project.name)
                    _           <- projectId.traverse(projectId => ticketRepo.createTicket(projectId)(ticket))
                    foundTicket <- projectId.traverse(projectId => ticketRepo.findTicket(projectId)(ticket.number))
                } yield foundTicket.getOrElse(None)
                test.map { foundTicket =>
                    foundTicket match {
                        case None => fail("Created ticket not found!")
                        case Some(foundTicket) =>
                            assertEquals(
                                foundTicket,
                                ticket.copy(createdAt = foundTicket.createdAt, updatedAt = foundTicket.updatedAt)
                            )
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("deleteTicket must remove the ticket from the database".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _           <- createProjectOwner(owner)
                    _           <- ticket.submitter.traverse(createTicketsSubmitter)
                    _           <- createTicketsProject(project)
                    projectId   <- loadProjectId(owner.uid, project.name)
                    _           <- projectId.traverse(projectId => ticketRepo.createTicket(projectId)(ticket))
                    _           <- projectId.traverse(projectId => ticketRepo.deleteTicket(projectId)(ticket))
                    foundTicket <- projectId.traverse(projectId => ticketRepo.findTicket(projectId)(ticket.number))
                } yield foundTicket.getOrElse(None)
                test.map { foundTicket =>
                    assertEquals(foundTicket, None, "Ticket was not deleted from database!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findTicket must find existing tickets".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTickets.sample) match {
            case (Some(owner), Some(generatedProject), Some(tickets)) =>
                val expectedTicket = tickets(scala.util.Random.nextInt(tickets.size))
                val submitters     = tickets.map(_.submitter).flatten
                val project        = generatedProject.copy(owner = owner)
                val dbConfig       = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- submitters.traverse(createTicketsSubmitter)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    _ <- projectId match {
                        case None            => IO.pure(Nil)
                        case Some(projectId) => tickets.traverse(ticket => ticketRepo.createTicket(projectId)(ticket))
                    }
                    foundTicket <- projectId.traverse(projectId =>
                        ticketRepo.findTicket(projectId)(expectedTicket.number)
                    )
                } yield foundTicket.getOrElse(None)
                test.map { foundTicket =>
                    foundTicket match {
                        case None => fail("Ticket not found!")
                        case Some(foundTicket) =>
                            assertEquals(
                                foundTicket,
                                expectedTicket.copy(
                                    createdAt = foundTicket.createdAt,
                                    updatedAt = foundTicket.updatedAt
                                )
                            )
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findTicketId must find the unique internal id of existing tickets".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTickets.sample) match {
            case (Some(owner), Some(generatedProject), Some(tickets)) =>
                val expectedTicket = tickets(scala.util.Random.nextInt(tickets.size))
                val submitters     = tickets.map(_.submitter).flatten
                val project        = generatedProject.copy(owner = owner)
                val dbConfig       = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- submitters.traverse(createTicketsSubmitter)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => IO.pure((None, None))
                        case Some(projectId) =>
                            for {
                                _ <- tickets.traverse(ticket => ticketRepo.createTicket(projectId)(ticket))
                                expectedTicketId <- loadTicketId(projectId, expectedTicket.number)
                                foundTicketId    <- ticketRepo.findTicketId(projectId)(expectedTicket.number)
                            } yield (expectedTicketId, foundTicketId)
                    }
                } yield result
                test.map { result =>
                    val (expectedTicketId, foundTicketId) = result
                    assert(expectedTicketId.nonEmpty, "Expected ticket id not found!")
                    assertEquals(foundTicketId, expectedTicketId)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("loadAssignees must return all assignees of a ticket".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genTicketsUsers.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(users)) =>
                val assignees =
                    users.map(user => Assignee(AssigneeId(user.uid.toUUID), AssigneeName(user.name.toString)))
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- users.traverse(createTicketsUser)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    _         <- ticket.submitter.traverse(createTicketsSubmitter)
                    foundAssignees <- projectId match {
                        case None => fail("Project ID not found in database!")
                        case Some(projectId) =>
                            for {
                                _ <- ticketRepo.createTicket(projectId)(ticket)
                                _ <- assignees.traverse(assignee =>
                                    ticketRepo.addAssignee(projectId)(ticket.number)(assignee)
                                )
                                foundAssignees <- ticketRepo.loadAssignees(projectId)(ticket.number).compile.toList
                            } yield foundAssignees
                    }
                } yield foundAssignees
                test.map { foundAssignees =>
                    assertEquals(foundAssignees.sortBy(_.name), assignees.sortBy(_.name))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("loadLabels must return all labels of a ticket".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genLabels.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(labels)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo  = new DoobieLabelRepository[IO](tx)
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => fail("Project ID not found in database!")
                        case Some(projectId) =>
                            for {
                                _             <- ticket.submitter.traverse(createTicketsSubmitter)
                                _             <- ticketRepo.createTicket(projectId)(ticket)
                                _             <- labels.traverse(label => labelRepo.createLabel(projectId)(label))
                                createdLabels <- labelRepo.allLabels(projectId).compile.toList
                                _ <- createdLabels.traverse(cl => ticketRepo.addLabel(projectId)(ticket.number)(cl))
                                foundLabels <- ticketRepo.loadLabels(projectId)(ticket.number).compile.toList
                            } yield (createdLabels, foundLabels)
                    }
                } yield result
                test.map { result =>
                    val (createdLabels, foundLabels) = result
                    assertEquals(foundLabels.sortBy(_.id), createdLabels.sortBy(_.id))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("loadMilestones must return all milestones of a ticket".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genMilestones.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(milestones)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val ticketRepo    = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    result <- projectId match {
                        case None => fail("Project ID not found in database!")
                        case Some(projectId) =>
                            for {
                                _ <- ticket.submitter.traverse(createTicketsSubmitter)
                                _ <- ticketRepo.createTicket(projectId)(ticket)
                                _ <- milestones.traverse(milestone =>
                                    milestoneRepo.createMilestone(projectId)(milestone)
                                )
                                createdMilestones <- milestoneRepo.allMilestones(projectId).compile.toList
                                _ <- createdMilestones.traverse(cm =>
                                    ticketRepo.addMilestone(projectId)(ticket.number)(cm)
                                )
                                foundMilestones <- ticketRepo.loadMilestones(projectId)(ticket.number).compile.toList
                            } yield (createdMilestones, foundMilestones)
                    }
                } yield result
                test.map { result =>
                    val (createdMilestones, foundMilestones) = result
                    assertEquals(foundMilestones.sortBy(_.title), createdMilestones.sortBy(_.title))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("removeAssignee must remove the assignees from the ticket".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genTicketsUser.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(user)) =>
                val assignee = Assignee(AssigneeId(user.uid.toUUID), AssigneeName(user.name.toString))
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- createTicketsUser(user)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    _         <- ticket.submitter.traverse(createTicketsSubmitter)
                    foundAssignees <- projectId match {
                        case None => IO.pure(Nil)
                        case Some(projectId) =>
                            for {
                                _              <- ticketRepo.createTicket(projectId)(ticket)
                                _              <- ticketRepo.addAssignee(projectId)(ticket.number)(assignee)
                                _              <- ticketRepo.removeAssignee(projectId)(ticket)(assignee)
                                foundAssignees <- ticketRepo.loadAssignees(projectId)(ticket.number).compile.toList
                            } yield foundAssignees
                    }
                } yield foundAssignees
                test.map { foundAssignees =>
                    assertEquals(foundAssignees, Nil)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("removeLabel must remove the label from the ticket".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genLabel.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(label)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo  = new DoobieLabelRepository[IO](tx)
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    foundLabels <- projectId match {
                        case None => fail("Project ID not found in database!")
                        case Some(projectId) =>
                            for {
                                _            <- ticket.submitter.traverse(createTicketsSubmitter)
                                _            <- labelRepo.createLabel(projectId)(label)
                                createdLabel <- labelRepo.findLabel(projectId)(label.name)
                                _            <- ticketRepo.createTicket(projectId)(ticket)
                                _ <- createdLabel.traverse(cl => ticketRepo.addLabel(projectId)(ticket.number)(cl))
                                _ <- createdLabel.traverse(cl => ticketRepo.removeLabel(projectId)(ticket)(cl))
                                foundLabels <- loadTicketLabelIds(projectId, ticket.number)
                            } yield foundLabels
                    }
                } yield foundLabels
                test.map { foundLabels =>
                    assertEquals(foundLabels, Nil)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("removeMilestone must remove the milestone from the ticket".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genMilestone.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(milestone)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val milestoneRepo = new DoobieMilestoneRepository[IO](tx)
                val ticketRepo    = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    foundMilestones <- projectId match {
                        case None => fail("Project ID not found in database!")
                        case Some(projectId) =>
                            for {
                                _                <- ticket.submitter.traverse(createTicketsSubmitter)
                                _                <- milestoneRepo.createMilestone(projectId)(milestone)
                                createdMilestone <- milestoneRepo.findMilestone(projectId)(milestone.title)
                                _                <- ticketRepo.createTicket(projectId)(ticket)
                                _ <- createdMilestone.traverse(ms =>
                                    ticketRepo.addMilestone(projectId)(ticket.number)(ms)
                                )
                                _ <- createdMilestone.traverse(ms => ticketRepo.removeMilestone(projectId)(ticket)(ms))
                                foundMilestones <- loadTicketMilestoneIds(projectId, ticket.number)
                            } yield foundMilestones
                    }
                } yield foundMilestones
                test.map { foundMilestones =>
                    assertEquals(foundMilestones, Nil)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("updateTicket must update the ticket in the database".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genTicket.sample, genTicket.sample) match {
            case (Some(owner), Some(generatedProject), Some(ticket), Some(anotherTicket)) =>
                val project = generatedProject.copy(owner = owner)
                val updatedTicket =
                    ticket.copy(
                        title = anotherTicket.title,
                        content = anotherTicket.content,
                        submitter = anotherTicket.submitter
                    )
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val ticketRepo = new DoobieTicketRepository[IO](tx)
                val test = for {
                    _           <- createProjectOwner(owner)
                    _           <- ticket.submitter.traverse(createTicketsSubmitter)
                    _           <- updatedTicket.submitter.traverse(createTicketsSubmitter)
                    _           <- createTicketsProject(project)
                    projectId   <- loadProjectId(owner.uid, project.name)
                    _           <- projectId.traverse(projectId => ticketRepo.createTicket(projectId)(ticket))
                    _           <- projectId.traverse(projectId => ticketRepo.updateTicket(projectId)(updatedTicket))
                    foundTicket <- projectId.traverse(projectId => ticketRepo.findTicket(projectId)(ticket.number))
                } yield foundTicket.getOrElse(None)
                test.map { foundTicket =>
                    foundTicket match {
                        case None => fail("Created ticket not found!")
                        case Some(foundTicket) =>
                            assertEquals(
                                foundTicket,
                                updatedTicket.copy(createdAt = foundTicket.createdAt, updatedAt = foundTicket.updatedAt)
                            )
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

}
