/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.i18n

import cats.*

import scala.util.matching.Regex

opaque type LanguageCode = String
object LanguageCode {
    given Eq[LanguageCode] = Eq.fromUniversalEquals

    val FormatIso639: Regex = "^[a-z]{2,3}$".r

    /** Create an instance of LanguageCode from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a LanguageCode.
      * @return
      *   The appropriate instance of LanguageCode.
      */
    def apply(source: String): LanguageCode = source

    /** Try to create an instance of LanguageCode from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a LanguageCode.
      * @return
      *   An option to the successfully converted LanguageCode.
      */
    def from(source: String): Option[LanguageCode] =
        Option(source).map(FormatIso639.matches) match {
            case Some(true) => Option(source)
            case _          => None
        }

    extension (code: LanguageCode) {

        /** Convert the language code into a [[java.util.Locale]] to be used for internationalisation functions.
          *
          * @return
          *   A locale retrieved via the `forLanguageTag` method of the Java Locale implementation.
          */
        def toLocale: java.util.Locale = java.util.Locale.forLanguageTag(code.toString)
    }
}
