/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.i18n

import java.text.MessageFormat
import java.util.Locale
import java.util.ResourceBundle

import org.slf4j.LoggerFactory

import scala.util.Failure
import scala.util.Success
import scala.util.Try

object Messages {
    private val log = LoggerFactory.getLogger(getClass)

    val DefaultLocale: Locale = Locale.ENGLISH
    val Filename: String      = "messages"

    /** Provide a way to simply write `Messages("my.key", foo, bar)` to translate stuff. This function will catch
      * possible exceptions underneath and in case of an error will log an error message and the underlying exception
      * and will return the given message string instead a translation!
      *
      * @param message
      *   The message to be translated, which should be a valid property key.
      * @param args
      *   A list of arguments which will be used in the [[java.text.MessageFormat]] to fill in placeholders.
      * @param locale
      *   The implicitly provided locale which determines which resource bundle will be used.
      * @return
      *   The translated and formatted (placeholders filled in) string.
      */
    def apply(message: String, args: Any*)(using locale: Locale): String =
        Try {
            val usedLocale =
                Option(locale).getOrElse(DefaultLocale) // Fallback to default locale if provided one is null.
            val formatter = new MessageFormat(getProperty(usedLocale)(message), usedLocale)
            formatter.format(args.toArray)
        } match {
            case Failure(error) =>
                log.error(s"Error during translation of '$message' for locale $locale!", error)
                message
            case Success(translation) => translation
        }

    /** Provide a way to simply write `Messages("my.key", foo, bar)` to translate stuff.
      *
      * @param message
      *   The message to be translated, which should be a valid property key.
      * @param args
      *   A list of arguments which will be used in the [[java.text.MessageFormat]] to fill in placeholders.
      * @param locale
      *   The implicitly provided locale which determines which resource bundle will be used.
      * @return
      *   The translated and formatted (placeholders filled in) string.
      */
    @throws[java.util.MissingResourceException]("if the message key cannot be found in the resource bundle")
    def unsafeApply(message: String, args: Any*)(using locale: Locale): String = {
        val usedLocale =
            Option(locale).getOrElse(DefaultLocale) // Fallback to default locale if provided one is null.
        val formatter = new MessageFormat(getProperty(usedLocale)(message), usedLocale)
        formatter.format(args.toArray)
    }

    /** Return the string stored in the properties bundle for the given locale and key.
      *
      * If the resource bundle for the desired locale cannot be found then the function falls back to the default one.
      *
      * @param locale
      *   The locale for which the string shall be returned.
      * @param key
      *   A key which must be present in the resource bundle of the given locale.
      * @return
      *   The string defined in the resource bundle of the locale at the given key.
      */
    @throws[java.lang.NullPointerException]("if base filename or locale is null")
    @throws[java.util.MissingResourceException]("if the message key cannot be found in the resource bundle")
    @throws[java.lang.ClassCastException]("if the object found for the given key is not a string")
    def getProperty(locale: Locale)(key: String): String = {
        val filename = s"messages_${locale.getLanguage.toString}.properties"
        log.debug("Checking for resource bundle: {}", filename)
        val usedLocale = Option(getClass.getClassLoader.getResource(filename)).map(_ => locale).getOrElse(DefaultLocale)
        log.debug("Using resource bundle for locale: {}", usedLocale)
        val bundle = ResourceBundle.getBundle(Filename, usedLocale)
        bundle.getString(key)
    }
}
