/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.i18n

import java.util.Locale
import java.util.MissingResourceException

import munit.*

import org.scalacheck.*
import org.scalacheck.Prop.*

final class MessagesTest extends ScalaCheckSuite {
    val genLocale: Gen[Locale] = Gen.oneOf(Locale.getAvailableLocales.toList)
    given Arbitrary[Locale]    = Arbitrary(genLocale)

    property("Messages must fall back to default locale if expected one is missing") {
        forAll { (locale: Locale) =>
            given Locale = locale
            val message  = "test.main.only"
            val expected = "This should only exist in the main properties file."
            val obtained = Messages(message)
            assertEquals(obtained, expected)
        }
    }

    test("Messages must translate correctly") {
        given Locale = Locale.GERMAN
        val message  = "test.title"
        val expected = "Das ist eine deutsche Übersetzung!"
        val obtained = Messages(message)
        assertEquals(obtained, expected)
    }

    property("Messages must return the given key if it is not defined in the resource bundle") {
        forAll { (message: String) =>
            given Locale    = Messages.DefaultLocale
            val translation = Messages(message)
            assertEquals(translation, message)
        }
    }

    property(
        "Messages.unsafeApply must fail with MissingResourceException if key is not defined in the resource bundle"
    ) {
        forAll { (message: String) =>
            given Locale = Messages.DefaultLocale
            intercept[MissingResourceException] {
                Messages.unsafeApply(message)
            }
            assert(true) // Workaround for the intercept part.
        }
    }

}
