/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.security

import cats.*
import cats.syntax.all.*

/** Generic permissions on resources used to provide privileges to users.
  *
  * @param encoded
  *   An internal value used for decoding and encoding permissions as well as permission sets. This follows the octal
  *   approach of encoding POSIX file system permissions and the idea to store only encoded permission sets in storage
  *   layers (database) to allow faster querying for permissions.
  */
enum Permission(val encoded: Int) {

    /** Allow executing tasks related to the data e.g. start build pipeline. */
    case Execute extends Permission(encoded = 1)

    /** Allow reading data (non modifying operations) e.g. view a repository or organisation. */
    case Read extends Permission(encoded = 4)

    /** Allow writing data including deletion e.g. push commits, delete a project or modify an organisation. */
    case Write extends Permission(encoded = 2)
}

object Permission {
    given Eq[Permission] = Eq.instance((a, b) => a.encoded === b.encoded)

    /** Decode the given integer value into a set of permissions. Illegal values that cannot be mapped will return an
      * empty set.
      *
      * @param encoded
      *   An integer value that holds an encoded permission set.
      * @return
      *   A set of permissions that may be empty.
      */
    def decode(encoded: Int): Set[Permission] =
        encoded match {
            case 0 => Set.empty
            case 1 => Set(Execute)
            case 2 => Set(Write)
            case 3 => Set(Execute, Write)
            case 4 => Set(Read)
            case 5 => Set(Execute, Read)
            case 6 => Set(Read, Write)
            case 7 => Set(Execute, Read, Write)
            case _ => Set.empty // We return an empty set for illegal values!
        }

    /** Return the permission named after the given string.
      *
      * @param name
      *   A string containing the name of a permission e.g. Read.
      * @return
      *   An option to the permission instance.
      */
    def from(name: String): Option[Permission] = Permission.values.find(_.toString === name)

    /** Return the permission that is encoded by the given integer value.
      *
      * @param encoded
      *   An integer value that holds a single encoded permission.
      * @return
      *   An option to the permission instance.
      */
    def fromInt(encoded: Int): Option[Permission] = Permission.values.find(_.encoded === encoded)

    extension (ps: Set[Permission]) {

        /** Encode this set of permissions into an integer value.
          *
          * @return
          *   An integer value containing the encoded representation of this set of permissions.
          */
        def encode: Int = ps.foldLeft(0)((encodedSet, permission) => encodedSet | permission.encoded)
    }
}
