/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.security

import cats.*
import cats.data.*
import cats.syntax.all.*

import scala.util.matching.Regex

/** A username for an account has to obey several restrictions which are similiar to the ones found for Unix usernames.
  * It must start with a letter, contain only alphanumeric characters, be at least 2 and at most 31 characters long and
  * be all lowercase.
  */
opaque type Username = String
object Username {
    given Eq[Username]       = Eq.fromUniversalEquals
    given Ordering[Username] = (x: Username, y: Username) => x.compareTo(y)
    given Order[Username]    = Order.fromOrdering

    val MinimumLength: Int    = 2
    val MaximumLength: Int    = 31
    val isAlphanumeric: Regex = "^[a-z][a-z0-9]+$".r

    /** Create an instance of Username from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a Username.
      * @return
      *   The appropriate instance of Username.
      */
    def apply(source: String): Username = source

    /** Try to create an instance of Username from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a Username.
      * @return
      *   An option to the successfully converted Username.
      */
    def from(s: String): Option[Username] = validate(s).toOption

    /** Validate the given string and return either the validated username or a list of errors.
      *
      * @param s
      *   An arbitrary string which should be a username.
      * @return
      *   Either a list of errors or the validated username.
      */
    def validate(s: String): ValidatedNec[String, Username] =
        Option(s).map(_.trim.nonEmpty) match {
            case Some(true) =>
                val input = s.trim
                val miniumLength =
                    if (input.length >= MinimumLength)
                        input.validNec
                    else
                        s"Username too short (min. $MinimumLength characters)!".invalidNec
                val maximumLength =
                    if (input.length <= MaximumLength)
                        input.validNec
                    else
                        s"Username too long (max. $MaximumLength characters)!".invalidNec
                val alphanumeric =
                    if (isAlphanumeric.matches(input))
                        input.validNec
                    else
                        "Username must be all lowercase alphanumeric characters and start with a character.".invalidNec
                (miniumLength, maximumLength, alphanumeric).mapN { case (_, _, name) =>
                    name
                }
            case _ => "Username must not be empty!".invalidNec
        }
}
